Package: ErrorTracer
Type: Package
Title: Bayesian Error Propagation and Forecast Uncertainty
        Decomposition
Version: 1.1.0
Date: 2026-05-23
Authors@R: c(
    person(given = "Luis Javier", family = "Madrigal-Roca", role = c("aut", "cre"), email = "madrigalrocalj@yahoo.com"),
    person(given = "John", family = "Kelly", role = "aut", email = "jkk@ku.edu")
  )
Description: Provides a full pipeline from regularized or standard regression
    models (elastic net, linear models, generalized linear models, random
    forests) to informed Bayesian priors, structured forecast uncertainty
    decomposition (parameter / environmental / residual, plus a temporal
    component when the model carries an autocorrelation term), and forecast
    shelf life analysis (the quantification of when a forecast becomes
    uninformative). Designed for ecological and genomic forecasting with
    climate or environmental covariates. Methods build on Bürkner (2017)
    <doi:10.18637/jss.v080.i01> for Bayesian regression via 'Stan',
    Friedman, Hastie, and Tibshirani (2010) <doi:10.18637/jss.v033.i01>
    for elastic net regularization, Wright and Ziegler (2017)
    <doi:10.18637/jss.v077.i01> for random forests, and Vehtari, Gelman,
    and Gabry (2017) <doi:10.1007/s11222-016-9696-4> for leave-one-out
    cross-validation.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Depends: R (>= 4.1.0)
Imports: brms (>= 2.20.0), ggplot2 (>= 3.4.0), rlang (>= 1.1.0), stats,
        utils
Suggests: loo (>= 2.6.0), bayesplot (>= 1.10.0), scales (>= 1.3.0),
        tidyr (>= 1.3.0), glmnet (>= 4.1.0), ranger (>= 0.15.0), dplyr
        (>= 1.1.0), testthat (>= 3.0.0), knitr, rmarkdown, covr
Config/testthat/edition: 3
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-25 12:51:48 UTC; madrigalrocalj
Author: Luis Javier Madrigal-Roca [aut, cre],
  John Kelly [aut]
Maintainer: Luis Javier Madrigal-Roca <madrigalrocalj@yahoo.com>
Repository: CRAN
Date/Publication: 2026-05-25 13:00:02 UTC
Built: R 4.6.0; ; 2026-05-25 15:30:42 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(decompose_uncertainty,default)
S3method(decompose_uncertainty,et_prediction)
S3method(decompose_uncertainty,et_prediction_list)
S3method(et_calibrate,et_prediction)
S3method(et_calibrate,et_prediction_list)
S3method(et_diagnose,et_model)
S3method(et_diagnose,et_model_list)
S3method(et_plot_coefficients,et_model)
S3method(et_plot_coefficients,et_model_list)
S3method(et_predict,et_model)
S3method(et_predict,et_model_list)
S3method(extract_priors,cv.glmnet)
S3method(extract_priors,glm)
S3method(extract_priors,glmnet)
S3method(extract_priors,lm)
S3method(extract_priors,ranger)
S3method(print,et_model)
S3method(print,et_model_list)
S3method(print,et_prediction)
S3method(print,et_prediction_list)
S3method(print,et_prior_spec)
S3method(print,et_sensitivity)
S3method(print,et_shelf_life)
S3method(shelf_life,default)
S3method(shelf_life,et_prediction)
S3method(shelf_life,et_prediction_list)
S3method(summary,et_model)
S3method(summary,et_model_list)
S3method(summary,et_shelf_life)
export(decompose_uncertainty)
export(et_calibrate)
export(et_diagnose)
export(et_fit)
export(et_plot_calibration)
export(et_plot_coefficients)
export(et_plot_decomposition)
export(et_plot_forecast)
export(et_plot_prior_posterior)
export(et_plot_sensitivity)
export(et_plot_shelf_life)
export(et_predict)
export(et_sensitivity_profile)
export(et_theme)
export(extract_priors)
export(shelf_life)
export(standardize)
export(unstandardize)
importFrom(rlang,.data)
importFrom(stats,coef)
importFrom(stats,lm)
importFrom(stats,median)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,var)
importFrom(utils,head)

# ErrorTracer: Bayesian Error Propagation and Forecast Uncertainty Decomposition

## Overview

<img src="images/logo.png" align="right" height="200" alt="ErrorTracer logo" />

**ErrorTracer** provides a complete pipeline for ecological and genomic forecasting with climate or environmental covariates. It bridges the gap between regularized regression and fully Bayesian workflows in three steps:

1. **Extract informed priors** from a fitted regularized or standard model (`glmnet`, `lm`, `glm`, `ranger`) and carry those estimates forward as prior means into a Bayesian model — rather than discarding them.
2. **Decompose forecast uncertainty** into three independent sources — parameter uncertainty, environmental/covariate measurement uncertainty, and residual process variance — plus a fourth temporal-autocorrelation component whenever the model formula carries an `ar()`, `ma()`, `arma()`, `cosy()`, `unstr()`, `sar()`, or `car()` term.
3. **Quantify forecast shelf life** — the time horizon at which a forecast's credible interval becomes wider than the biologically plausible response range, rendering it uninformative. This concept has no equivalent in any existing R package.

The package is designed for ecological and genomic time-series forecasting in which a regularized (or plain) regression model is refit as a Bayesian model and predictions must be accompanied by a principled uncertainty budget. The API is fully general — response, predictors, and grouping are formula-driven.

---

## Key Features

| Feature | Description |
|---|---|
| **Regularized → Bayesian prior pipeline** | `extract_priors()` converts `glmnet`, `lm`, `glm`, or `ranger` coefficients into `brms`-compatible priors |
| **Uncertainty decomposition** | Partitions forecast variance into parameter / environmental / residual components, with a fourth temporal-autocorrelation component added automatically when the model formula contains `ar()`, `ma()`, `arma()`, `cosy()`, `unstr()`, `sar()`, or `car()` |
| **Forecast shelf life** | Quantifies the exact time point at which a forecast becomes uninformative |
| **Group-level fitting** | One function call fits and predicts across all groups (e.g., SNP clusters, species, sites) |
| **Calibration assessment** | Observed vs. nominal coverage probability at multiple CI levels |
| **Rich visualizations** | Six `ggplot2`-based plotting functions, all returning customizable objects |
| **Full S3 class system** | Consistent `print`, `summary`, and `plot`-family methods across all objects |

---

## Installation

Install the development version from GitHub using `remotes`:

```r
# install.packages("remotes")
remotes::install_github("madrigalrocalj/ErrorTracer")
```

ErrorTracer requires a Stan backend. Install either `cmdstanr` (recommended) or `rstan` separately:

```r
# Option A: cmdstanr (recommended)
install.packages("cmdstanr", repos = c("https://mc-stan.org/r-packages/", getOption("repos")))
cmdstanr::install_cmdstan()

# Option B: rstan
install.packages("rstan")
```

---

## Core Workflow

### Step 1 — Extract informed priors from a regularized model

`extract_priors()` is an S3 generic that dispatches on the class of the fitted model. Coefficient estimates (or importance weights for `ranger`) are used as prior means; prior SDs are proportional to coefficient magnitude or standard error.

```r
library(ErrorTracer)
library(glmnet)

# --- Elastic net (cv.glmnet) ---
cv_fit <- cv.glmnet(x_train, y_train, alpha = 0.5)

prior_spec <- extract_priors(
  model              = cv_fit,
  lambda             = "lambda.min",  # or "lambda.1se"
  multiplier         = 2.0,           # prior SD = multiplier * |coef|
  min_sd             = 0.1,           # floor for prior SD
  intercept_prior_sd = 2.5,           # intercept prior SD
  sigma_prior_scale  = 1.0            # half-Cauchy scale for sigma
)

print(prior_spec)
# ErrorTracer prior specification
#   Method      : glmnet
#   Predictors  : 5
#   Multiplier  : 2
#   Min SD      : 0.1
#   Coefficients:
#     Tmean                 mean =   0.3241  sd =   0.6482
#     PPT                   mean =  -0.1087  sd =   0.2174
#     ...
```

**Supported model classes and their prior construction:**

| Class | Prior mean | Prior SD |
|---|---|---|
| `cv.glmnet` / `glmnet` | Non-zero coefficients | `multiplier × |coef|` (floor: `min_sd`) |
| `lm` | Coefficient estimates | `multiplier × SE` (floor: `min_sd`) |
| `glm` | Coefficient estimates | `multiplier × SE` (floor: `min_sd`) |
| `ranger` | Zero (direction unknown) | `multiplier × normalised_importance` |

For `ranger`, only variables with positive permutation importance are included. Because random forests do not produce signed coefficients, priors are centered at zero and scaled by importance — encoding "this predictor matters" without a directional claim.

> **Note on centering:** The glmnet intercept is intentionally excluded from the prior — it reflects the centering of training data, which may not apply to new data. Center predictors upstream and document this in your analysis.

---

### Step 2 — Fit the Bayesian model

`et_fit()` wraps `brms::brm()` and attaches the prior specification and training data for downstream use.

```r
fit <- et_fit(
  formula       = z_diff ~ Tmean + PPT + SWE,
  data          = train_df,
  priors        = prior_spec,       # from extract_priors(); or NULL for brms defaults
  chains        = 4L,
  iter          = 2000L,
  warmup        = 1000L,
  cores         = 4L,
  seed          = 42L,
  adapt_delta   = 0.95,
  max_treedepth = 12L
)

print(fit)
# ErrorTracer model (et_model)
#   Formula : z_diff ~ Tmean + PPT + SWE
#   n obs   : 48
#   Chains  : 4  Iter: 2000  Warmup: 1000
#   Priors  : informed (glmnet, 3 predictors)
#   Rhat max: 1.002

summary(fit)
```

#### Group-level fitting

Pass a `grouping` column name to fit one model per group and receive an `et_model_list`:

```r
# Fits one brms model per SNP cluster
fit_list <- et_fit(
  formula  = z_diff ~ Tmean + PPT + SWE,
  data     = all_clusters_df,
  priors   = prior_spec,        # single spec applied to all groups
  grouping = "cluster_id",      # or pass a named list of et_prior_spec for per-group priors
  chains   = 4L,
  iter     = 2000L
)

print(fit_list)
# ErrorTracer grouped model list (et_model_list)
#   Grouping : cluster_id
#   Formula  : z_diff ~ Tmean + PPT + SWE
#   Groups   : 12
#   Fitted   : 12 / 12
```

---

### Step 3 — Predict with uncertainty propagation

`et_predict()` draws from the posterior predictive distribution, propagates environmental measurement noise through the model, and computes credible intervals.

```r
pred <- et_predict(
  model     = fit,
  newdata   = forecast_df,
  env_noise = list(Tmean = 0.5, PPT = 10.0, SWE = 5.0),
  # env_noise can also be a single fraction of each predictor's SD:
  # env_noise = 0.10   →  10% noise applied uniformly
  n_draws   = 2000L,
  ci_levels = c(0.5, 0.8, 0.90, 0.95),
  n_perturb = 500L    # draws used in the perturbation step; reduce for speed
)

print(pred)
# ErrorTracer prediction (et_prediction)
#   Observations : 24
#   Draws        : 2000
#   CI levels    : 0.5, 0.8, 0.9, 0.95
#   Mean var decomposition (across observations):
#     Parameter    : 0.0412
#     Environmental: 0.0083
#     Residual     : 0.1204
#     Total        : 0.1619
```

The returned `et_prediction` object contains:

| Slot | Description |
|---|---|
| `posterior_predict` | `[n_draws × n_obs]` matrix — full posterior predictive draws |
| `posterior_linpred` | `[n_draws × n_obs]` matrix — linear predictor draws (parameter uncertainty only) |
| `lp_perturbed` | `[n_perturb × n_obs]` matrix — linear predictor under environmentally perturbed inputs |
| `sigma_draws` | Numeric vector of posterior sigma draws |
| `credible_intervals` | `data.frame` with columns `row_id, ci_level, lower, median, upper, width` |
| `decomposition` | `data.frame` with `obs_id, param_var, env_var, residual_var, total_var` (plus `temporal_var` when the model carries an autocorrelation term) |
| `newdata` | The input forecast data |
| `model` | Reference to the source `et_model` |

---

### Step 4 — Decompose uncertainty

`decompose_uncertainty()` extracts the variance decomposition table from a prediction object:

```r
decomp <- decompose_uncertainty(pred)
head(decomp)
#   obs_id param_var  env_var residual_var total_var
# 1      1    0.0389  0.0071       0.1204    0.1601
# 2      2    0.0441  0.0098       0.1204    0.1637
# ...
```

**Decomposition method:**

| Component | Estimator |
|---|---|
| `param_var` | `Var[posterior_linpred]` across draws — coefficient uncertainty |
| `env_var` | `Var[lp_perturbed] − Var[lp_unperturbed]` — additional variance from predictor noise |
| `residual_var` | `E[σ²]` across posterior draws — biological process noise |
| `total_var` | `Var[posterior_predict]` — full predictive variance |

All components are guaranteed non-negative. Using the posterior mean of σ² (not the median) ensures `param_var + residual_var ≈ total_var` by the law of total variance.

For models whose formula carries an autocorrelation term (`ar()`, `ma()`, `arma()`, `cosy()`, `unstr()`, `sar()`, `car()`), `decompose_uncertainty()` adds a fourth column `temporal_var`, defined as `pmax(0, total_var − (param_var + env_var + residual_var))`. It captures the autocorrelation-induced spread that `brms::posterior_predict()` accumulates iteratively beyond a single innovation, and `residual_var` is then to be read as the *innovation* variance rather than the stationary marginal variance. The four components reconstruct `total_var` modulo Monte Carlo error.

---

### Step 5 — Forecast shelf life

`shelf_life()` computes the ratio of credible interval width to the plausible response range at each forecast time point, and flags when the forecast becomes uninformative:

```r
# For arcsin-sqrt transformed responses, derive range from training data:
#   plausible_range = range(train_df$z_diff)
# For bounded responses, supply explicit bounds:
#   plausible_range = c(lower, upper)
sl <- shelf_life(
  predictions              = pred,
  plausible_range          = range(train_df$z_diff),
  ci_level                 = 0.90,   # must be present in the et_prediction object
  threshold                = 1.0,    # CI/range ratio above which forecast is uninformative
  time_col                 = "year", # column in newdata to use as time axis
  max_extrapolation_factor = 10      # cap on linear projection beyond observed window
)

print(sl)
# ErrorTracer shelf life analysis
#   Observations    : 24
#   Plausible range : 2
#   Informative     : 18 / 24
#   Mean CI/range   : 0.847
#   Max CI/range    : 1.231

as.data.frame(sl)
#   obs_id  time ci_width plausible_range  ratio informative
# 1      1  2024    0.841               2  0.421        TRUE
# 2      2  2025    1.103               2  0.552        TRUE
# 3      3  2026    2.461               2  1.231       FALSE
```

A `threshold` of 1.0 (default) means the forecast is considered uninformative when the CI spans the entire plausible response range. Lower thresholds (e.g., 0.8) impose stricter requirements.

---

### Step 6 — Calibrate against validation data

`et_calibrate()` computes observed coverage probability at each nominal CI level. A well-calibrated model produces coverage that matches the nominal level:

```r
cal <- et_calibrate(
  predictions  = pred,
  observed     = validation_df,  # same n_rows as newdata; matched positionally
  response_col = "z_diff",       # inferred from formula if omitted
  ci_levels    = c(0.5, 0.8, 0.90, 0.95)
)

print(cal)
#   ci_level nominal observed_coverage n_obs calibration_error
# 1     0.50    0.50             0.542    24             0.042
# 2     0.80    0.80             0.792    24             0.008
# 3     0.90    0.90             0.875    24             0.025
# 4     0.95    0.95             0.958    24             0.008
```

---

### Step 7 — Diagnose convergence

`et_diagnose()` reports Rhat, effective sample size ratios, divergent transitions, and LOO-CV in a single call:

```r
diag <- et_diagnose(fit, loo = TRUE)

# Convergence
diag$convergence$rhat_max     # should be < 1.05
diag$convergence$neff_all_ok  # all Neff ratios > 0.1
diag$convergence$n_divergences

# LOO-CV
diag$loo$elpd_loo
diag$loo$n_bad_pareto_k  # Pareto k > 0.7 indicate influential observations
```

For grouped models, `et_diagnose()` returns a `per_group` list and a `summary` data.frame with one row per group.

---

## Visualization

All plotting functions return `ggplot2` objects that can be further customized with standard `ggplot2` additions (`+ theme(...)`, `+ labs(...)`, etc.).

```r
# Stacked bar: proportional contribution of each variance component
et_plot_decomposition(decomp, proportional = TRUE)

# Line chart: CI width / plausible range over time, with uninformative threshold
et_plot_shelf_life(sl, show_ratio = TRUE)

# Calibration: observed vs. nominal coverage (points on the 1:1 diagonal = well-calibrated)
et_plot_calibration(cal)

# Fan chart: nested CI ribbons with optional observed values overlaid
et_plot_forecast(pred, observed = validation_df,
                 response_col = "z_diff", time_col = "year")

# Density overlay: prior vs. posterior for each regression coefficient
et_plot_prior_posterior(fit, max_preds = 8L)

# Forest plot: Bayesian 95% CI (blue) vs. regularized coefficient (red ×)
et_plot_coefficients(fit)
```

---

## S3 Class System

| Class | Created by | Methods |
|---|---|---|
| `et_prior_spec` | `extract_priors()` | `print` |
| `et_model` | `et_fit()` | `print`, `summary` |
| `et_model_list` | `et_fit(..., grouping = ...)` | `print`, `summary` |
| `et_prediction` | `et_predict()` | `print` |
| `et_prediction_list` | `et_predict()` on `et_model_list` | `print` |
| `et_shelf_life` | `shelf_life()` | `print`, `summary` |

---

## Full Example: Allele Frequency Forecasting

```r
library(ErrorTracer)
library(glmnet)

# --- Prepare data ---
# train_df: historical years with columns z_diff, Tmean, PPT, SWE
# forecast_df: future years with predictor values only

# 1. Fit elastic net
x_mat <- as.matrix(train_df[, c("Tmean", "PPT", "SWE")])
cv_fit <- cv.glmnet(x_mat, train_df$z_diff, alpha = 0.5)

# 2. Extract informed priors
prior_spec <- extract_priors(cv_fit, multiplier = 2.0, min_sd = 0.1)

# 3. Fit Bayesian model
fit <- et_fit(
  formula = z_diff ~ Tmean + PPT + SWE,
  data    = train_df,
  priors  = prior_spec,
  chains  = 4L, iter = 2000L, cores = 4L
)

# 4. Predict with environmental noise propagation
pred <- et_predict(
  model     = fit,
  newdata   = forecast_df,
  env_noise = list(Tmean = 0.5, PPT = 10.0, SWE = 5.0),
  ci_levels = c(0.5, 0.80, 0.90, 0.95)
)

# 5. Decompose uncertainty
decomp <- decompose_uncertainty(pred)

# 6. Assess forecast shelf life
# For arcsin-sqrt responses use the observed training range; for bounded
# responses supply explicit bounds.
sl <- shelf_life(
  pred,
  plausible_range          = range(train_df$z_diff),
  ci_level                 = 0.90,
  threshold                = 1.0,
  time_col                 = "year",
  max_extrapolation_factor = 10
)

# 7. Calibrate (if validation data available)
cal <- et_calibrate(pred, observed = valid_df, response_col = "z_diff")

# 8. Diagnose
et_diagnose(fit)

# 9. Visualize
library(ggplot2)
et_plot_decomposition(decomp)
et_plot_shelf_life(sl)
et_plot_calibration(cal)
et_plot_forecast(pred, observed = valid_df,
                 response_col = "z_diff", time_col = "year")
et_plot_prior_posterior(fit)
et_plot_coefficients(fit)
```

---

## Group-Level Analysis Example

```r
# Fit one model per SNP cluster, then predict and summarize shelf life by cluster
fit_list <- et_fit(
  formula  = z_diff ~ Tmean + PPT + SWE,
  data     = all_df,
  priors   = prior_spec,
  grouping = "cluster_id",
  chains = 4L, iter = 2000L, cores = 4L
)

pred_list <- et_predict(
  fit_list,
  newdata   = forecast_df,
  env_noise = list(Tmean = 0.5, PPT = 10.0),
  ci_levels = c(0.50, 0.90, 0.95)
)

# Per-group decomposition and shelf life
decomp_all <- decompose_uncertainty(pred_list)
sl_all     <- shelf_life(pred_list,
                         plausible_range          = range(train_df$z_diff),
                         ci_level                 = 0.90,
                         time_col                 = "year",
                         max_extrapolation_factor = 10)

# Per-group diagnostics summary
diag_list  <- et_diagnose(fit_list)
diag_list$summary  # data.frame: rhat_ok, neff_ok, n_divergences, elpd_loo per cluster

# Faceted shelf life plot
et_plot_shelf_life(sl_all) +
  ggplot2::facet_wrap(~ group, ncol = 4)
```

---

## Comparison with Existing R Packages

| Package | What it does | What is missing |
|---|---|---|
| `brms` | Bayesian regression | No shelf life, no decomposition, no enet→prior pipeline |
| `propagate` | Analytical error propagation | No posterior sampling, no Bayesian framework |
| `rstanarm` | Bayesian regression | Same gaps as `brms` |
| `forecast` / `fable` | Time series forecasting | No ecological/genomic data structures, no decomposition |
| `INLA` | Spatial/temporal Bayes | No shelf life concept, very different user base |

**ErrorTracer's niche:** the full pipeline from regularized feature selection → informed Bayesian refitting → structured uncertainty decomposition → forecast shelf life, designed for ecology and genomics.

---

## Package Structure

```
ErrorTracer/
├── R/
│   ├── priors.R      # extract_priors() — dispatch for glmnet, lm, glm, ranger
│   ├── fit.R         # et_fit() — Bayesian model fitting via brms
│   ├── predict.R     # et_predict() — posterior prediction with decomposition
│   ├── decompose.R   # decompose_uncertainty() — param / env / residual split
│   ├── shelf_life.R  # shelf_life() — forecast horizon analysis
│   ├── calibrate.R   # et_calibrate(), et_diagnose()
│   ├── plot.R        # et_plot_*() — six ggplot2 visualizations
│   └── utils.R       # internal helpers (standardize, logging, etc.)
├── tests/
│   └── testthat/
├── vignettes/
│   ├── genomics.Rmd  # Drosophila allele frequency example
│   └── ecology.Rmd   # Phenology / species abundance example
├── DESCRIPTION
├── NAMESPACE
```

---

## Dependencies

**Imports** (installed automatically):

| Package | Version | Purpose |
|---|---|---|
| `brms` | >= 2.20.0 | Bayesian regression via Stan |
| `loo` | >= 2.6.0 | LOO cross-validation |
| `bayesplot` | >= 1.10.0 | Posterior predictive checks |
| `ggplot2` | >= 3.4.0 | Visualization |
| `tidyr` | >= 1.3.0 | Data reshaping |
| `rlang` | >= 1.1.0 | Tidy evaluation |

**Suggests** (install separately as needed):

| Package | Purpose |
|---|---|
| `glmnet` | Elastic net / lasso prior extraction |
| `ranger` | Random forest prior extraction |
| `dplyr` | Data manipulation in vignettes |
| `testthat` | Unit testing |
| `knitr`, `rmarkdown` | Vignette building |

**Backend** (required; install separately):
- `cmdstanr` (recommended) or `rstan`

---

## Caveats and Limitations

- **Small-n regime:** When `n < 10` observations, informed priors from `extract_priors()` dominate the posterior. This is the setting where the enet→prior pipeline provides the most value, but is also most sensitive to prior misspecification. Document prior choices carefully.
- **Intercept handling:** The glmnet intercept is excluded from the prior by default, as it reflects upstream centering of training data. If your data is not pre-centered, verify that the intercept prior (default `Normal(0, 2.5)`) is appropriate.
- **Environmental uncertainty is additive:** `env_var` is estimated via perturbation and is reported separately from `total_var`, which is computed from the full posterior predictive (parameter + residual only). Interpret the decomposition table accordingly.
- **`ranger` priors are undirected:** Because random forests provide no signed coefficients, ranger-derived priors are centered at zero. This is conservative but avoids false directional assumptions.

---

## License

MIT License. See `LICENSE` for details.

---

## Authors

- **Luis Javier Madrigal-Roca** (Creator, Maintainer) — madrigalrocalj@yahoo.com
- **John Kelly** (Author) — jkk@ku.edu

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = FALSE,  # Stan/brms chunks are too slow for automated builds
  fig.width = 7,
  fig.height = 4.5
)

## ----libraries----------------------------------------------------------------
# library(ErrorTracer)
# library(glmnet)      # for elastic net (Suggests)
# library(ggplot2)
# 
# # Load the bundled simulated dataset
# data(et_sim)

## ----demo-settings------------------------------------------------------------
# # Use small MCMC settings for this tutorial.
# # In a real analysis: CHAINS = 4, ITER = 2000 (or more).
# CHAINS <- 2L
# ITER   <- 1000L
# SEED   <- 42L

## ----data-overview------------------------------------------------------------
# str(et_sim, max.level = 2)
# # List of 7
# #  $ train         :'data.frame': 40 obs. of 6 variables:
# #  $ forecast      :'data.frame': 10 obs. of 5 variables:
# #  $ validation    :'data.frame': 10 obs. of 3 variables:
# #  $ true_params   :List of 2
# #  $ env_noise     :List of 3
# #  $ standardization:List of 3
# #  $ description   : chr ...

## ----head-train---------------------------------------------------------------
# head(et_sim$train)
# #   year cluster_id      Tmean        PPT        SWE     z_diff
# # 1 1995          A -0.6775963  0.9119435 -0.1386249 -0.4917161
# # 2 1996          A  0.4866542 -1.7284276 -1.4391959  0.4088745
# # 3 1997          A -0.4935765 -0.3945083 -0.7267905  0.2352018
# # 4 1998          A -0.2397691 -0.9805886 -0.8206722  0.3194194
# # 5 1999          A -0.5193449  1.1304845  1.6453631 -0.2554886
# # 6 2000          A -1.5333598  0.6437605  0.6252043 -0.5864645

## ----true-params--------------------------------------------------------------
# # True coefficients — known because the data are simulated
# et_sim$true_params
# # $A
# #   intercept     Tmean       PPT       SWE     sigma
# #        0.00      0.50     -0.30      0.20      0.40
# #
# # $B
# #   intercept     Tmean       PPT       SWE     sigma
# #        0.10      0.30     -0.20     -0.10      0.50

## ----forecast-predictors------------------------------------------------------
# # Forecast predictors are slightly above training range in Tmean (warming trend)
# et_sim$forecast[et_sim$forecast$cluster_id == "A", ]
# #   year cluster_id     Tmean       PPT        SWE
# # 1 2015          A 0.6918714  0.291...  0.102...
# # 2 2016          A 0.7012... ...
# # ...
# # Forecast Tmean is ~0.7-1.4 SDs above training mean — mild extrapolation.

## ----plot-training-data-------------------------------------------------------
# ggplot(et_sim$train, aes(x = year, y = z_diff, colour = cluster_id)) +
#   geom_line() +
#   geom_point(size = 2) +
#   labs(
#     title   = "Simulated allele-frequency change (training period)",
#     x       = "Year",
#     y       = expression(italic(z)[diff]),
#     colour  = "Cluster"
#   ) +
#   et_theme()

## ----plot-predictors----------------------------------------------------------
# # Climate predictors over the full period — warming trend visible in Tmean
# all_years <- rbind(
#   transform(et_sim$train[et_sim$train$cluster_id == "A",
#                           c("year", "Tmean", "PPT", "SWE")],
#             period = "Training"),
#   transform(et_sim$forecast[et_sim$forecast$cluster_id == "A",
#                              c("year", "Tmean", "PPT", "SWE")],
#             period = "Forecast")
# )
# 
# all_long <- reshape(all_years, direction = "long",
#                     varying = c("Tmean", "PPT", "SWE"),
#                     v.names = "value", timevar = "predictor",
#                     times   = c("Tmean", "PPT", "SWE"))
# 
# ggplot(all_long, aes(x = year, y = value,
#                      colour = period, group = period)) +
#   geom_line() +
#   geom_point(size = 1.8) +
#   facet_wrap(~ predictor, ncol = 3) +
#   labs(
#     title  = "Climate predictors: training and forecast periods",
#     x      = "Year", y = "Standardised value", colour = "Period"
#   ) +
#   et_theme()

## ----enet-fit-----------------------------------------------------------------
# train_A <- et_sim$train[et_sim$train$cluster_id == "A", ]
# 
# x_train <- as.matrix(train_A[, c("Tmean", "PPT", "SWE")])
# y_train <- train_A$z_diff
# 
# set.seed(42)
# cv_fit_A <- glmnet::cv.glmnet(
#   x     = x_train,
#   y     = y_train,
#   alpha = 0.5          # 0 = ridge, 1 = lasso, 0.5 = elastic net
# )
# 
# # Coefficient estimates at lambda.min
# coef(cv_fit_A, s = "lambda.min")
# # 4 x 1 sparse Matrix of class "dgCMatrix"
# #             lambda.min
# # (Intercept)  0.0430
# # Tmean        0.6658
# # PPT         -0.4591
# # SWE          0.3775

## ----enet-plot----------------------------------------------------------------
# plot(cv_fit_A)
# title("Cross-validated elastic net — Cluster A", line = 2.5)

## ----extract-priors-----------------------------------------------------------
# prior_spec_A <- extract_priors(
#   model      = cv_fit_A,
#   lambda     = "lambda.min",
#   multiplier = 2.0,   # prior SD = 2 × |elastic-net coef|
#   min_sd     = 0.10   # floor: avoids spike priors on near-zero coefficients
# )
# 
# print(prior_spec_A)
# # ErrorTracer prior specification
# #   Method      : glmnet
# #   Predictors  : 3
# #   Multiplier  : 2
# #   Min SD      : 0.1
# #   Coefficients:
# #     Tmean                 mean =   0.6658  sd =   1.3316
# #     PPT                   mean =  -0.4591  sd =   0.9182
# #     SWE                   mean =   0.3775  sd =   0.7550

## ----et-fit-------------------------------------------------------------------
# fit_A <- et_fit(
#   formula = z_diff ~ Tmean + PPT + SWE,
#   data    = train_A,
#   priors  = prior_spec_A,
#   chains  = CHAINS,
#   iter    = ITER,
#   warmup  = ITER %/% 2L,
#   cores   = CHAINS,
#   seed    = SEED
# )
# 
# print(fit_A)
# # ErrorTracer model (et_model)
# #   Formula : z_diff ~ Tmean + PPT + SWE
# #   n obs   : 20
# #   Chains  : 2  Iter: 1000  Warmup: 500
# #   Priors  : informed (glmnet, 3 predictors)
# #   Rhat max: 1.004

## ----et-summary---------------------------------------------------------------
# summary(fit_A)
# # === ErrorTracer model summary ===
# #
# # --- Fixed effects ---
# #              Estimate  Est.Error     Q2.5    Q97.5
# # Intercept    0.041      0.104    -0.162    0.246
# # Tmean        0.666      0.113     0.438    0.888
# # PPT         -0.487      0.200    -0.909   -0.123
# # SWE          0.406      0.206     0.023    0.827

## ----et-diagnose--------------------------------------------------------------
# diag_A <- et_diagnose(fit_A, loo = TRUE)
# 
# # Convergence
# diag_A$convergence
# # $rhat_max
# # [1] 1.004         # all chains well below 1.05 threshold
# #
# # $rhat_all_ok
# # [1] TRUE
# #
# # $neff_min
# # [1] 0.275         # lowest effective-sample-size ratio
# #
# # $neff_all_ok
# # [1] TRUE          # all Neff ratios > 0.10
# #
# # $n_divergences
# # [1] 0             # no divergent transitions

## ----loo-results--------------------------------------------------------------
# diag_A$loo[c("elpd_loo", "p_loo", "looic", "n_bad_pareto_k")]
# # $elpd_loo
# # [1] -14.55
# #
# # $p_loo
# # [1] 4.00          # effective parameters ≈ 4 (intercept, 3 betas); sigma
# #                   # is implicit — consistent with a 4-parameter model
# # $looic
# # [1] 29.10
# #
# # $n_bad_pareto_k
# # [1] 0             # no influential observations (Pareto k > 0.7)

## ----et-predict---------------------------------------------------------------
# fcast_A <- et_sim$forecast[et_sim$forecast$cluster_id == "A", ]
# 
# pred_A <- et_predict(
#   model     = fit_A,
#   newdata   = fcast_A,
#   env_noise = et_sim$env_noise,   # list(Tmean = 0.30, PPT = 0.20, SWE = 0.15)
#   n_draws   = 1000L,
#   ci_levels = c(0.50, 0.80, 0.90, 0.95),
#   n_perturb = 300L                # draws for the perturbation step
# )
# 
# print(pred_A)
# # ErrorTracer prediction (et_prediction)
# #   Observations : 5
# #   Draws        : 1000
# #   CI levels    : 0.5, 0.8, 0.9, 0.95
# #   Mean var decomposition (across observations):
# #     Parameter    : 0.0441
# #     Environmental: 0.0579
# #     Residual     : 0.2173
# #     Total        : 0.2673

## ----ci-table-----------------------------------------------------------------
# # 90% credible intervals by forecast year
# ci90 <- pred_A$credible_intervals[pred_A$credible_intervals$ci_level == 0.90, ]
# ci90$year <- fcast_A$year
# ci90[, c("year", "lower", "median", "upper", "width")]
# #   year    lower  median   upper  width
# # 1 2015   -0.507   0.381   1.213  1.720
# # 2 2016    0.009   0.794   1.660  1.651
# # 3 2017   -0.009   0.875   1.631  1.640
# # 4 2018   -0.179   0.649   1.460  1.639
# # 5 2019   -0.767   0.131   1.002  1.769

## ----decompose----------------------------------------------------------------
# decomp_A <- decompose_uncertainty(pred_A)
# decomp_A
# #   obs_id total_var param_var  env_var residual_var
# # 1      1    0.2627    0.0202   0.0541       0.2173
# # 2      2    0.2667    0.0361   0.0588       0.2173
# # 3      3    0.2573    0.0364   0.0541       0.2173
# # 4      4    0.2529    0.0427   0.0653       0.2173
# # 5      5    0.2972    0.0849   0.0574       0.2173

## ----plot-decomp--------------------------------------------------------------
# et_plot_decomposition(decomp_A, proportional = TRUE) +
#   labs(subtitle = "Cluster A, forecast 2015–2019")

## ----shelf-life---------------------------------------------------------------
# # Derive plausible range from training data (arcsin-sqrt scale — unbounded)
# pr_A <- range(train_A$z_diff)   # e.g. c(-1.50, 2.04) => width ~ 3.54
# 
# sl_A <- shelf_life(
#   predictions              = pred_A,
#   plausible_range          = pr_A,
#   ci_level                 = 0.90,
#   threshold                = 1.0,        # uninformative when CI width >= plausible range
#   time_col                 = "year",
#   min_slope_for_projection = 1e-4,       # positive slope required to project
#   max_extrapolation_factor = 10          # cap: project <= last_year + 10 * window_width
# )
# 
# print(sl_A)
# # ErrorTracer shelf life analysis
# #   Observations     : 5
# #   Plausible range  : 3.54
# #   Threshold        : 1
# #   Informative      : 5 / 5
# #   Mean CI/range    : 0.488
# #   Max CI/range     : 0.499
# #   Shelf life       : > 2019 (lower bound — all periods informative, no trend)
# #   Detail           : All 5 forecast periods informative. Linear trend
# #                      (slope = 0.00241 per time unit) projects threshold
# #                      crossing at ~2234.3, but this exceeds the extrapolation
# #                      cap (2059). Shelf life > 2019.

## ----shelf-life-table---------------------------------------------------------
# as.data.frame(sl_A)
# #   obs_id  time ci_width plausible_range  ratio informative
# # 1      1  2015    1.720            3.54  0.486        TRUE
# # 2      2  2016    1.651            3.54  0.466        TRUE
# # 3      3  2017    1.640            3.54  0.463        TRUE
# # 4      4  2018    1.639            3.54  0.463        TRUE
# # 5      5  2019    1.769            3.54  0.499        TRUE

## ----plot-shelf-life----------------------------------------------------------
# et_plot_shelf_life(sl_A, show_ratio = TRUE) +
#   labs(subtitle = "Cluster A — 90% CI width / training range (arcsin-sqrt scale)")

## ----calibrate----------------------------------------------------------------
# valid_A <- et_sim$validation[et_sim$validation$cluster_id == "A", ]
# 
# cal_A <- et_calibrate(
#   predictions  = pred_A,
#   observed     = valid_A,
#   response_col = "z_diff",
#   ci_levels    = c(0.50, 0.80, 0.90, 0.95)
# )
# 
# print(cal_A)
# #   ci_level nominal observed_coverage n_obs calibration_error
# # 1     0.50    0.50              0.60     5              0.10
# # 2     0.80    0.80              0.80     5              0.00
# # 3     0.90    0.90              1.00     5              0.10
# # 4     0.95    0.95              1.00     5              0.05

## ----plot-calibration---------------------------------------------------------
# et_plot_calibration(cal_A) +
#   labs(subtitle = "Cluster A — posterior predictive calibration")

## ----plot-forecast------------------------------------------------------------
# et_plot_forecast(
#   predictions  = pred_A,
#   observed     = valid_A,
#   response_col = "z_diff",
#   time_col     = "year"
# ) +
#   labs(subtitle = "Cluster A — shaded ribbons: 50%, 80%, 90%, 95% CI")

## ----plot-prior-posterior-----------------------------------------------------
# et_plot_prior_posterior(fit_A, max_preds = 3L) +
#   labs(subtitle = "Cluster A — how 20 observations update the elastic-net prior")

## ----plot-coefficients--------------------------------------------------------
# et_plot_coefficients(fit_A) +
#   labs(subtitle = "Cluster A — Bayesian 95% CI (blue) vs. enet coefficient (red ×)")

## ----parameter-recovery-------------------------------------------------------
# post    <- brms::fixef(fit_A$fit)
# true_A  <- et_sim$true_params$A
# 
# post_df <- data.frame(
#   parameter  = c("Intercept", "Tmean", "PPT", "SWE"),
#   true_value = c(true_A["intercept"], true_A["Tmean"],
#                  true_A["PPT"],       true_A["SWE"]),
#   post_mean  = post[, "Estimate"],
#   lower_95   = post[, "Q2.5"],
#   upper_95   = post[, "Q97.5"]
# )
# post_df$covered <- with(post_df,
#                         true_value >= lower_95 & true_value <= upper_95)
# 
# post_df[, c("parameter", "true_value", "post_mean", "lower_95", "upper_95", "covered")]
# #   parameter true_value post_mean lower_95 upper_95 covered
# # 1 Intercept       0.00     0.041   -0.162    0.246    TRUE
# # 2     Tmean       0.50     0.666    0.438    0.888    TRUE
# # 3       PPT      -0.30    -0.487   -0.909   -0.123    TRUE
# # 4       SWE       0.20     0.406    0.023    0.827    TRUE

## ----grouped-priors-----------------------------------------------------------
# # Build one et_prior_spec per cluster from individual elastic nets
# prior_list <- lapply(c("A", "B"), function(cl) {
#   df <- et_sim$train[et_sim$train$cluster_id == cl, ]
#   x  <- as.matrix(df[, c("Tmean", "PPT", "SWE")])
#   set.seed(42)
#   suppressWarnings(cv <- glmnet::cv.glmnet(x, df$z_diff, alpha = 0.5))
#   extract_priors(cv, multiplier = 2.0, min_sd = 0.10)
# })
# names(prior_list) <- c("A", "B")

## ----grouped-fit--------------------------------------------------------------
# # One brms model per cluster, returned as et_model_list
# fit_all <- et_fit(
#   formula  = z_diff ~ Tmean + PPT + SWE,
#   data     = et_sim$train,
#   priors   = prior_list,       # named list: one et_prior_spec per cluster
#   grouping = "cluster_id",
#   chains   = CHAINS,
#   iter     = ITER,
#   warmup   = ITER %/% 2L,
#   cores    = CHAINS,
#   seed     = SEED
# )
# 
# print(fit_all)
# # ErrorTracer grouped model list (et_model_list)
# #   Grouping : cluster_id
# #   Formula  : z_diff ~ Tmean + PPT + SWE
# #   Groups   : 2
# #   Fitted   : 2 / 2
# 
# summary(fit_all)
# # --- Per-group Rhat max ---
# #   A                     Rhat max = 1.004
# #   B                     Rhat max = 1.007

## ----grouped-diagnose---------------------------------------------------------
# diag_all <- et_diagnose(fit_all, loo = TRUE)
# diag_all$summary
# #   group rhat_ok neff_ok n_divergences elpd_loo n_bad_pareto_k
# # 1     A    TRUE    TRUE             0   -14.55              0
# # 2     B    TRUE    TRUE             0   -17.15              1

## ----grouped-predict----------------------------------------------------------
# pred_all <- et_predict(
#   model     = fit_all,
#   newdata   = et_sim$forecast,
#   env_noise = et_sim$env_noise,
#   n_draws   = 1000L,
#   ci_levels = c(0.50, 0.80, 0.90, 0.95),
#   n_perturb = 300L
# )
# 
# print(pred_all)
# # ErrorTracer grouped predictions (et_prediction_list)
# #   Grouping : cluster_id
# #   Groups   : 2 / 2 predicted

## ----grouped-decompose--------------------------------------------------------
# decomp_all <- decompose_uncertainty(pred_all)
# head(decomp_all)
# #   group obs_id total_var param_var  env_var residual_var
# # 1     A      1    0.2627    0.0202   0.0541       0.2173
# # 2     A      2    0.2667    0.0361   0.0588       0.2173
# # ...
# # 6     B      1    0.3374    0.0252   0.0122       0.2999
# # 7     B      2    0.3472    0.0488   0.0180       0.2999

## ----grouped-shelf-life-------------------------------------------------------
# # Use the pooled training range across both clusters (arcsin-sqrt scale)
# pr_all <- range(et_sim$train$z_diff)   # e.g. c(-1.67, 2.04) => width ~ 3.71
# 
# sl_all <- shelf_life(
#   pred_all,
#   plausible_range          = pr_all,
#   ci_level                 = 0.90,
#   threshold                = 1.0,
#   time_col                 = "year",
#   max_extrapolation_factor = 10
# )
# 
# print(sl_all)
# # ErrorTracer shelf life analysis
# #   Observations     : 10
# #   Plausible range  : 3.71
# #   Threshold        : 1
# #   [A]  100.0% informative  mean ratio = 0.465  shelf life > 2019 (lower bound)
# #   [B]  100.0% informative  mean ratio = 0.544  shelf life > 2019 (lower bound)

## ----plot-grouped-shelf-life--------------------------------------------------
# et_plot_shelf_life(sl_all, show_ratio = TRUE) +
#   labs(
#     subtitle = "Both clusters — 90% CI width / plausible range",
#     colour   = "Cluster"
#   )

## ----grouped-calibrate--------------------------------------------------------
# cal_all <- et_calibrate(
#   pred_all,
#   observed     = et_sim$validation,
#   response_col = "z_diff",
#   ci_levels    = c(0.50, 0.80, 0.90, 0.95)
# )
# 
# et_plot_calibration(cal_all) +
#   labs(subtitle = "Both clusters — posterior predictive calibration")

## ----grouped-coef-plot--------------------------------------------------------
# # Forest plot for all groups in a single faceted figure
# et_plot_coefficients(fit_all) +
#   labs(subtitle = "Bayesian 95% CI (blue) vs. enet coefficient (red ×)")

## ----gcm-workflow, eval = FALSE-----------------------------------------------
# # Suppose you have GCM ensemble-mean predictors for 2025-2075,
# # standardised using your training-period statistics.
# gcm_years <- 2025:2075
# 
# gcm_df <- data.frame(
#   year  = gcm_years,
#   Tmean = ...,    # standardised GCM ensemble-mean temperature
#   PPT   = ...,    # standardised GCM ensemble-mean precipitation
#   SWE   = ...     # standardised GCM ensemble-mean snowpack
# )
# 
# # GCM ensemble spread (SD across runs) grows roughly linearly with time.
# # Here we model spread as starting at the validation-period noise level
# # and increasing by ~0.01 SD per year for temperature.
# base_year   <- 2025
# tmean_noise <- 0.30 + 0.01 * (gcm_years - base_year)  # 0.30 → 0.80
# ppt_noise   <- 0.20 + 0.005 * (gcm_years - base_year) # 0.20 → 0.45
# swe_noise   <- 0.15 + 0.005 * (gcm_years - base_year) # 0.15 → 0.40
# 
# # Predict over the 50-year GCM window with time-varying noise
# pred_gcm <- et_predict(
#   model     = fit_A,
#   newdata   = gcm_df,
#   env_noise = list(
#     Tmean = tmean_noise,   # per-row vector, length = nrow(gcm_df)
#     PPT   = ppt_noise,
#     SWE   = swe_noise
#   ),
#   ci_levels = c(0.90, 0.95),
#   n_draws   = 1000L
# )
# 
# # Shelf life over the 50-year horizon
# sl_gcm <- shelf_life(
#   pred_gcm,
#   plausible_range          = range(train_A$z_diff),
#   time_col                 = "year",
#   max_extrapolation_factor = Inf    # allow projection beyond the 50-yr window
# )
# 
# print(sl_gcm)
# # If the 90% CI stays below the plausible range throughout 2025-2075:
# #   Shelf life : > 2075 (lower bound — all periods informative, no trend)
# # If it crosses (say, around 2058):
# #   Shelf life : ~2058 (observed — threshold first exceeded)

## ----gcm-decompose, eval = FALSE----------------------------------------------
# decomp_gcm <- decompose_uncertainty(pred_gcm)
# # env_var should be small in 2025 and growing toward 2075 as GCM spread
# # accumulates; param_var and residual_var remain roughly constant.
# plot(gcm_years, decomp_gcm$env_var,
#      type = "l", xlab = "Year", ylab = "Environmental variance",
#      main = "Growing GCM ensemble uncertainty propagated into forecast")

## ----lm-priors----------------------------------------------------------------
# lm_fit_A <- lm(z_diff ~ Tmean + PPT + SWE, data = train_A)
# 
# prior_lm_A <- extract_priors(
#   model      = lm_fit_A,
#   multiplier = 2.0,
#   min_sd     = 0.10
# )
# 
# print(prior_lm_A)
# # ErrorTracer prior specification
# #   Method      : lm
# #   Predictors  : 3
# #   Multiplier  : 2
# #   Min SD      : 0.1
# #   Coefficients:
# #     Tmean                 mean =   0.6669  sd =   1.3338
# #     PPT                   mean =  -0.4677  sd =   0.9354
# #     SWE                   mean =   0.3859  sd =   0.7719

## ----back-transform-----------------------------------------------------------
# # Standardisation constants used when generating et_sim
# et_sim$standardization
# # $Tmean
# #      mean        sd
# # 15.5422  0.6797
# #
# # $PPT
# #      mean        sd
# # 110.5460  14.6855
# #
# # $SWE
# #      mean        sd
# # 86.9190  15.1885
# 
# # A Tmean of +1 SD corresponds to:
# unstandardize(z = 1.0,
#               mu = et_sim$standardization$Tmean["mean"],
#               s  = et_sim$standardization$Tmean["sd"])
# # [1] 16.22   ← 1 SD above training mean = 16.2 °C

## ----session-info, eval = TRUE, echo = FALSE----------------------------------
sessionInfo()

# R/sensitivity.R -- et_sensitivity_profile(): noise-level sensitivity of
# decomposition and shelf life.

#' Sensitivity profile of the environmental uncertainty component
#'
#' Sweeps a grid of noise magnitudes, re-runs \code{\link{et_predict}} for
#' each, and summarises how the environmental variance component and the
#' forecast shelf life respond.  This turns the subjective choice of
#' \code{env_noise} into an auditable \emph{curve}: instead of reporting a
#' single shelf life at one (often hand-picked) measurement-error level, the
#' user sees how the horizon degrades as assumed noise grows.
#'
#' Three argument styles are supported for the grid:
#' \itemize{
#'   \item \code{fraction_grid}: scalar fractions of each predictor's SD in
#'     \code{newdata}.  These are passed straight to the scalar form of
#'     \code{\link{et_predict}}'s \code{env_noise}.  Use this for an apples-
#'     to-apples sweep that scales all predictors together.
#'   \item \code{absolute_grid}: a \code{list} of named numeric lists /
#'     vectors; each list element becomes a single \code{env_noise} argument
#'     (so you can sweep absolute SDs for one predictor while holding others
#'     fixed).
#'   \item Neither supplied (\code{NULL}): a default log-spaced fraction grid
#'     \code{c(0, 0.01, 0.025, 0.05, 0.1, 0.2, 0.4, 0.8)} is used.
#' }
#'
#' Each iteration calls \code{et_predict(..., env_noise = g)} then
#' \code{\link{shelf_life}} and records:
#' \itemize{
#'   \item Mean parameter / environmental / residual / total variance across
#'     forecast observations.
#'   \item The horizon description (\code{"observed"}, \code{"projected"}, or
#'     \code{"lower_bound"}) and numeric horizon if any.
#'   \item Mean / max CI width / plausible-range ratio.
#' }
#'
#' @param model An \code{et_model} from \code{\link{et_fit}}.
#' @param newdata \code{data.frame} passed to \code{et_predict}.
#' @param response_scale Two-element numeric vector \code{c(min, max)}, as in
#'   \code{\link{shelf_life}}.  The effective range is
#'   \code{diff(response_scale)}.
#' @param plausible_range Deprecated.  Use \code{response_scale} instead.
#' @param fraction_grid Optional numeric vector of scalar noise fractions.
#' @param absolute_grid Optional \code{list} of \code{env_noise} arguments
#'   (each a named list / vector).  If supplied together with
#'   \code{fraction_grid}, \code{fraction_grid} is ignored.
#' @param env_cov,env_dist Passed through to \code{et_predict}.
#' @param include_env_in_ci Logical.  If \code{TRUE} (default), credible
#'   intervals — and hence the shelf-life horizon — are computed from
#'   environmentally inflated draws (\code{et_predict(..., include_env_in_ci
#'   = TRUE)}).  Set \code{FALSE} to see how \code{env_var} evolves while
#'   holding the CI fixed at parameter + residual.
#' @param ci_level Numeric.  Used for shelf life and CI width tracking
#'   (default 0.90).  Must be in \code{ci_levels}.
#' @param ci_levels Numeric vector of CI levels computed per iteration
#'   (default \code{c(0.5, 0.8, 0.9, 0.95)}).
#' @param threshold Numeric.  Shelf-life threshold (default 1.0).
#' @param time_col Character.  Optional time column for shelf life.
#' @param n_draws,n_perturb Passed to \code{et_predict}.
#' @param max_extrapolation_factor Passed to \code{shelf_life}.
#' @param verbose Logical.  If \code{TRUE} (default), logs each iteration.
#'
#' @return An \code{et_sensitivity} object: a \code{data.frame} with one row
#'   per grid point and columns
#'   \describe{
#'     \item{\code{grid_id}}{1-indexed step.}
#'     \item{\code{label}}{Short descriptor of the env_noise used
#'       (e.g.\ \code{"fraction = 0.05"}).}
#'     \item{\code{fraction}}{Scalar fraction when applicable; \code{NA}
#'       for absolute-grid rows.}
#'     \item{\code{param_var, env_var, residual_var, total_var}}{Mean across
#'       forecast observations.}
#'     \item{\code{env_share}}{\code{env_var / (env_var + total_var)} ---
#'       the fraction of combined predictive variance attributable to
#'       predictor noise (bounded in \eqn{[0, 1]}).}
#'     \item{\code{ci_width_mean, ci_width_max}}{At \code{ci_level}.}
#'     \item{\code{ratio_mean, ratio_max}}{CI width / plausible range.}
#'     \item{\code{horizon_type}}{One of \code{"observed"}, \code{"projected"},
#'       \code{"lower_bound"}.}
#'     \item{\code{horizon}}{Numeric horizon (observed or projected) or
#'       \code{NA} for lower-bound rows.}
#'     \item{\code{horizon_description}}{The long description returned by
#'       \code{shelf_life}.}
#'   }
#'   The returned object carries the original call and \code{response_scale}
#'   as attributes for use by \code{\link{et_plot_sensitivity}}.
#'
#' @seealso \code{\link{et_predict}}, \code{\link{shelf_life}},
#'   \code{\link{et_plot_sensitivity}}
#' @export
et_sensitivity_profile <- function(model,
                                   newdata,
                                   response_scale,
                                   fraction_grid = NULL,
                                   absolute_grid = NULL,
                                   env_cov = NULL,
                                   env_dist = NULL,
                                   include_env_in_ci = TRUE,
                                   ci_level = 0.90,
                                   ci_levels = c(0.5, 0.8, 0.9, 0.95),
                                   threshold = 1.0,
                                   time_col = NULL,
                                   n_draws = 2000L,
                                   n_perturb = NULL,
                                   max_extrapolation_factor = 10,
                                   verbose = TRUE,
                                   plausible_range = NULL) {

  # Deprecated alias: plausible_range -> response_scale
  if (!is.null(plausible_range)) {
    if (!missing(response_scale)) {
      stop("Cannot specify both 'response_scale' and 'plausible_range'.")
    }
    warning(
      "'plausible_range' is deprecated; use 'response_scale' instead.",
      call. = FALSE
    )
    response_scale <- plausible_range
  }

  if (!inherits(model, "et_model")) {
    stop("et_sensitivity_profile() expects an et_model object. ",
         "For grouped models, iterate over model$models yourself.")
  }
  if (!ci_level %in% ci_levels) {
    ci_levels <- sort(unique(c(ci_levels, ci_level)))
  }

  # Build grid (list of env_noise args and matching labels/fractions).
  grid <- .build_sensitivity_grid(fraction_grid, absolute_grid)

  rows <- vector("list", length(grid$noise))
  for (i in seq_along(grid$noise)) {
    if (verbose) {
      .et_info("Sensitivity step ", i, "/", length(grid$noise), ": ",
               grid$label[[i]])
    }

    pred <- et_predict(
      model             = model,
      newdata           = newdata,
      env_noise         = grid$noise[[i]],
      env_cov           = env_cov,
      env_dist          = env_dist,
      include_env_in_ci = include_env_in_ci,
      n_draws           = n_draws,
      ci_levels         = ci_levels,
      n_perturb         = n_perturb
    )
    sl <- shelf_life(
      predictions              = pred,
      response_scale           = response_scale,
      ci_level                 = ci_level,
      threshold                = threshold,
      time_col                 = time_col,
      max_extrapolation_factor = max_extrapolation_factor
    )
    hor <- attr(sl, "horizon")

    rows[[i]] <- data.frame(
      grid_id             = i,
      label               = grid$label[[i]],
      fraction            = grid$fraction[[i]],
      param_var           = mean(pred$decomposition$param_var),
      env_var             = mean(pred$decomposition$env_var),
      residual_var        = mean(pred$decomposition$residual_var),
      total_var           = mean(pred$decomposition$total_var),
      env_share           = mean(pred$decomposition$env_var /
                                   pmax(pred$decomposition$env_var +
                                        pred$decomposition$total_var, 1e-12)),
      ci_width_mean       = mean(sl$ci_width),
      ci_width_max        = max(sl$ci_width),
      ratio_mean          = mean(sl$ratio),
      ratio_max           = max(sl$ratio),
      horizon_type        = if (is.null(hor)) NA_character_ else hor$type,
      horizon             = if (is.null(hor) || is.na(hor$value))
                              NA_real_
                            else as.numeric(hor$value),
      horizon_description = if (is.null(hor)) NA_character_
                            else hor$description,
      stringsAsFactors = FALSE
    )
  }

  out <- do.call(rbind, rows)
  out <- structure(out, class = c("et_sensitivity", "data.frame"))
  attr(out, "response_scale") <- response_scale
  attr(out, "ci_level")       <- ci_level
  attr(out, "threshold")      <- threshold
  out
}

# ******************************************************************************
# Internal: build noise-argument grid from user input
# ______________________________________________________________________________

.build_sensitivity_grid <- function(fraction_grid, absolute_grid) {
  if (!is.null(absolute_grid)) {
    if (!is.list(absolute_grid)) {
      stop("absolute_grid must be a list of env_noise arguments.")
    }
    labels <- vapply(seq_along(absolute_grid), function(i) {
      g <- absolute_grid[[i]]
      lab <- names(absolute_grid)[i]
      if (!is.null(lab) && nzchar(lab)) return(lab)
      if (is.list(g) || (is.numeric(g) && !is.null(names(g)))) {
        paste(paste(names(g), signif(unlist(g), 3), sep = "="), collapse = ", ")
      } else if (is.numeric(g) && length(g) == 1L) {
        paste0("fraction = ", signif(g, 3))
      } else {
        paste0("step_", i)
      }
    }, character(1))
    fractions <- vapply(absolute_grid, function(g) {
      if (is.numeric(g) && length(g) == 1L && is.null(names(g))) g
      else NA_real_
    }, numeric(1))
    return(list(noise = absolute_grid, label = labels, fraction = fractions))
  }

  if (is.null(fraction_grid)) {
    fraction_grid <- c(0, 0.01, 0.025, 0.05, 0.1, 0.2, 0.4, 0.8)
  }
  if (!is.numeric(fraction_grid) || any(fraction_grid < 0)) {
    stop("fraction_grid must be a numeric vector of non-negative values.")
  }

  list(
    noise    = as.list(fraction_grid),
    label    = paste0("fraction = ", signif(fraction_grid, 3)),
    fraction = fraction_grid
  )
}

# ******************************************************************************
# S3 methods
# ______________________________________________________________________________

#' @export
print.et_sensitivity <- function(x, ...) {
  cat("ErrorTracer sensitivity profile (et_sensitivity)\n")
  cat("  Grid steps        :", nrow(x), "\n")
  cat("  CI level          :", attr(x, "ci_level"), "\n")
  cat("  Response scale    :", diff(attr(x, "response_scale")), "\n")
  cat("  Threshold         :", attr(x, "threshold"), "\n")
  tab <- as.data.frame(x)[, c("label", "env_var", "env_share", "ratio_mean",
                              "horizon_type", "horizon"), drop = FALSE]
  print(tab, row.names = FALSE)
  invisible(x)
}

# R/predict.R -- et_predict(): posterior prediction with uncertainty propagation

# ******************************************************************************
# S3 generic
# ______________________________________________________________________________

#' Posterior prediction with uncertainty decomposition
#'
#' Generates posterior predictive draws for new observations, propagates
#' environmental measurement uncertainty through the model, and computes
#' credible intervals.  The resulting \code{et_prediction} object is the
#' input to \code{\link{decompose_uncertainty}}, \code{\link{shelf_life}},
#' and the plotting functions.
#'
#' @param model An \code{et_model} or \code{et_model_list} object from
#'   \code{\link{et_fit}}.
#' @param newdata A \code{data.frame} containing the predictor columns
#'   named in the model formula.  For grouped models, must also contain
#'   the grouping column.
#' @param env_noise Environmental measurement / prediction uncertainty.
#'   Can be:
#'   \itemize{
#'     \item \code{NULL} (default): no environmental noise.
#'     \item A single numeric: applied as a fraction of each predictor's
#'       empirical SD in \code{newdata} (e.g.\ \code{0.1} means 10\% noise,
#'       constant across all observations).
#'     \item A named \code{list} or named numeric vector with one scalar per
#'       predictor: constant absolute noise SD per predictor, e.g.\
#'       \code{list(Tmean = 0.5, PPT = 10)}.
#'     \item A named \code{list} where each entry is a \strong{numeric vector
#'       of length \code{nrow(newdata)}}: \emph{time-varying} (per-row) noise
#'       SDs.  Use this when predictor uncertainty grows with forecast horizon,
#'       e.g.\ from a GCM ensemble spread that increases over time:
#'       \code{list(Tmean = 0.30 + 0.01 * (years - base_year))}.
#'       Entries not supplied default to zero (no noise for that predictor).
#'   }
#' @param env_cov Correlation structure of the environmental noise.  The
#'   \emph{magnitudes} of the noise come from \code{env_noise}; \code{env_cov}
#'   supplies the \emph{correlation} between predictors, so that a perturbation
#'   on row \eqn{i} is drawn from
#'   \eqn{\mathcal{N}(0,\; D_i R D_i)} with
#'   \eqn{D_i = \mathrm{diag}(\sigma_{i1}, \ldots, \sigma_{ip})} and
#'   \eqn{R} = the correlation matrix.  One of:
#'   \itemize{
#'     \item \code{NULL} (default): independent noise, \eqn{R = I} --- equivalent
#'       to ErrorTracer behaviour prior to this feature and the right choice
#'       when predictor measurement errors are genuinely independent (e.g.\
#'       separate instruments on unrelated variables).
#'     \item \code{"empirical"}: compute the correlation of the predictor
#'       columns in the \strong{training data} (\code{model$data}).  Use this
#'       when predictor \emph{errors} are expected to inherit the correlation
#'       structure of the predictors themselves --- e.g.\ temperature and
#'       humidity that co-vary in the underlying climate system.
#'     \item \code{"newdata"}: compute the correlation of the predictor columns
#'       in \code{newdata}.  Useful when the forecast window has a different
#'       covariance structure than training (e.g.\ scenario runs).
#'     \item A numeric \eqn{p \times p} matrix with \code{dimnames} matching the
#'       model's predictors.  Entries with an off-diagonal exceeding 1 are
#'       rescaled to a correlation matrix.  Use this to supply an independent
#'       estimate of the \emph{error} correlation structure (e.g.\ from a
#'       reanalysis product or a sensor covariance report).
#'   }
#'   A correlation derived from training data is a working assumption: the
#'   structure of the \emph{errors} is assumed to mirror the structure of the
#'   \emph{values}.  When this is implausible, pass a matrix directly.
#' @param env_dist Distributional form of the per-predictor noise.  The
#'   \code{env_noise} SDs set the \emph{magnitude} of the perturbation;
#'   \code{env_dist} sets its \emph{shape}.  For every distribution other than
#'   \code{"gaussian"}, the noise is calibrated so that (approximately)
#'   \eqn{E[\tilde x] = x} and \eqn{\mathrm{Var}[\tilde x] = \sigma^2}, using a
#'   Gaussian copula to honour \code{env_cov}.  One of:
#'   \itemize{
#'     \item \code{NULL} (default): \code{"gaussian"} for every predictor ---
#'       additive Gaussian noise, legacy behaviour.
#'     \item A single string (\code{"gaussian"}, \code{"lognormal"},
#'       \code{"gamma"}, \code{"beta"}): applied to all predictors.
#'     \item A named list / character vector with one entry per predictor to
#'       override the default, e.g.\
#'       \code{list(PPT = "gamma", tmax = "gaussian")}.
#'   }
#'   Distributions:
#'   \describe{
#'     \item{\code{"gaussian"}}{Additive normal noise (\eqn{\tilde x = x + \varepsilon},
#'       \eqn{\varepsilon \sim N(0, \sigma^2)}).  Appropriate for symmetric
#'       measurement error on a continuous, potentially negative scale
#'       (temperature, anomalies).}
#'     \item{\code{"lognormal"}}{Multiplicative noise: \eqn{\log \tilde x \sim
#'       N(\log x - s^2/2,\; s^2)} with \eqn{s^2 = \log(1 + (\sigma/x)^2)}.
#'       Preserves positivity; right-tail skewed.  Natural for strictly
#'       positive continuous variables whose error scales with magnitude
#'       (e.g.\ enzyme activity, biomass).  Rows with \eqn{x \le 0} are
#'       left unperturbed.}
#'     \item{\code{"gamma"}}{\eqn{\tilde x \sim \mathrm{Gamma}(\mathrm{shape} =
#'       (x/\sigma)^2,\; \mathrm{rate} = x/\sigma^2)}.  Positive support,
#'       right-skewed, analytic mean/variance match.  Natural for precipitation,
#'       rates, and other non-negative continuous variables.  Rows with
#'       \eqn{x \le 0} are left unperturbed.}
#'     \item{\code{"beta"}}{\eqn{\tilde x \sim \mathrm{Beta}(\alpha,\beta)} with
#'       \eqn{\alpha + \beta = x(1-x)/\sigma^2 - 1}.  Support in \eqn{(0,1)};
#'       appropriate for proportions and probabilities (allele frequencies,
#'       presence rates).  Rows with \eqn{x \not\in (0,1)} or
#'       \eqn{\sigma^2 \ge x(1-x)} are left unperturbed.}
#'   }
#'   Correlation (\code{env_cov}) is applied to the latent standard-normal
#'   draws before the marginal quantile transform, so rank correlations are
#'   preserved across distributions.
#' @param n_draws Integer.  Number of posterior draws to use (default 2000;
#'   capped at the number of draws available in the fit).
#' @param ci_levels Numeric vector.  Credible interval levels to compute
#'   (default \code{c(0.5, 0.8, 0.9, 0.95)}).
#' @param n_perturb Integer.  Number of posterior draws used for the
#'   environmental perturbation step (default \code{min(500, n_draws)}).
#'   Reducing this speeds up computation.
#' @param n_env_draws Integer.  Number of independent environmental
#'   perturbations averaged \emph{per posterior draw} when estimating
#'   \code{env_var} (default \code{1}).  Increasing this reduces Monte Carlo
#'   noise on the environmental-variance estimate at the cost of proportional
#'   computation.  The decomposition always reports a Monte Carlo SE
#'   (\code{v_env_mcse}) alongside \code{env_var}; it decreases roughly like
#'   \eqn{1/\sqrt{n\_env\_draws \cdot n\_perturb}}.
#' @param interval_type Character.  Which draws to use when computing credible
#'   intervals:
#'   \itemize{
#'     \item \code{"predictive"} (default): draws from \code{posterior_predict},
#'       which include sigma (residual noise).  Use this when forecasting
#'       \strong{individual observations} --- e.g. a single population's allele
#'       frequency, one site's ozone reading on a specific day.
#'     \item \code{"linpred"}: draws from \code{posterior_linpred}, which
#'       capture only parameter uncertainty (no sigma).  Use this when
#'       forecasting the \strong{mean response} --- e.g. the expected ozone
#'       across many similar days, or mean delta f across replicate populations.
#'       These intervals are always narrower; they will under-cover individual
#'       observations unless sigma is negligible.
#'   }
#'   The decomposition components and \code{posterior_predict} /
#'   \code{posterior_linpred} matrices are always computed regardless of this
#'   setting.
#' @param include_env_in_ci Logical.  When \code{TRUE} and
#'   \code{interval_type = "predictive"}, credible intervals are constructed
#'   from \strong{environmentally inflated} draws
#'   \eqn{\tilde y = \tilde{\mathrm{lp}} + \varepsilon}, with
#'   \eqn{\tilde{\mathrm{lp}}} the perturbed linear predictor and
#'   \eqn{\varepsilon \sim N(0, \sigma^2)} using posterior draws of
#'   \eqn{\sigma}.  This folds the environmental uncertainty component back
#'   into the CI, which is typically what you want for sensitivity analyses
#'   or whenever the reported interval should cover predictor-measurement
#'   error.  When \code{FALSE} (default, backward compatible), CIs are based
#'   on \code{posterior_predict} only --- parameter + residual, without
#'   predictor noise.
#' @param ... Passed to methods.
#'
#' @return An \code{et_prediction} object (list) containing:
#' \describe{
#'   \item{\code{posterior_predict}}{Matrix \code{[n_draws x n_obs]}: full
#'     posterior predictive draws (parameter + residual uncertainty).}
#'   \item{\code{posterior_linpred}}{Matrix \code{[n_draws x n_obs]}: linear
#'     predictor draws on the \strong{link scale} (parameter uncertainty only).}
#'   \item{\code{lp_perturbed}}{Matrix \code{[n_perturb x n_obs]}: linear
#'     predictor on the link scale computed on environmentally perturbed inputs.}
#'   \item{\code{sigma_draws}}{Numeric vector: posterior draws of sigma
#'     (\code{NA} for families without a sigma parameter, e.g.\ Binomial).}
#'   \item{\code{credible_intervals}}{data.frame with columns
#'     \code{row_id, ci_level, lower, median, upper, width}.}
#'   \item{\code{decomposition}}{data.frame with columns
#'     \code{obs_id, total_var, param_var, env_var, v_env_mcse, residual_var}.
#'     All components are on the \strong{response scale}.
#'     \code{v_env_mcse} is the Monte Carlo SE of \code{env_var}.
#'     \code{residual_var} is per-observation for non-Gaussian families
#'     (e.g.\ Binomial: \eqn{E_s[\mu^{(s)}(1-\mu^{(s)})]}) and constant
#'     for Gaussian.}
#'   \item{\code{newdata}}{The input \code{newdata}.}
#'   \item{\code{model}}{Reference to the \code{et_model} used.}
#'   \item{\code{env_cov}}{The \eqn{p \times p} correlation matrix actually
#'     used for perturbation (identity for \code{env_cov = NULL}).}
#'   \item{\code{env_dist}}{Named character vector mapping each predictor to
#'     the distribution actually used for its perturbation.}
#' }
#'
#' @seealso \code{\link{decompose_uncertainty}}, \code{\link{shelf_life}},
#'   \code{\link{et_calibrate}}
#' @export
et_predict <- function(model, newdata, env_noise = NULL, env_cov = NULL,
                        env_dist = NULL,
                        n_draws = 2000L, ci_levels = c(0.5, 0.8, 0.9, 0.95),
                        n_perturb = NULL,
                        n_env_draws = 1L,
                        interval_type = c("predictive", "linpred"),
                        include_env_in_ci = FALSE, ...) {
  UseMethod("et_predict")
}

# ******************************************************************************
# Method: et_model (single model)
# ______________________________________________________________________________

#' @export
et_predict.et_model <- function(model, newdata, env_noise = NULL,
                                  env_cov = NULL,
                                  env_dist = NULL,
                                  n_draws = 2000L,
                                  ci_levels = c(0.5, 0.8, 0.9, 0.95),
                                  n_perturb = NULL,
                                  n_env_draws = 1L,
                                  interval_type = c("predictive", "linpred"),
                                  include_env_in_ci = FALSE,
                                  ...) {

  interval_type <- match.arg(interval_type)
  n_env_draws   <- max(1L, as.integer(n_env_draws))
  fit           <- model$fit
  pred_names    <- .brms_pred_names(fit)
  n_perturb     <- if (is.null(n_perturb)) min(500L, n_draws) else as.integer(n_perturb)

  # For EIV-fit models, posterior_predict() on newdata requires each se_<pred>
  # column present. Copy them over from the training data (recycled to the
  # length of newdata) when the user hasn't supplied them.
  if (!is.null(model$eiv_spec)) {
    for (p in names(model$eiv_spec)) {
      se_col <- paste0("se_", p)
      if (!se_col %in% colnames(newdata)) {
        v <- as.numeric(model$eiv_spec[[p]])
        if (length(v) == 1L) v <- rep(v, nrow(newdata))
        if (length(v) != nrow(newdata)) {
          stop("eiv_spec[['", p, "']] has length ", length(v),
               " but newdata has ", nrow(newdata), " row(s). ",
               "Supply a se_", p, " column in newdata explicitly.")
        }
        newdata[[se_col]] <- v
      }
    }
  }

  # Resolve environmental noise SDs, correlation structure, and per-predictor
  # perturbation distribution.
  noise_sds <- .resolve_env_noise(env_noise, pred_names, newdata)
  cor_mat   <- .resolve_env_cor(env_cov, pred_names,
                                training_data = model$data,
                                newdata       = newdata)
  dist_spec <- .resolve_env_dist(env_dist, pred_names)

  # If the model was fit with errors-in-variables (me() terms), the fit's
  # beta posteriors already absorb predictor measurement noise; perturbing
  # those predictors again would double-count. Warn, so the user can
  # null-out env_noise for EIV-modelled predictors.
  eiv_preds <- names(model$eiv_spec %||% list())
  if (length(eiv_preds) && !is.null(env_noise)) {
    overlap <- intersect(eiv_preds, names(noise_sds))
    overlap <- overlap[vapply(overlap, function(p) any(noise_sds[[p]] != 0),
                              logical(1))]
    if (length(overlap)) {
      .et_warn("env_noise is non-zero for predictor(s) also modelled ",
               "under eiv (", paste(overlap, collapse = ", "),
               "); env_var may double-count predictor uncertainty.")
    }
  }

  # Detect Gaussian identity link once; used to avoid a redundant
  # posterior_linpred(transform=TRUE) call and to stay backward-compatible.
  is_gauss_id <- .is_gaussian_identity(fit$family)

  # --- 1. Posterior predictive (full uncertainty, response scale) ---
  pp <- brms::posterior_predict(fit, newdata = newdata, ndraws = n_draws)

  # --- 2. Linear predictor draws (link scale, parameter uncertainty only) ---
  lp <- brms::posterior_linpred(fit, newdata = newdata, ndraws = n_draws)

  # --- 3. Response-scale parameter draws ---
  # For Gaussian identity g^{-1}(eta) = eta, so mu_draws == lp — reuse it.
  # For all other families, apply the inverse link to get the response scale.
  mu_draws <- if (is_gauss_id) {
    lp
  } else {
    brms::posterior_linpred(fit, newdata = newdata, ndraws = n_draws,
                            transform = TRUE)
  }

  # --- 4. Draws matrix and dispersion parameters ---
  draws_mat   <- .brms_draws_matrix(fit, max_draws = max(n_draws, n_perturb))
  sigma_draws <- if ("sigma" %in% colnames(draws_mat)) {
    draws_mat[seq_len(n_draws), "sigma"]
  } else {
    rep(NA_real_, n_draws)
  }
  disp_draws <- .extract_disp_draws(draws_mat, n_draws)

  # --- 5. Perturbed linear predictor (environmental uncertainty, link scale) ---
  # Short-circuit: if every predictor's noise is identically zero for every
  # observation, lp_perturbed == lp and env_var will be zero -- skip the loop.
  all_zero_noise <- all(vapply(noise_sds,
                               function(v) all(v == 0, na.rm = TRUE),
                               logical(1)))
  lp_perturbed <- if (all_zero_noise) {
    lp[seq_len(n_perturb), , drop = FALSE]
  } else {
    .compute_lp_perturbed(
      draws_mat   = draws_mat[seq_len(n_perturb), , drop = FALSE],
      newdata     = newdata,
      pred_names  = pred_names,
      noise_sds   = noise_sds,
      cor_mat     = cor_mat,
      dist_spec   = dist_spec,
      n_env_draws = n_env_draws
    )
  }

  # --- 6. Response-scale perturbed draws (for decomposition) ---
  mu_perturbed <- if (is_gauss_id) {
    lp_perturbed
  } else {
    .apply_inv_link(lp_perturbed, fit$family)
  }
  mu_draws_sub <- mu_draws[seq_len(n_perturb), , drop = FALSE]

  # --- 7. Credible intervals ---
  # Route to posterior_predict (includes sigma) or posterior_linpred (no sigma)
  # depending on interval_type. When include_env_in_ci = TRUE and interval_type
  # is predictive, rebuild predictive draws as lp_perturbed + sigma * N(0,1)
  # so predictor noise is folded back into the credible interval.
  ci_draws <- if (interval_type == "predictive") {
    if (isTRUE(include_env_in_ci) && !all_zero_noise) {
      .inflate_env_predictive(lp_perturbed, sigma_draws)
    } else {
      pp
    }
  } else {
    lp
  }
  ci_df <- .compute_ci(ci_draws, ci_levels)

  # --- 8. Decomposition (all components on response scale) ---
  has_autocor <- .formula_has_autocor(fit$formula)
  decomp <- .decompose_from_arrays(
    pp           = pp,
    mu_draws     = mu_draws,
    mu_perturbed = mu_perturbed,
    mu_draws_sub = mu_draws_sub,
    family       = fit$family,
    disp_draws   = disp_draws,
    has_autocor  = has_autocor
  )

  structure(
    list(
      posterior_predict  = pp,
      posterior_linpred  = lp,
      lp_perturbed       = lp_perturbed,
      sigma_draws        = sigma_draws,
      credible_intervals = ci_df,
      decomposition      = decomp,
      newdata            = newdata,
      model              = model,
      env_noise          = env_noise,
      env_cov            = cor_mat,
      env_dist           = dist_spec,
      n_draws            = n_draws,
      interval_type      = interval_type,
      include_env_in_ci  = isTRUE(include_env_in_ci)
    ),
    class = "et_prediction"
  )
}

# ******************************************************************************
# Method: et_model_list (grouped models)
# ______________________________________________________________________________

#' @export
et_predict.et_model_list <- function(model, newdata, env_noise = NULL,
                                      env_cov = NULL,
                                      env_dist = NULL,
                                      n_draws = 2000L,
                                      ci_levels = c(0.5, 0.8, 0.9, 0.95),
                                      n_perturb = NULL,
                                      n_env_draws = 1L,
                                      interval_type = c("predictive", "linpred"),
                                      include_env_in_ci = FALSE,
                                      ...) {
  interval_type <- match.arg(interval_type)

  grouping <- model$grouping
  if (!grouping %in% colnames(newdata)) {
    stop("grouping column '", grouping, "' not found in newdata.")
  }

  groups <- names(model$models)
  preds  <- vector("list", length(groups))
  names(preds) <- groups

  for (g in groups) {
    m <- model$models[[g]]
    if (is.null(m)) {
      .et_warn("Skipping group ", g, " (model is NULL)")
      next
    }
    sub_nd <- newdata[newdata[[grouping]] == g, , drop = FALSE]
    if (nrow(sub_nd) == 0) {
      .et_warn("No newdata rows for group ", g, " -- skipping")
      next
    }

    preds[[g]] <- tryCatch(
      et_predict.et_model(
        model             = m,
        newdata           = sub_nd,
        env_noise         = env_noise,
        env_cov           = env_cov,
        env_dist          = env_dist,
        n_draws           = n_draws,
        ci_levels         = ci_levels,
        n_perturb         = n_perturb,
        n_env_draws       = n_env_draws,
        interval_type     = interval_type,
        include_env_in_ci = include_env_in_ci,
        ...
      ),
      error = function(e) {
        .et_error("Prediction failed for group ", g, ": ", e$message)
        NULL
      }
    )
  }

  structure(
    list(
      predictions = preds,
      grouping    = grouping,
      newdata     = newdata
    ),
    class = "et_prediction_list"
  )
}

# ******************************************************************************
# Internal helpers
# ______________________________________________________________________________

# Compute lp for environmentally perturbed predictor values.
# draws_mat   : [n_perturb x n_params] posterior draws
# newdata     : data.frame of predictors
# pred_names  : character vector
# noise_sds   : named list; each element is a numeric vector of length n_obs
#               giving the per-observation noise SD for that predictor.
# cor_mat     : p x p correlation matrix over pred_names.
# dist_spec   : named character vector (predictor -> distribution).
# n_env_draws : number of perturbations to average per posterior draw.
#               When > 1, the within-draw MC noise on the linear predictor is
#               reduced by averaging k independent perturbations before taking
#               the cross-draw variance.
.compute_lp_perturbed <- function(draws_mat, newdata, pred_names, noise_sds,
                                   cor_mat = NULL, dist_spec = NULL,
                                   n_env_draws = 1L) {
  n_perturb <- nrow(draws_mat)
  n_obs     <- nrow(newdata)
  lp_mat    <- matrix(NA_real_, n_perturb, n_obs)

  intercept_col <- "b_Intercept"
  beta_cols     <- paste0("b_", pred_names)
  avail_betas   <- intersect(beta_cols, colnames(draws_mat))
  avail_preds   <- sub("^b_", "", avail_betas)

  if (length(avail_betas) == 0) {
    .et_warn("No beta columns found in draws matrix -- returning zero LP.")
    return(matrix(0, n_perturb, n_obs))
  }

  int_vals <- if (intercept_col %in% colnames(draws_mat)) {
    draws_mat[, intercept_col]
  } else {
    rep(0, n_perturb)
  }

  # Cholesky of the correlation sub-matrix restricted to available predictors.
  L <- NULL
  if (!is.null(cor_mat)) {
    R_sub <- cor_mat[avail_preds, avail_preds, drop = FALSE]
    if (!isTRUE(all.equal(R_sub, diag(length(avail_preds)),
                          check.attributes = FALSE))) {
      L <- tryCatch(chol(R_sub),
                    error = function(e) {
                      .et_warn("Cholesky of env_cov failed (", e$message,
                               "); falling back to independent noise.")
                      NULL
                    })
    }
  }

  # Stack per-predictor per-obs SDs into an n_obs x p matrix.
  sd_mat <- do.call(cbind, lapply(avail_preds, function(p) {
    v <- noise_sds[[p]]
    if (is.null(v)) rep(0, n_obs) else v
  }))
  colnames(sd_mat) <- avail_preds

  # Fill in any missing distribution entries with "gaussian".
  if (is.null(dist_spec)) {
    dist_spec <- stats::setNames(rep("gaussian", length(avail_preds)),
                                 avail_preds)
  } else {
    missing_preds <- setdiff(avail_preds, names(dist_spec))
    if (length(missing_preds)) dist_spec[missing_preds] <- "gaussian"
  }

  x_mat <- as.matrix(newdata[, avail_preds, drop = FALSE])
  p     <- length(avail_preds)

  for (i in seq_len(n_perturb)) {
    betas <- draws_mat[i, avail_betas]

    if (n_env_draws <= 1L) {
      # Single perturbation (original fast path)
      Z <- matrix(stats::rnorm(n_obs * p), n_obs, p)
      if (!is.null(L)) Z <- Z %*% L
      xmat_p <- x_mat
      for (k in seq_along(avail_preds)) {
        xmat_p[, k] <- .perturb_predictor(
          x     = x_mat[, k],
          sigma = sd_mat[, k],
          Z_std = Z[, k],
          dist  = dist_spec[[avail_preds[k]]]
        )
      }
      lp_mat[i, ] <- int_vals[i] + as.numeric(xmat_p %*% betas)
    } else {
      # Average n_env_draws perturbations; add the intercept once at the end.
      lp_acc <- rep(0, n_obs)
      for (j in seq_len(n_env_draws)) {
        Z <- matrix(stats::rnorm(n_obs * p), n_obs, p)
        if (!is.null(L)) Z <- Z %*% L
        xmat_p <- x_mat
        for (k in seq_along(avail_preds)) {
          xmat_p[, k] <- .perturb_predictor(
            x     = x_mat[, k],
            sigma = sd_mat[, k],
            Z_std = Z[, k],
            dist  = dist_spec[[avail_preds[k]]]
          )
        }
        lp_acc <- lp_acc + as.numeric(xmat_p %*% betas)
      }
      lp_mat[i, ] <- int_vals[i] + lp_acc / n_env_draws
    }
  }

  lp_mat
}

# Fold environmental noise into the predictive draws by adding fresh
# residual noise on top of the perturbed linear predictor.
.inflate_env_predictive <- function(lp_perturbed, sigma_draws) {
  n_p  <- nrow(lp_perturbed)
  n_o  <- ncol(lp_perturbed)
  s    <- sigma_draws[seq_len(n_p)]
  s[is.na(s)] <- 0
  eps  <- matrix(stats::rnorm(n_p * n_o, sd = rep(s, n_o)), nrow = n_p)
  lp_perturbed + eps
}

# Build credible interval data.frame from posterior predictive matrix.
# pp: [n_draws x n_obs]
.compute_ci <- function(pp, ci_levels) {
  n_obs <- ncol(pp)
  rows  <- vector("list", length(ci_levels))

  for (k in seq_along(ci_levels)) {
    lv    <- ci_levels[k]
    alpha <- 1 - lv
    lower  <- apply(pp, 2, stats::quantile, probs = alpha / 2)
    upper  <- apply(pp, 2, stats::quantile, probs = 1 - alpha / 2)
    med    <- apply(pp, 2, stats::median)
    rows[[k]] <- data.frame(
      row_id   = seq_len(n_obs),
      ci_level = lv,
      lower    = lower,
      median   = med,
      upper    = upper,
      width    = upper - lower,
      stringsAsFactors = FALSE
    )
  }

  do.call(rbind, rows)
}

# Core uncertainty decomposition from arrays.
# All components are computed on the RESPONSE SCALE (g^{-1}(eta)) so that
# the law of total variance holds: V_total ~ V_param + V_resid (when
# env_noise = 0), and the components are dimensionally consistent. When
# has_autocor = TRUE a fourth temporal_var component absorbs the
# autocorrelation-induced spread (V_total - (V_param + V_env + V_resid)),
# clamped at 0 for Monte Carlo robustness.
#
# pp           : [n_draws  x n_obs] posterior predictive draws (response scale)
# mu_draws     : [n_draws  x n_obs] response-scale posterior means
#                (= posterior_linpred(transform=TRUE); equals lp for Gaussian)
# mu_perturbed : [n_perturb x n_obs] response-scale draws under env perturbation
# mu_draws_sub : [n_perturb x n_obs] first n_perturb rows of mu_draws
# family       : brms family object (used to dispatch residual variance)
# disp_draws   : named list of dispersion parameter draws (sigma, phi, nu, shape)
# has_autocor  : logical; if TRUE, add a temporal_var component capturing the
#                AR/MA/ARMA-induced predictive variance beyond the iid sum.
.decompose_from_arrays <- function(pp, mu_draws, mu_perturbed, mu_draws_sub,
                                    family, disp_draws,
                                    has_autocor = FALSE) {
  n_obs <- ncol(pp)
  n_p   <- nrow(mu_perturbed)

  # Parameter uncertainty: variance of the response-scale posterior mean.
  param_var <- apply(mu_draws, 2, stats::var)

  # Environmental variance: subtraction estimator on the response scale.
  v_perturbed <- apply(mu_perturbed,   2, stats::var, na.rm = TRUE)
  v_param_sub <- apply(mu_draws_sub,   2, stats::var)
  v_env_raw   <- v_perturbed - v_param_sub
  env_var     <- pmax(v_env_raw, 0)

  # Monte Carlo SE of env_var.
  # Uses the chi-squared approximation SE(Var) ~= Var * sqrt(2 / (n - 1)).
  # When v_env_raw == 0 exactly the two matrices are from the same sample
  # (all_zero_noise path); env_var = 0 by construction, so SE = 0.
  f_se       <- sqrt(2 / max(n_p - 1, 1))
  v_env_mcse <- ifelse(
    v_env_raw == 0,
    0,
    sqrt((v_perturbed * f_se)^2 + (v_param_sub * f_se)^2)
  )

  # Residual variance: family-specific expected within-draw variance.
  # For models with an autocor term this is the *innovation* variance
  # (E[sigma^2] for Gaussian AR(1)); the additional autocorrelation-induced
  # spread is captured in temporal_var below.
  residual_var <- .family_residual_var(family, mu_draws, disp_draws)

  # Total variance: from the posterior predictive (response scale, unchanged).
  # brms's posterior_predict() forecasts iteratively under AR/MA/ARMA, so
  # total_var already includes the autocorrelation contribution.
  total_var <- apply(pp, 2, stats::var)

  out <- data.frame(
    obs_id       = seq_len(n_obs),
    total_var    = total_var,
    param_var    = param_var,
    env_var      = env_var,
    v_env_mcse   = v_env_mcse,
    residual_var = residual_var,
    stringsAsFactors = FALSE
  )

  if (isTRUE(has_autocor)) {
    # Temporal (autocorrelation-induced) variance: the gap between the
    # posterior-predictive total and the iid-equivalent (param + residual)
    # sum. For Gaussian AR(p) this corresponds to the autocorrelated
    # accumulation of innovations beyond a single sigma^2; for ARMA /
    # MA / cosy() / sar() / car() it absorbs whatever residual dependence
    # structure brms is modelling. env_var is deliberately excluded
    # because it is an *additive* perturbation-based augmentation
    # measured outside of posterior_predict (see docs); subtracting it
    # here would systematically zero out temporal_var. Clamped at 0 to
    # absorb small Monte Carlo and posterior-mean approximation noise.
    out$temporal_var <- pmax(0, total_var - (param_var + residual_var))
  }

  out
}

# ******************************************************************************
# Internal: family-aware helpers
# ______________________________________________________________________________

# Returns TRUE for Gaussian family with identity link.
# For this case posterior_linpred == posterior_linpred(transform=TRUE) and
# the decomposition is identical to the pre-response-scale implementation.
.is_gaussian_identity <- function(family) {
  if (is.null(family)) return(TRUE)
  fname <- tolower(family$family %||% "")
  lname <- tolower(family$link   %||% "identity")
  fname %in% c("gaussian", "normal") && lname == "identity"
}

# Apply the inverse link function g^{-1} to a matrix of link-scale values.
# Returns a matrix of the same dimensions on the response scale.
# Falls back to the identity (no transform) when the link is unknown.
.apply_inv_link <- function(lp_mat, family) {
  if (is.null(family)) return(lp_mat)
  link_str <- family$link %||% "identity"
  if (link_str == "identity") return(lp_mat)

  linkinv <- tryCatch(
    family$linkinv %||% stats::make.link(link_str)$linkinv,
    error = function(e) NULL
  )
  if (is.null(linkinv)) {
    linkinv <- tryCatch(stats::make.link(link_str)$linkinv,
                        error = function(e) NULL)
  }
  if (is.null(linkinv)) {
    .et_warn("Could not determine inverse link for '", link_str,
             "'; response-scale decomposition may be on the link scale.")
    return(lp_mat)
  }

  matrix(linkinv(as.vector(lp_mat)), nrow = nrow(lp_mat), ncol = ncol(lp_mat))
}

# Compute the expected within-draw variance E_s[Var(y | mu^(s))] for the
# supported brms families. Returns a numeric vector of length n_obs.
#
# For Gaussian the residual variance is constant across observations (sigma^2);
# for all other supported families it is observation-specific because the
# variance function V(mu) depends on the fitted mean.
#
# Supported families and their variance functions V(mu):
#   gaussian / normal : sigma^2                        (scalar, constant)
#   student           : sigma^2 * nu / (nu - 2)       (scalar, nu > 2)
#   binomial / bernoulli : mu * (1 - mu)
#   poisson           : mu
#   negbinomial       : mu + mu^2 / shape
#   beta              : mu * (1 - mu) / (phi + 1)
#   gamma             : mu^2 / shape
#
# Unsupported families return NA with a warning.
.family_residual_var <- function(family, mu_draws, disp_draws) {
  n_obs <- ncol(mu_draws)
  n_s   <- nrow(mu_draws)
  fname <- tolower(family$family %||% "gaussian")

  # Helper: fetch first n_s rows of a named dispersion vector.
  get_d <- function(name) {
    v <- disp_draws[[name]]
    if (is.null(v) || all(is.na(v))) return(NULL)
    v[seq_len(min(n_s, length(v)))]
  }

  if (fname %in% c("gaussian", "normal")) {
    sigma <- get_d("sigma")
    v <- if (!is.null(sigma)) mean(sigma^2, na.rm = TRUE) else NA_real_
    return(rep(v, n_obs))
  }

  if (fname == "student") {
    sigma <- get_d("sigma")
    nu    <- get_d("nu")
    if (is.null(sigma) || is.null(nu)) return(rep(NA_real_, n_obs))
    valid <- !is.na(nu) & nu > 2
    v <- if (any(valid)) {
      mean(sigma[valid]^2 * nu[valid] / (nu[valid] - 2), na.rm = TRUE)
    } else NA_real_
    return(rep(v, n_obs))
  }

  if (fname %in% c("binomial", "bernoulli")) {
    # V(mu) = mu * (1 - mu), averaged over posterior draws.
    return(colMeans(mu_draws * (1 - mu_draws), na.rm = TRUE))
  }

  if (fname == "poisson") {
    # V(mu) = mu (mean == variance for Poisson).
    return(colMeans(mu_draws, na.rm = TRUE))
  }

  if (fname %in% c("negbinomial", "negbinomial2", "neg_binomial_2")) {
    # V(mu, shape) = mu + mu^2 / shape.
    shape <- get_d("shape")
    if (is.null(shape)) {
      .et_warn("negbinomial 'shape' draws not found; falling back to Poisson residual.")
      return(colMeans(mu_draws, na.rm = TRUE))
    }
    # shape is length n_s; broadcast over n_obs columns via sweep.
    return(colMeans(
      mu_draws + sweep(mu_draws^2, 1, shape, "/"),
      na.rm = TRUE
    ))
  }

  if (fname == "beta") {
    # V(mu, phi) = mu * (1 - mu) / (phi + 1) — brms uses precision phi.
    phi <- get_d("phi")
    if (is.null(phi)) {
      .et_warn("beta 'phi' draws not found; returning mu*(1-mu) approximation.")
      return(colMeans(mu_draws * (1 - mu_draws), na.rm = TRUE))
    }
    return(colMeans(
      sweep(mu_draws * (1 - mu_draws), 1, 1 + phi, "/"),
      na.rm = TRUE
    ))
  }

  if (fname %in% c("gamma", "gamma2")) {
    # V(mu, shape) = mu^2 / shape.
    shape <- get_d("shape")
    if (is.null(shape)) return(rep(NA_real_, n_obs))
    return(colMeans(
      sweep(mu_draws^2, 1, shape, "/"),
      na.rm = TRUE
    ))
  }

  .et_warn(
    "Family '", fname, "' is not supported for the response-scale residual ",
    "variance decomposition. residual_var will be NA. ",
    "Supported: gaussian, student, binomial, bernoulli, poisson, ",
    "negbinomial, beta, gamma."
  )
  rep(NA_real_, n_obs)
}

# Extract dispersion parameter draws from the full draws matrix.
# Always returns a named list with elements sigma, phi, shape, nu.
# Missing columns produce NULL entries.
.extract_disp_draws <- function(draws_mat, n_draws) {
  cols    <- colnames(draws_mat)
  n_use   <- min(n_draws, nrow(draws_mat))
  get_col <- function(name) {
    if (name %in% cols) draws_mat[seq_len(n_use), name] else NULL
  }
  list(
    sigma = get_col("sigma"),
    phi   = get_col("phi"),
    shape = get_col("shape"),
    nu    = get_col("nu")
  )
}

# ******************************************************************************
# S3 methods for et_prediction
# ______________________________________________________________________________

#' @export
print.et_prediction <- function(x, ...) {
  cat("ErrorTracer prediction (et_prediction)\n")
  cat("  Observations  :", ncol(x$posterior_predict), "\n")
  cat("  Draws         :", nrow(x$posterior_predict), "\n")
  cat("  CI levels     :", paste(unique(x$credible_intervals$ci_level), collapse = ", "), "\n")
  cat("  Interval type :", if (is.null(x$interval_type)) "predictive" else x$interval_type, "\n")
  decomp <- x$decomposition
  cat("  Mean var decomposition (across observations):\n")
  cat(sprintf("    Parameter    : %.4f\n", mean(decomp$param_var)))
  cat(sprintf("    Environmental: %.4f  (MC SE: %.4f)\n",
              mean(decomp$env_var), mean(decomp$v_env_mcse)))
  cat(sprintf("    Residual     : %.4f\n", mean(decomp$residual_var, na.rm = TRUE)))
  if ("temporal_var" %in% colnames(decomp)) {
    cat(sprintf("    Temporal     : %.4f\n", mean(decomp$temporal_var, na.rm = TRUE)))
  }
  cat(sprintf("    Total        : %.4f\n", mean(decomp$total_var)))
  invisible(x)
}

#' @export
print.et_prediction_list <- function(x, ...) {
  cat("ErrorTracer grouped predictions (et_prediction_list)\n")
  cat("  Grouping :", x$grouping, "\n")
  n_ok <- sum(!vapply(x$predictions, is.null, logical(1)))
  cat("  Groups   :", n_ok, "/", length(x$predictions), "predicted\n")
  invisible(x)
}

# tests/testthat/test-decompose-nongaussian.R
#
# Tests for B1.1: response-scale decomposition for non-Gaussian families.
# These tests exercise .decompose_from_arrays() directly with mock mu_draws
# that simulate what posterior_linpred(transform=TRUE) returns for each family.
# No real Stan fit is needed.

# ---------------------------------------------------------------------------
# Shared helpers
# ---------------------------------------------------------------------------

# Build a mock disp_draws list from named numeric vectors.
.mock_disp <- function(...) {
  args <- list(...)
  base <- list(sigma = NULL, phi = NULL, shape = NULL, nu = NULL)
  base[names(args)] <- args
  base
}

# Call .decompose_from_arrays() with a given family and mu_draws,
# returning the decomposition data.frame.
.decomp_with_family <- function(family, mu_draws, disp_draws,
                                 n_obs = NULL, n_draws = NULL) {
  if (is.null(n_obs))    n_obs    <- ncol(mu_draws)
  if (is.null(n_draws))  n_draws  <- nrow(mu_draws)
  set.seed(7)
  pp <- matrix(rnorm(n_draws * n_obs), nrow = n_draws)
  # mu_perturbed: slightly perturbed version of mu_draws for env_var > 0
  mu_perturbed <- mu_draws + matrix(rnorm(n_draws * n_obs, sd = 0.02),
                                    nrow = n_draws)
  # For non-Gaussian families, clip mu_perturbed to the valid range
  fname <- tolower(family$family)
  if (fname %in% c("binomial", "bernoulli", "beta")) {
    mu_perturbed <- pmin(pmax(mu_perturbed, 1e-6), 1 - 1e-6)
  }
  if (fname %in% c("poisson", "negbinomial", "gamma")) {
    mu_perturbed <- pmax(mu_perturbed, 1e-6)
  }

  ErrorTracer:::.decompose_from_arrays(
    pp           = pp,
    mu_draws     = mu_draws,
    mu_perturbed = mu_perturbed,
    mu_draws_sub = mu_draws,
    family       = family,
    disp_draws   = disp_draws
  )
}

# ---------------------------------------------------------------------------
# Gaussian (identity link) — baseline, should match legacy behaviour
# ---------------------------------------------------------------------------

test_that("Gaussian identity: residual_var equals mean(sigma^2) for all obs", {
  set.seed(1)
  n_draws <- 200; n_obs <- 4
  mu_draws    <- matrix(rnorm(n_draws * n_obs), nrow = n_draws)
  sigma_draws <- abs(rnorm(n_draws, mean = 1, sd = 0.1))
  family <- list(family = "gaussian", link = "identity", linkinv = identity)

  d <- .decomp_with_family(family, mu_draws, .mock_disp(sigma = sigma_draws))

  expected_resid <- mean(sigma_draws^2)
  expect_equal(unique(round(d$residual_var, 8)), round(expected_resid, 8))
})

# ---------------------------------------------------------------------------
# Binomial (logit link) — key non-Gaussian target family
# ---------------------------------------------------------------------------

test_that("Binomial: residual_var = colMeans(mu*(1-mu)) varies per observation", {
  set.seed(2)
  n_draws <- 300; n_obs <- 5
  # mu_draws on response scale (probabilities): simulate logit^{-1}(eta)
  mu_draws <- matrix(plogis(rnorm(n_draws * n_obs, sd = 1.5)), nrow = n_draws)
  family   <- list(family = "binomial", link = "logit", linkinv = plogis)

  d <- .decomp_with_family(family, mu_draws, .mock_disp())

  expected <- colMeans(mu_draws * (1 - mu_draws))
  expect_equal(d$residual_var, expected, tolerance = 1e-10)
  # residual_var should differ across observations (not constant)
  expect_gt(diff(range(d$residual_var)), 0)
})

test_that("Binomial: param_var is on response (probability) scale, not logit scale", {
  set.seed(3)
  n_draws <- 200; n_obs <- 3
  eta      <- matrix(rnorm(n_draws * n_obs, sd = 2), nrow = n_draws)
  mu_draws <- matrix(plogis(as.vector(eta)), nrow = n_draws)
  family   <- list(family = "binomial", link = "logit", linkinv = plogis)

  d <- .decomp_with_family(family, mu_draws, .mock_disp())
  expected_param_var <- apply(mu_draws, 2, var)

  expect_equal(d$param_var, expected_param_var, tolerance = 1e-10)
  # param_var on probability scale must be much smaller than on logit scale
  link_scale_var <- apply(eta, 2, var)
  expect_true(all(d$param_var < link_scale_var))
})

test_that("Binomial: all decomposition components are non-negative", {
  set.seed(4)
  n_draws <- 150; n_obs <- 6
  mu_draws <- matrix(plogis(rnorm(n_draws * n_obs)), nrow = n_draws)
  family   <- list(family = "binomial", link = "logit", linkinv = plogis)
  d <- .decomp_with_family(family, mu_draws, .mock_disp())

  expect_true(all(d$param_var    >= 0))
  expect_true(all(d$env_var      >= 0))
  expect_true(all(d$v_env_mcse   >= 0))
  expect_true(all(d$residual_var >= 0))
  expect_true(all(d$total_var    >= 0))
})

# ---------------------------------------------------------------------------
# Poisson (log link)
# ---------------------------------------------------------------------------

test_that("Poisson: residual_var = colMeans(mu) per observation", {
  set.seed(5)
  n_draws <- 200; n_obs <- 4
  mu_draws <- matrix(exp(rnorm(n_draws * n_obs, mean = 1, sd = 0.5)),
                     nrow = n_draws)
  family   <- list(family = "poisson", link = "log", linkinv = exp)

  d <- .decomp_with_family(family, mu_draws, .mock_disp())

  expected <- colMeans(mu_draws)
  expect_equal(d$residual_var, expected, tolerance = 1e-10)
})

# ---------------------------------------------------------------------------
# Student-t
# ---------------------------------------------------------------------------

test_that("Student-t: residual_var = mean(sigma^2 * nu / (nu - 2))", {
  set.seed(6)
  n_draws     <- 200; n_obs <- 3
  mu_draws    <- matrix(rnorm(n_draws * n_obs), nrow = n_draws)
  sigma_draws <- abs(rnorm(n_draws, mean = 1, sd = 0.1))
  nu_draws    <- abs(rnorm(n_draws, mean = 10, sd = 1)) + 2 + 0.1  # nu > 2
  family      <- list(family = "student", link = "identity", linkinv = identity)

  d <- .decomp_with_family(family, mu_draws,
                            .mock_disp(sigma = sigma_draws, nu = nu_draws))

  expected <- mean(sigma_draws^2 * nu_draws / (nu_draws - 2))
  expect_equal(unique(round(d$residual_var, 6)), round(expected, 6))
})

# ---------------------------------------------------------------------------
# Unknown family — should warn and return NA
# ---------------------------------------------------------------------------

test_that("Unknown family warns and returns NA for residual_var", {
  set.seed(8)
  n_draws <- 50; n_obs <- 3
  mu_draws <- matrix(rnorm(n_draws * n_obs), nrow = n_draws)
  family   <- list(family = "xyzzy_unknown", link = "identity",
                   linkinv = identity)

  expect_message(
    d <- .decomp_with_family(family, mu_draws, .mock_disp()),
    regexp = "not supported"
  )
  expect_true(all(is.na(d$residual_var)))
})

# ---------------------------------------------------------------------------
# .is_gaussian_identity() helper
# ---------------------------------------------------------------------------

test_that(".is_gaussian_identity returns TRUE only for Gaussian identity", {
  expect_true(ErrorTracer:::.is_gaussian_identity(
    list(family = "gaussian", link = "identity")))
  expect_true(ErrorTracer:::.is_gaussian_identity(
    list(family = "normal", link = "identity")))
  expect_false(ErrorTracer:::.is_gaussian_identity(
    list(family = "gaussian", link = "log")))
  expect_false(ErrorTracer:::.is_gaussian_identity(
    list(family = "binomial", link = "logit")))
  expect_true(ErrorTracer:::.is_gaussian_identity(NULL))
})

# ---------------------------------------------------------------------------
# .apply_inv_link() helper
# ---------------------------------------------------------------------------

test_that(".apply_inv_link identity passes matrix through unchanged", {
  family <- list(family = "gaussian", link = "identity", linkinv = identity)
  m <- matrix(rnorm(20), 4, 5)
  expect_identical(ErrorTracer:::.apply_inv_link(m, family), m)
})

test_that(".apply_inv_link logit applies plogis element-wise", {
  family <- list(family = "binomial", link = "logit", linkinv = plogis)
  m <- matrix(c(-2, 0, 2, 1), 2, 2)
  result <- ErrorTracer:::.apply_inv_link(m, family)
  expect_equal(result, matrix(plogis(c(-2, 0, 2, 1)), 2, 2), tolerance = 1e-10)
  expect_true(all(result > 0 & result < 1))
})

test_that(".apply_inv_link log applies exp element-wise", {
  family <- list(family = "poisson", link = "log", linkinv = exp)
  m <- matrix(c(0, 1, -1, 2), 2, 2)
  result <- ErrorTracer:::.apply_inv_link(m, family)
  expect_equal(result, matrix(exp(c(0, 1, -1, 2)), 2, 2), tolerance = 1e-10)
  expect_true(all(result > 0))
})

# tests/testthat/test-shelf_life.R

# Build a minimal mock et_prediction for shelf life tests
.mock_pred_for_sl <- function(n_obs = 6, ci_widths = NULL) {
  set.seed(42)
  n_draws <- 200
  # Widths grow with obs_id (simulating increasing uncertainty over forecast horizon)
  if (is.null(ci_widths)) ci_widths <- seq(0.2, 1.4, length.out = n_obs)

  # Construct a pp matrix whose quantile widths approximate ci_widths
  pp <- matrix(NA_real_, n_draws, n_obs)
  for (j in seq_len(n_obs)) {
    pp[, j] <- stats::rnorm(n_draws, sd = ci_widths[j] / (2 * 1.645))
  }

  lp  <- pp + matrix(rnorm(n_draws * n_obs, sd = 0.01), n_draws, n_obs)
  lp_p <- lp
  sigma_draws <- rep(0.1, n_draws)

  gauss_family <- list(family = "gaussian", link = "identity", linkinv = identity)
  disp_draws   <- list(sigma = sigma_draws, phi = NULL, shape = NULL, nu = NULL)
  decomp <- ErrorTracer:::.decompose_from_arrays(
    pp = pp, mu_draws = lp, mu_perturbed = lp_p, mu_draws_sub = lp,
    family = gauss_family, disp_draws = disp_draws
  )
  ci_df  <- ErrorTracer:::.compute_ci(pp, c(0.90, 0.95))

  nd <- data.frame(year = 2020 + seq_len(n_obs) - 1L)

  structure(
    list(
      posterior_predict  = pp,
      posterior_linpred  = lp,
      lp_perturbed       = lp_p,
      sigma_draws        = sigma_draws,
      credible_intervals = ci_df,
      decomposition      = decomp,
      newdata            = nd,
      model              = NULL
    ),
    class = "et_prediction"
  )
}

test_that("shelf_life returns et_shelf_life with correct columns", {
  pred <- .mock_pred_for_sl()
  sl   <- shelf_life(pred, response_scale = c(-1, 1), ci_level = 0.90)

  expect_s3_class(sl, c("et_shelf_life", "data.frame"))
  expect_true(all(c("obs_id", "time", "ci_width", "plausible_range",
                    "ratio", "informative") %in% colnames(sl)))
})

test_that("shelf_life row count matches n_obs", {
  n <- 8
  pred <- .mock_pred_for_sl(n_obs = n)
  sl   <- shelf_life(pred, response_scale = c(-2, 2))
  expect_equal(nrow(sl), n)
})

test_that("shelf_life ratio = ci_width / response_scale", {
  pred <- .mock_pred_for_sl()
  sl   <- shelf_life(pred, response_scale = c(-1, 1))
  expect_equal(sl$ratio, sl$ci_width / 2, tolerance = 1e-10)
})

test_that("shelf_life informative flag respects threshold", {
  pred <- .mock_pred_for_sl()
  sl   <- shelf_life(pred, response_scale = c(-1, 1), threshold = 0.5)
  # informative == TRUE when ratio < 0.5
  expect_equal(sl$informative, sl$ratio < 0.5)
})

test_that("shelf_life uses time_col from newdata", {
  pred <- .mock_pred_for_sl(n_obs = 4)
  sl   <- shelf_life(pred, response_scale = c(-1, 1), time_col = "year")
  expect_equal(sl$time, 2020:2023)
})

test_that("shelf_life errors if ci_level not in predictions", {
  pred <- .mock_pred_for_sl()
  expect_error(
    shelf_life(pred, response_scale = c(-1, 1), ci_level = 0.99),
    "ci_level"
  )
})

test_that("shelf_life errors if response_scale has length != 2", {
  pred <- .mock_pred_for_sl()
  expect_error(shelf_life(pred, response_scale = c(1, 2, 3)), "length 2")
})

test_that("shelf_life errors if response_scale min == max", {
  pred <- .mock_pred_for_sl()
  expect_error(shelf_life(pred, response_scale = c(1, 1)), "equal")
})

test_that("shelf_life print method runs without error", {
  pred <- .mock_pred_for_sl()
  sl   <- shelf_life(pred, response_scale = c(-1, 1))
  expect_output(print(sl), "ErrorTracer shelf life")
  expect_output(print(sl), "Informative")
})

test_that("shelf_life plausible_range output column is constant (scalar diff)", {
  pred <- .mock_pred_for_sl(n_obs = 5)
  sl   <- shelf_life(pred, response_scale = c(-2, 2))
  expect_true(all(sl$plausible_range == 4))
})

# ── Deprecated plausible_range alias ────────────────────────────────────────

test_that("shelf_life warns when deprecated plausible_range is used", {
  pred <- .mock_pred_for_sl(n_obs = 4)
  expect_warning(
    shelf_life(pred, plausible_range = c(-1, 1)),
    "deprecated"
  )
})

test_that("shelf_life plausible_range alias produces same result as response_scale", {
  pred <- .mock_pred_for_sl(n_obs = 4)
  sl_new  <- shelf_life(pred, response_scale = c(-1, 1))
  sl_old  <- suppressWarnings(shelf_life(pred, plausible_range = c(-1, 1)))
  expect_equal(sl_new$ratio, sl_old$ratio, tolerance = 1e-10)
})

# ── Horizon attribute tests ──────────────────────────────────────────────────

test_that("horizon type is 'observed' when threshold crossed in window", {
  # widths grow 0.2 → 1.4 over 6 obs; threshold = 0.5 => crossed early
  pred <- .mock_pred_for_sl(n_obs = 6)
  sl   <- shelf_life(pred, response_scale = c(-1, 1), threshold = 0.5)
  hor  <- attr(sl, "horizon")
  expect_equal(hor$type, "observed")
  expect_true(is.numeric(hor$value) && !is.na(hor$value))
})

test_that("horizon type is 'lower_bound' when all informative and trend flat", {
  # Constant widths => zero slope => lower bound
  pred <- .mock_pred_for_sl(n_obs = 6, ci_widths = rep(0.2, 6))
  sl   <- shelf_life(pred, response_scale = c(-1, 1), threshold = 1.0)
  hor  <- attr(sl, "horizon")
  expect_equal(hor$type, "lower_bound")
  expect_true(is.na(hor$value))
})

test_that("horizon type is 'projected' when positive slope within cap", {
  # Use many obs and a steep, clear trend so quantile-based widths reliably
  # increase; large max_extrapolation_factor ensures projection is not capped.
  pred <- .mock_pred_for_sl(n_obs = 10,
                            ci_widths = seq(0.20, 0.90, length.out = 10))
  sl   <- shelf_life(pred, response_scale = c(-1, 1), threshold = 1.0,
                     min_slope_for_projection = 1e-10,
                     max_extrapolation_factor = Inf)
  hor  <- attr(sl, "horizon")
  expect_equal(hor$type, "projected")
  # Projected value must be a finite number beyond the last observed time
  expect_true(is.numeric(hor$value) && !is.na(hor$value) &&
              hor$value > hor$last_informative)
})

test_that("max_extrapolation_factor cap converts projected to lower_bound", {
  # Same gentle slope as above but cap = 0 => always lower bound
  pred <- .mock_pred_for_sl(n_obs = 6, ci_widths = seq(0.1, 0.35, length.out = 6))
  sl_cap <- shelf_life(pred, response_scale = c(-1, 1), threshold = 1.0,
                       min_slope_for_projection = 1e-10,
                       max_extrapolation_factor = 0)
  expect_equal(attr(sl_cap, "horizon")$type, "lower_bound")
})

test_that("horizon attribute has required fields", {
  pred <- .mock_pred_for_sl(n_obs = 6)
  sl   <- shelf_life(pred, response_scale = c(-1, 1))
  hor  <- attr(sl, "horizon")
  expect_true(all(c("type", "value", "last_informative", "description") %in% names(hor)))
})

# tests/testthat/test-shelf-life-projection.R
#
# Tests for B1.5: shelf-life projection formula and se_t_star.
# Verifies:
#   1. Projected t* = (tau - a) / b (correct inversion of linear fit).
#   2. se_t_star is present and positive in the projected horizon.
#   3. se_t_star recovers the known shelf life within ±2 SE on synthetic data.
#   4. se_t_star is zero / near-zero when the linear fit is exact.
#   5. Observed and lower-bound modes do not return se_t_star.
# No real Stan fit is needed.

# ---------------------------------------------------------------------------
# Helper: build an et_prediction mock where the CI widths follow a known
# linear trend: width_j = a_true + b_true * time_j.
# ---------------------------------------------------------------------------

.mock_pred_linear <- function(n_obs = 12,
                               a_true = 0.05,   # intercept (ratio at t=0)
                               b_true = 0.06,   # slope (rate of ratio increase)
                               noise_sd = 0,    # width noise; 0 = exact line
                               pr = 1.0,        # response_scale width
                               threshold = 1.0,
                               seed = 42) {
  set.seed(seed)
  # 301 draws so stats::quantile(., 0.05) lands on order statistic 16 and
  # stats::quantile(., 0.95) on order statistic 286 (type 7, no interpolation).
  n_draws <- 301L
  times <- seq_len(n_obs)

  # Target CI widths from the linear model (on ratio scale, times pr)
  target_widths <- (a_true + b_true * times) * pr +
    rnorm(n_obs, sd = noise_sd * pr)
  target_widths <- pmax(target_widths, 0.01 * pr)

  # Build pp deterministically so the empirical 90% CI width equals
  # target_widths[j] exactly — no Monte Carlo noise.
  # Linear spread with x[16] = -w/2 and x[286] = +w/2.
  k <- seq_len(n_draws)
  base <- (k - 16L) / 270  # base[16] = 0, base[286] = 1
  pp <- matrix(NA_real_, n_draws, n_obs)
  for (j in seq_len(n_obs)) {
    pp[, j] <- target_widths[j] * (base - 0.5)
  }

  lp  <- pp + matrix(rnorm(n_draws * n_obs, sd = 1e-4), n_draws, n_obs)
  sigma_draws <- rep(0.1, n_draws)
  gauss_family <- list(family = "gaussian", link = "identity", linkinv = identity)
  disp_draws   <- list(sigma = sigma_draws, phi = NULL, shape = NULL, nu = NULL)
  decomp <- ErrorTracer:::.decompose_from_arrays(
    pp = pp, mu_draws = lp, mu_perturbed = lp, mu_draws_sub = lp,
    family = gauss_family, disp_draws = disp_draws
  )
  ci_df <- ErrorTracer:::.compute_ci(pp, c(0.90))

  structure(
    list(
      posterior_predict  = pp,
      posterior_linpred  = lp,
      lp_perturbed       = lp,
      sigma_draws        = sigma_draws,
      credible_intervals = ci_df,
      decomposition      = decomp,
      newdata            = data.frame(time = times),
      model              = NULL
    ),
    class = "et_prediction"
  )
}

# ---------------------------------------------------------------------------
# 1. Projected t* matches (tau - a) / b for an exact linear CI-width trend
# ---------------------------------------------------------------------------

test_that("projected t* equals (tau - a) / b for exact linear width trend", {
  # With zero noise, the ratio vs time relationship is exactly linear.
  # t* = (1.0 - a_true) / b_true (threshold = 1.0, pr = 1 so ratio = width/pr)
  a_true <- 0.10; b_true <- 0.08; threshold <- 1.0
  pred <- .mock_pred_linear(n_obs = 10, a_true = a_true, b_true = b_true,
                             noise_sd = 0, pr = 1.0, threshold = threshold,
                             seed = 1)

  sl  <- suppressWarnings(  # lm() warns "essentially perfect fit" on exact data
    shelf_life(pred, response_scale = c(0, 1), threshold = threshold,
               min_slope_for_projection = 1e-10,
               max_extrapolation_factor = Inf)
  )
  hor <- attr(sl, "horizon")

  expect_equal(hor$type, "projected")

  # Recover the linear fit parameters from the ratios directly
  lm_fit <- lm(sl$ratio ~ sl$time)
  a_fit  <- unname(coef(lm_fit)[1])
  b_fit  <- unname(coef(lm_fit)[2])
  t_star_expected <- (threshold - a_fit) / b_fit

  expect_equal(hor$value, t_star_expected, tolerance = 1e-6)
})

# ---------------------------------------------------------------------------
# 2. se_t_star is present and positive in projected mode
# ---------------------------------------------------------------------------

test_that("projected horizon includes se_t_star and it is positive", {
  pred <- .mock_pred_linear(n_obs = 10, a_true = 0.10, b_true = 0.07,
                             noise_sd = 0.005, seed = 7)
  sl  <- shelf_life(pred, response_scale = c(0, 1),
                    min_slope_for_projection = 1e-10,
                    max_extrapolation_factor = Inf)
  hor <- attr(sl, "horizon")

  expect_equal(hor$type, "projected")
  expect_true(!is.null(hor$se_t_star))
  expect_true(is.finite(hor$se_t_star))
  expect_true(hor$se_t_star >= 0)
})

# ---------------------------------------------------------------------------
# 3. se_t_star recovers known shelf life within ±2 SE (noisy case)
# ---------------------------------------------------------------------------

test_that("true t* lies within projected t* ± 2 se_t_star", {
  # True t*: (1 - 0.10) / 0.06 = 15.0
  a_true <- 0.10; b_true <- 0.06; threshold <- 1.0
  t_star_true <- (threshold - a_true) / b_true   # = 15

  set.seed(123)
  # Small noise so the test is reliable without many replicates
  pred <- .mock_pred_linear(n_obs = 14, a_true = a_true, b_true = b_true,
                             noise_sd = 0.003, pr = 1.0, threshold = threshold,
                             seed = 123)

  sl  <- shelf_life(pred, response_scale = c(0, 1), threshold = threshold,
                    min_slope_for_projection = 1e-10,
                    max_extrapolation_factor = Inf)
  hor <- attr(sl, "horizon")

  expect_equal(hor$type, "projected")
  expect_true(abs(hor$value - t_star_true) <= 2 * hor$se_t_star)
})

# ---------------------------------------------------------------------------
# 4. se_t_star is near zero for an exact (no-noise) linear trend
# ---------------------------------------------------------------------------

test_that("se_t_star is near zero when CI widths follow an exact linear trend", {
  pred <- .mock_pred_linear(n_obs = 8, a_true = 0.10, b_true = 0.09,
                             noise_sd = 0, seed = 5)
  sl  <- suppressWarnings(  # lm() warns "essentially perfect fit" on exact data
    shelf_life(pred, response_scale = c(0, 1),
               min_slope_for_projection = 1e-10,
               max_extrapolation_factor = Inf)
  )
  hor <- attr(sl, "horizon")

  expect_equal(hor$type, "projected")
  # With zero noise, the lm fit is (nearly) perfect — vcov entries are tiny
  expect_lt(hor$se_t_star, 0.5)
})

# ---------------------------------------------------------------------------
# 5. Observed and lower-bound modes do NOT return se_t_star
# ---------------------------------------------------------------------------

test_that("observed horizon has no se_t_star field", {
  # Widths exceed threshold = 0.2 immediately
  pred <- .mock_pred_linear(n_obs = 6, a_true = 0.5, b_true = 0.1,
                             noise_sd = 0, pr = 1.0, threshold = 0.2)
  sl  <- shelf_life(pred, response_scale = c(0, 1), threshold = 0.2)
  hor <- attr(sl, "horizon")

  expect_equal(hor$type, "observed")
  expect_null(hor$se_t_star)
})

test_that("lower_bound horizon has no se_t_star field", {
  pred <- .mock_pred_linear(n_obs = 6, a_true = 0.1, b_true = 0.0,
                             noise_sd = 0, pr = 1.0)
  sl  <- shelf_life(pred, response_scale = c(0, 1), threshold = 1.0,
                    min_slope_for_projection = 0.01)   # slope < min_slope
  hor <- attr(sl, "horizon")

  expect_equal(hor$type, "lower_bound")
  expect_null(hor$se_t_star)
})

# tests/testthat/test-decompose.R

# Minimal mock et_prediction (Gaussian identity) for tests that don't need
# a real Stan fit.  All decomposition is computed via the internal helpers so
# that the mock stays in sync with the production code path.
.mock_et_prediction <- function(n_obs = 5, n_draws = 100) {
  set.seed(99)
  pp <- matrix(rnorm(n_draws * n_obs, mean = 0.5), nrow = n_draws)
  lp <- matrix(rnorm(n_draws * n_obs, mean = 0.5, sd = 0.3), nrow = n_draws)
  # Perturbed LP — slightly higher variance than lp; for Gaussian identity
  # mu_draws == lp and mu_perturbed == lp_p.
  lp_p <- lp + matrix(rnorm(n_draws * n_obs, sd = 0.1), nrow = n_draws)
  sigma_draws <- abs(rnorm(n_draws, mean = 0.5, sd = 0.1))

  gauss_family <- list(family = "gaussian", link = "identity",
                       linkinv = identity)
  disp_draws   <- list(sigma = sigma_draws, phi = NULL,
                       shape = NULL, nu = NULL)

  decomp <- ErrorTracer:::.decompose_from_arrays(
    pp           = pp,
    mu_draws     = lp,       # Gaussian: mu_draws == lp
    mu_perturbed = lp_p,
    mu_draws_sub = lp,       # n_perturb == n_draws in mock
    family       = gauss_family,
    disp_draws   = disp_draws
  )

  ci_df <- ErrorTracer:::.compute_ci(pp, c(0.5, 0.9, 0.95))

  structure(
    list(
      posterior_predict  = pp,
      posterior_linpred  = lp,
      lp_perturbed       = lp_p,
      sigma_draws        = sigma_draws,
      credible_intervals = ci_df,
      decomposition      = decomp,
      newdata            = data.frame(obs_id = seq_len(n_obs)),
      model              = NULL,
      env_noise          = NULL,
      n_draws            = n_draws
    ),
    class = "et_prediction"
  )
}

test_that("decompose_uncertainty returns correct columns", {
  pred   <- .mock_et_prediction()
  decomp <- decompose_uncertainty(pred)

  expect_s3_class(decomp, "data.frame")
  expect_true(all(c("obs_id", "param_var", "env_var", "v_env_mcse",
                    "residual_var", "total_var") %in% colnames(decomp)))
})

test_that("decompose_uncertainty has non-negative variance components", {
  pred   <- .mock_et_prediction(n_obs = 10)
  decomp <- decompose_uncertainty(pred)

  expect_true(all(decomp$param_var    >= 0))
  expect_true(all(decomp$env_var      >= 0))
  expect_true(all(decomp$v_env_mcse   >= 0))
  expect_true(all(decomp$residual_var >= 0))
  expect_true(all(decomp$total_var    >= 0))
})

test_that("decompose_uncertainty row count matches n_obs", {
  n_obs <- 7
  pred  <- .mock_et_prediction(n_obs = n_obs)
  decomp <- decompose_uncertainty(pred)
  expect_equal(nrow(decomp), n_obs)
})

test_that("decompose_uncertainty.default raises an error", {
  expect_error(decompose_uncertainty(list(a = 1)), "et_prediction")
})

test_that("param_var < total_var on average (residual inflates total)", {
  pred   <- .mock_et_prediction(n_obs = 20, n_draws = 500)
  decomp <- decompose_uncertainty(pred)
  expect_true(mean(decomp$total_var) > mean(decomp$param_var))
})

test_that("residual_var is constant across observations for Gaussian", {
  # For Gaussian identity sigma is constant across obs (scalar sigma^2)
  pred   <- .mock_et_prediction(n_obs = 6)
  decomp <- decompose_uncertainty(pred)
  expect_equal(length(unique(round(decomp$residual_var, 10))), 1L)
})

test_that("v_env_mcse is finite and positive when env noise is present", {
  pred   <- .mock_et_prediction(n_obs = 5, n_draws = 50)
  decomp <- decompose_uncertainty(pred)
  # Some env noise was injected in the mock (lp_p = lp + noise)
  expect_true(all(is.finite(decomp$v_env_mcse)))
  expect_true(all(decomp$v_env_mcse >= 0))
})

test_that("temporal_var column is absent unless has_autocor = TRUE", {
  pred   <- .mock_et_prediction()
  decomp <- decompose_uncertainty(pred)
  expect_false("temporal_var" %in% colnames(decomp))
})

test_that("temporal_var captures the autocor gap when has_autocor = TRUE", {
  # Build a mock where pp has variance much greater than param + env + residual,
  # mimicking AR-induced predictive variance growth.
  set.seed(101)
  n_obs   <- 5
  n_draws <- 200
  # Tight linpred (small param), perturbed lp ~ lp (small env), residual sigma ~ 0.3,
  # but pp inflated by extra AR-style innovation (sd = 0.8) on top of lp.
  lp   <- matrix(rnorm(n_draws * n_obs, mean = 0.5, sd = 0.05), nrow = n_draws)
  lp_p <- lp + matrix(rnorm(n_draws * n_obs, sd = 0.01), nrow = n_draws)
  sigma_draws <- rep(0.3, n_draws)
  pp <- lp + matrix(rnorm(n_draws * n_obs, sd = sqrt(0.3^2 + 0.8^2)), nrow = n_draws)

  gauss_family <- list(family = "gaussian", link = "identity",
                       linkinv = identity)
  disp_draws   <- list(sigma = sigma_draws, phi = NULL,
                       shape = NULL, nu = NULL)

  d_iid <- ErrorTracer:::.decompose_from_arrays(
    pp = pp, mu_draws = lp, mu_perturbed = lp_p, mu_draws_sub = lp,
    family = gauss_family, disp_draws = disp_draws, has_autocor = FALSE)
  d_ar <- ErrorTracer:::.decompose_from_arrays(
    pp = pp, mu_draws = lp, mu_perturbed = lp_p, mu_draws_sub = lp,
    family = gauss_family, disp_draws = disp_draws, has_autocor = TRUE)

  expect_false("temporal_var" %in% colnames(d_iid))
  expect_true("temporal_var" %in% colnames(d_ar))
  expect_true(all(d_ar$temporal_var >= 0))
  # param + residual + temporal should reconstruct total_var (within MC
  # tolerance). env_var is excluded because it is an additive perturbation-
  # based augmentation measured outside of posterior_predict.
  recon <- d_ar$param_var + d_ar$residual_var + d_ar$temporal_var
  expect_true(all(abs(recon - d_ar$total_var) < 0.05 * d_ar$total_var))
  # And the temporal gap should be substantial relative to residual_var alone.
  expect_true(mean(d_ar$temporal_var) > mean(d_ar$residual_var))
})

test_that(".formula_has_autocor detects ar()/ma()/arma() and rejects plain formulas", {
  expect_false(ErrorTracer:::.formula_has_autocor(y ~ x))
  expect_true(ErrorTracer:::.formula_has_autocor(y ~ x + ar(time = t, p = 1)))
  expect_true(ErrorTracer:::.formula_has_autocor(y ~ x + ma(time = t, q = 1)))
})

test_that("decompose_uncertainty.et_prediction_list returns grouped data.frame", {
  p1 <- .mock_et_prediction(n_obs = 4)
  p2 <- .mock_et_prediction(n_obs = 3)

  pred_list <- structure(
    list(
      predictions = list(A = p1, B = p2),
      grouping    = "cluster_id",
      newdata     = data.frame(cluster_id = rep(c("A","B"), c(4,3)))
    ),
    class = "et_prediction_list"
  )

  decomp <- decompose_uncertainty(pred_list)
  expect_true("group" %in% colnames(decomp))
  expect_equal(nrow(decomp), 7L)
  expect_equal(sort(unique(decomp$group)), c("A", "B"))
})

test_that("decompose_uncertainty.et_prediction_list skips NULL groups", {
  p1 <- .mock_et_prediction(n_obs = 4)

  pred_list <- structure(
    list(
      predictions = list(A = p1, B = NULL),
      grouping    = "cluster_id",
      newdata     = data.frame(cluster_id = rep("A", 4))
    ),
    class = "et_prediction_list"
  )

  decomp <- decompose_uncertainty(pred_list)
  expect_equal(nrow(decomp), 4L)
  expect_false("B" %in% decomp$group)
})

# tests/testthat/test-v-env-stability.R
#
# Tests for B1.2: Monte Carlo stability of the environmental variance estimator.
# These tests verify:
#   1. v_env_mcse == 0 when env_noise = 0 (no perturbation, no MC noise).
#   2. v_env_mcse decreases as n_env_draws increases (averaging effect).
#   3. env_var == 0 when mu_perturbed == mu_draws_sub (zero noise path).
#   4. .compute_lp_perturbed() accepts n_env_draws and its output has the
#      same shape regardless of n_env_draws value.
# No real Stan fit is needed.

# ---------------------------------------------------------------------------
# Helper: build mu arrays with a controlled level of perturbation noise
# ---------------------------------------------------------------------------

.mock_mu_arrays <- function(n_draws = 200, n_obs = 5,
                             noise_sd = 0.1, seed = 42) {
  set.seed(seed)
  mu_draws     <- matrix(rnorm(n_draws * n_obs, mean = 0.5, sd = 0.3),
                          nrow = n_draws)
  mu_perturbed <- mu_draws + matrix(rnorm(n_draws * n_obs, sd = noise_sd),
                                     nrow = n_draws)
  list(mu_draws = mu_draws, mu_perturbed = mu_perturbed)
}

.gauss_family  <- list(family = "gaussian", link = "identity", linkinv = identity)
.gauss_disp    <- function(n) list(sigma = abs(rnorm(n, 0.5, 0.1)),
                                   phi = NULL, shape = NULL, nu = NULL)

# ---------------------------------------------------------------------------
# v_env_mcse == 0 when env_noise = 0 (mu_perturbed == mu_draws_sub)
# ---------------------------------------------------------------------------

test_that("env_var and v_env_mcse are zero when mu_perturbed equals mu_draws_sub", {
  set.seed(1)
  n_draws <- 100; n_obs <- 4
  mu_draws  <- matrix(rnorm(n_draws * n_obs), nrow = n_draws)
  pp        <- matrix(rnorm(n_draws * n_obs), nrow = n_draws)
  sigma     <- abs(rnorm(n_draws, 0.5))
  disp      <- list(sigma = sigma, phi = NULL, shape = NULL, nu = NULL)

  # Perturbed == draws_sub (same object) → v_env_raw == 0 exactly → SE = 0
  d <- ErrorTracer:::.decompose_from_arrays(
    pp           = pp,
    mu_draws     = mu_draws,
    mu_perturbed = mu_draws,   # identical → no env noise
    mu_draws_sub = mu_draws,
    family       = .gauss_family,
    disp_draws   = disp
  )

  expect_equal(d$env_var,    rep(0, n_obs), tolerance = 1e-10)
  expect_equal(d$v_env_mcse, rep(0, n_obs), tolerance = 1e-10)
})

# ---------------------------------------------------------------------------
# env_var > 0 and finite when perturbation noise is present
# ---------------------------------------------------------------------------

test_that("env_var and v_env_mcse are positive and finite with perturbation", {
  arrays <- .mock_mu_arrays(n_draws = 300, n_obs = 5, noise_sd = 0.3)
  pp     <- matrix(rnorm(300 * 5), nrow = 300)
  set.seed(10)
  sigma  <- abs(rnorm(300, 0.5))
  disp   <- list(sigma = sigma, phi = NULL, shape = NULL, nu = NULL)

  d <- ErrorTracer:::.decompose_from_arrays(
    pp           = pp,
    mu_draws     = arrays$mu_draws,
    mu_perturbed = arrays$mu_perturbed,
    mu_draws_sub = arrays$mu_draws,
    family       = .gauss_family,
    disp_draws   = disp
  )

  expect_true(all(is.finite(d$env_var)))
  expect_true(all(is.finite(d$v_env_mcse)))
  expect_true(mean(d$env_var) > 0)
  expect_true(mean(d$v_env_mcse) > 0)
})

# ---------------------------------------------------------------------------
# v_env_mcse decreases as n_env_draws increases
# (averaged perturbations reduce within-draw MC noise → lower v_env variance)
# ---------------------------------------------------------------------------

test_that("v_env_mcse decreases as n_env_draws increases", {
  skip_if_not_installed("brms")

  set.seed(99)
  n_obs     <- 3
  n_perturb <- 60
  # Synthetic draws matrix with two parameters
  draws_mat <- cbind(
    b_Intercept = rnorm(n_perturb, 0, 0.5),
    b_x1        = rnorm(n_perturb, 1, 0.3)
  )
  newdata   <- data.frame(x1 = rnorm(n_obs))
  noise_sds <- list(x1 = rep(0.5, n_obs))

  lp1 <- ErrorTracer:::.compute_lp_perturbed(
    draws_mat = draws_mat, newdata = newdata,
    pred_names = "x1", noise_sds = noise_sds,
    n_env_draws = 1L
  )
  lp5 <- ErrorTracer:::.compute_lp_perturbed(
    draws_mat = draws_mat, newdata = newdata,
    pred_names = "x1", noise_sds = noise_sds,
    n_env_draws = 5L
  )
  lp20 <- ErrorTracer:::.compute_lp_perturbed(
    draws_mat = draws_mat, newdata = newdata,
    pred_names = "x1", noise_sds = noise_sds,
    n_env_draws = 20L
  )

  # All have same shape
  expect_equal(dim(lp1),  c(n_perturb, n_obs))
  expect_equal(dim(lp5),  c(n_perturb, n_obs))
  expect_equal(dim(lp20), c(n_perturb, n_obs))

  # Variance across draws should decrease with n_env_draws because within-draw
  # noise is averaged out. Check for the first observation.
  var1  <- apply(lp1,  2, var)
  var5  <- apply(lp5,  2, var)
  var20 <- apply(lp20, 2, var)

  # With n_env_draws=20 the variance should be closer to the "true" parameter
  # variance (less inflated by MC noise). On average across obs, mean(var20)
  # should be <= mean(var1) — not guaranteed for every obs with finite n_perturb,
  # so we test the mean.
  expect_lte(mean(var20), mean(var1) * 1.1)   # allow 10% slack for randomness
})

# ---------------------------------------------------------------------------
# .compute_lp_perturbed output shape invariant to n_env_draws
# ---------------------------------------------------------------------------

test_that(".compute_lp_perturbed returns [n_perturb x n_obs] for any n_env_draws", {
  set.seed(5)
  n_perturb <- 30; n_obs <- 4
  draws_mat <- cbind(
    b_Intercept = rnorm(n_perturb),
    b_x1        = rnorm(n_perturb)
  )
  newdata   <- data.frame(x1 = rnorm(n_obs))
  noise_sds <- list(x1 = rep(0.2, n_obs))

  for (k in c(1L, 3L, 10L)) {
    lp <- ErrorTracer:::.compute_lp_perturbed(
      draws_mat   = draws_mat,
      newdata     = newdata,
      pred_names  = "x1",
      noise_sds   = noise_sds,
      n_env_draws = k
    )
    expect_equal(dim(lp), c(n_perturb, n_obs),
                 label = paste0("n_env_draws = ", k))
  }
})

# ---------------------------------------------------------------------------
# v_env_mcse formula: check it is sqrt(2/(n-1)) * sqrt(Vp^2 + Vs^2)
# ---------------------------------------------------------------------------

test_that("v_env_mcse matches the chi-squared SE formula analytically", {
  set.seed(20)
  n_draws <- 100; n_obs <- 3
  mu_draws     <- matrix(rnorm(n_draws * n_obs, sd = 0.5), nrow = n_draws)
  mu_perturbed <- mu_draws + matrix(rnorm(n_draws * n_obs, sd = 0.15),
                                     nrow = n_draws)
  pp           <- matrix(rnorm(n_draws * n_obs), nrow = n_draws)
  sigma        <- abs(rnorm(n_draws, 0.5))
  disp         <- list(sigma = sigma, phi = NULL, shape = NULL, nu = NULL)

  d <- ErrorTracer:::.decompose_from_arrays(
    pp           = pp,
    mu_draws     = mu_draws,
    mu_perturbed = mu_perturbed,
    mu_draws_sub = mu_draws,
    family       = .gauss_family,
    disp_draws   = disp
  )

  n_p   <- n_draws
  f_se  <- sqrt(2 / (n_p - 1))
  vp    <- apply(mu_perturbed, 2, var)
  vs    <- apply(mu_draws,     2, var)
  expected_mcse <- sqrt((vp * f_se)^2 + (vs * f_se)^2)

  expect_equal(d$v_env_mcse, expected_mcse, tolerance = 1e-10)
})

# R/priors.R — extract_priors(): regularized or standard regression model → 
# brms prior specification

# ******************************************************************************
# S3 generic
# ______________________________________________________________________________

#' Extract brms prior specification from a regularized or standard regression model
#'
#' Converts a fitted model object into a \code{brms} prior specification
#' suitable for \code{\link{et_fit}}.  The coefficient estimates (or
#' importance weights for \pkg{ranger}) from the regularized or standard fit are used as
#' informative prior means, so the Bayesian model starts close to the
#' regularized or standard solution while remaining open to data-driven revision.
#'
#' @param model A fitted model.  Supported classes:
#'   \itemize{
#'     \item \code{cv.glmnet} / \code{glmnet} — elastic net / lasso
#'     \item \code{lm} — ordinary least squares
#'     \item \code{glm} — generalized linear model
#'     \item \code{ranger} — random forest (importance-scaled flat priors)
#'   }
#' @param multiplier Numeric scalar.  Prior SD is set to
#'   \code{multiplier * |coef|} (for signed-coefficient methods) or
#'   \code{multiplier * importance_normalised} (for ranger).  Default 2.0.
#' @param min_sd Numeric scalar.  Minimum prior SD to avoid degenerate
#'   (spike) priors on near-zero coefficients.  Default 0.1.
#' @param intercept_prior_sd Optional prior SD for the intercept term.  The
#'   default \code{NULL} emits \emph{no} explicit intercept prior and lets
#'   \code{brms} pick its data-aware default
#'   (\code{student_t(3, median(y), mad(y))}).  If you supply a numeric value,
#'   the prior becomes \code{normal(0, intercept_prior_sd)} --- this is only
#'   appropriate when the response has been centred (or nearly so) before
#'   fitting, since \code{brms}'s \code{class = "Intercept"} refers to the
#'   intercept at the centre of the predictors, whose posterior mean equals
#'   \code{E[y]} there.  Supplying a small \code{intercept_prior_sd} for an
#'   uncentred response will pull the intercept toward zero and cripple the
#'   fit.
#' @param sigma_prior_scale Scale parameter for the half-Cauchy prior on the
#'   residual SD sigma.  Default 1.0.
#' @param ... Additional arguments passed to methods.
#' @return An \code{et_prior_spec} list containing:
#'   \describe{
#'     \item{\code{prior}}{A \code{brmsprior} object for use in \code{brms::brm()}.}
#'     \item{\code{pred_names}}{Character vector of included predictor names.}
#'     \item{\code{coefs}}{Named numeric vector of regularized coefficients
#'       (NULL for ranger, which uses importance instead).}
#'     \item{\code{method}}{Character: the dispatch method used.}
#'     \item{\code{multiplier}, \code{min_sd}}{Settings echo.}
#'   }
#' @examples
#' fit_lm <- lm(mpg ~ wt + hp + cyl, data = mtcars)
#' ps <- extract_priors(fit_lm, multiplier = 2, min_sd = 0.1)
#' print(ps)
#' @export
extract_priors <- function(model, multiplier = 2.0, min_sd = 0.1,
                           intercept_prior_sd = NULL, sigma_prior_scale = 1.0,
                           ...) {
  UseMethod("extract_priors")
}

# ******************************************************************************
# Method: lm
# ______________________________________________________________________________

#' @rdname extract_priors
#' @export
extract_priors.lm <- function(model, multiplier = 2.0, min_sd = 0.1,
                               intercept_prior_sd = NULL, sigma_prior_scale = 1.0,
                               ...) {
  coef_summary <- summary(model)$coefficients
  # Drop the intercept row — handled separately
  coef_rows <- coef_summary[rownames(coef_summary) != "(Intercept)", , drop = FALSE]

  if (nrow(coef_rows) == 0) stop("No non-intercept predictors found in lm model.")

  pred_names <- rownames(coef_rows)
  coef_ests <- coef_rows[, "Estimate"]
  coef_ses <- coef_rows[, "Std. Error"]

  .build_prior_spec(
    pred_names = pred_names,
    prior_means = coef_ests,
    prior_sds = pmax(multiplier * coef_ses, min_sd),
    coefs = stats::setNames(coef_ests, pred_names),
    method = "lm",
    multiplier = multiplier,
    min_sd = min_sd,
    intercept_prior_sd = intercept_prior_sd,
    sigma_prior_scale = sigma_prior_scale
  )
}

# ******************************************************************************
# Method: glm
# ______________________________________________________________________________

#' @rdname extract_priors
#' @export
extract_priors.glm <- function(model, multiplier = 2.0, min_sd = 0.1,
                                intercept_prior_sd = NULL, sigma_prior_scale = 1.0,
                                ...) {
  coef_summary <- summary(model)$coefficients
  coef_rows <- coef_summary[rownames(coef_summary) != "(Intercept)", , drop = FALSE]

  if (nrow(coef_rows) == 0) stop("No non-intercept predictors found in glm model.")

  pred_names <- rownames(coef_rows)
  coef_ests <- coef_rows[, "Estimate"]
  coef_ses <- coef_rows[, "Std. Error"]

  .build_prior_spec(
    pred_names = pred_names,
    prior_means = coef_ests,
    prior_sds = pmax(multiplier * coef_ses, min_sd),
    coefs = stats::setNames(coef_ests, pred_names),
    method = "glm",
    multiplier = multiplier,
    min_sd = min_sd,
    intercept_prior_sd = intercept_prior_sd,
    sigma_prior_scale = sigma_prior_scale
  )
}

# ******************************************************************************
# Method: cv.glmnet
# ______________________________________________________________________________

#' @rdname extract_priors
#' @param lambda Which lambda to extract coefficients from when the model is a
#'   \code{cv.glmnet} object.  Either \code{"lambda.min"} (default) or
#'   \code{"lambda.1se"}.
#' @export
extract_priors.cv.glmnet <- function(model, multiplier = 2.0, min_sd = 0.1,
                                      intercept_prior_sd = NULL,
                                      sigma_prior_scale = 1.0,
                                      lambda = "lambda.min", ...) {
  if (!requireNamespace("glmnet", quietly = TRUE)) {
    stop("Package 'glmnet' is required for extract_priors.cv.glmnet().")
  }
  lam_val <- model[[lambda]]
  coef_mat <- glmnet::coef.glmnet(model, s = lam_val)
  .extract_glmnet_coefs(
    coef_mat  = coef_mat,
    multiplier = multiplier,
    min_sd = min_sd,
    intercept_prior_sd = intercept_prior_sd,
    sigma_prior_scale = sigma_prior_scale
  )
}

# ******************************************************************************
# Method: glmnet
# ______________________________________________________________________________

#' @rdname extract_priors
#' @param s Index (integer) or value of lambda to extract coefficients at when
#'   the model is a plain \code{glmnet} object.  Defaults to the first lambda
#'   (smallest regularisation).
#' @export
extract_priors.glmnet <- function(model, multiplier = 2.0, min_sd = 0.1,
                                   intercept_prior_sd = NULL,
                                   sigma_prior_scale = 1.0,
                                   s = 1L, ...) {
  if (!requireNamespace("glmnet", quietly = TRUE)) {
    stop("Package 'glmnet' is required for extract_priors.glmnet().")
  }
  lam_val <- if (is.integer(s) || (is.numeric(s) && s == round(s))) {
    model$lambda[s]
  } else {
    s
  }
  coef_mat <- glmnet::coef.glmnet(model, s = lam_val)
  .extract_glmnet_coefs(
    coef_mat = coef_mat,
    multiplier = multiplier,
    min_sd = min_sd,
    intercept_prior_sd = intercept_prior_sd,
    sigma_prior_scale  = sigma_prior_scale
  )
}

# ******************************************************************************
# Method: ranger
# ______________________________________________________________________________

#' @rdname extract_priors
#' @details
#' For \code{ranger} models, signed coefficients are not available.  Priors
#' are centred at zero (direction unknown) and the prior SD for each predictor
#' is set to \code{multiplier * importance_normalised}, where importance is
#' normalised to the \code{[min_sd, 1]} interval.  Only variables with
#' positive permutation importance are included.
#' @export
extract_priors.ranger <- function(model, multiplier = 2.0, min_sd = 0.1,
                                   intercept_prior_sd = NULL,
                                   sigma_prior_scale = 1.0, ...) {
  if (!requireNamespace("ranger", quietly = TRUE)) {
    stop("Package 'ranger' is required for extract_priors.ranger().")
  }
  imp <- model$variable.importance
  if (is.null(imp)) {
    stop("ranger model has no variable importance. ",
         "Refit with importance = 'permutation' or 'impurity'.")
  }

  # Keep only variables with positive importance
  imp <- imp[imp > 0]
  if (length(imp) == 0) stop("No variables with positive importance in ranger model.")

  pred_names <- names(imp)

  # Normalise importance to [0, 1] then scale
  imp_norm  <- imp / max(imp)
  prior_sds <- pmax(multiplier * imp_norm, min_sd)
  prior_means <- rep(0, length(pred_names))

  .build_prior_spec(
    pred_names = pred_names,
    prior_means = prior_means,
    prior_sds = prior_sds,
    coefs = NULL,          # no signed coefficients
    method = "ranger",
    multiplier = multiplier,
    min_sd = min_sd,
    intercept_prior_sd = intercept_prior_sd,
    sigma_prior_scale = sigma_prior_scale
  )
}

# ******************************************************************************
# Internal helpers
# ______________________________________________________________________________

# Shared code for cv.glmnet and glmnet coefficient extraction
.extract_glmnet_coefs <- function(coef_mat, multiplier, min_sd,
                                   intercept_prior_sd, sigma_prior_scale) {
  # coef_mat is a dgCMatrix with rows = (Intercept, pred1, pred2, ...)
  coef_vec <- as.numeric(coef_mat)
  names(coef_vec) <- rownames(coef_mat)

  # Drop intercept and zero coefficients
  coef_vec <- coef_vec[names(coef_vec) != "(Intercept)"]
  coef_nz <- coef_vec[coef_vec != 0]

  if (length(coef_nz) == 0) {
    stop("All glmnet coefficients are zero at the chosen lambda. ",
         "Try a smaller lambda (less regularisation).")
  }

  pred_names <- names(coef_nz)
  prior_means <- as.numeric(coef_nz)
  prior_sds <- pmax(multiplier * abs(coef_nz), min_sd)

  .build_prior_spec(
    pred_names = pred_names,
    prior_means = prior_means,
    prior_sds = prior_sds,
    coefs = stats::setNames(as.numeric(coef_nz), pred_names),
    method = "glmnet",
    multiplier = multiplier,
    min_sd = min_sd,
    intercept_prior_sd = intercept_prior_sd,
    sigma_prior_scale = sigma_prior_scale
  )
}

# Assemble a brms prior list and return an et_prior_spec object
.build_prior_spec <- function(pred_names, prior_means, prior_sds, coefs,
                               method, multiplier, min_sd,
                               intercept_prior_sd, sigma_prior_scale) {
  if (length(pred_names) == 0) stop("No predictors to build priors for.")

  prior_list <- vector("list", length(pred_names))
  for (j in seq_along(pred_names)) {
    prior_list[[j]] <- brms::set_prior(
      paste0("normal(", round(prior_means[j], 6), ", ", round(prior_sds[j], 6), ")"),
      class = "b",
      coef  = pred_names[j]
    )
  }

  # Intercept: only emit an explicit prior when the user supplied a scale.
  # A NULL default lets brms use its data-aware default
  # (student_t(3, median(y), mad(y))), which is robust across response scales.
  # A numeric value preserves the previous behaviour -- normal(0, sd) -- which
  # is only appropriate for a centred response; see extract_priors() docs.
  if (!is.null(intercept_prior_sd)) {
    prior_list <- c(prior_list, list(brms::set_prior(
      paste0("normal(0, ", round(intercept_prior_sd, 4), ")"),
      class = "Intercept"
    )))
  }

  # Sigma: half-Cauchy
  prior_list <- c(prior_list, list(brms::set_prior(
    paste0("cauchy(0, ", round(sigma_prior_scale, 4), ")"),
    class = "sigma"
  )))

  combined_prior <- do.call(c, prior_list)

  structure(
    list(
      prior = combined_prior,
      pred_names = pred_names,
      coefs = coefs,
      method = method,
      multiplier = multiplier,
      min_sd = min_sd,
      intercept_prior_sd = intercept_prior_sd,
      sigma_prior_scale = sigma_prior_scale
    ),
    class = "et_prior_spec"
  )
}

# ******************************************************************************
# S3 methods for et_prior_spec
# ______________________________________________________________________________

#' @export
print.et_prior_spec <- function(x, ...) {
  cat("ErrorTracer prior specification\n")
  cat("  Method      :", x$method, "\n")
  cat("  Predictors  :", length(x$pred_names), "\n")
  cat("  Multiplier  :", x$multiplier, "\n")
  cat("  Min SD      :", x$min_sd, "\n")
  if (!is.null(x$coefs)) {
    cat("  Coefficients:\n")
    for (j in seq_along(x$pred_names)) {
      cat(sprintf("    %-20s  mean = %8.4f  sd = %8.4f\n",
                  x$pred_names[j],
                  x$coefs[j],
                  max(x$multiplier * abs(x$coefs[j]), x$min_sd)))
    }
  } else {
    cat("  Predictors  :", paste(x$pred_names, collapse = ", "), "\n")
  }
  invisible(x)
}

# tests/testthat/test-priors.R

test_that("extract_priors.lm returns et_prior_spec with correct structure", {
  set.seed(1)
  df  <- data.frame(y = rnorm(30), x1 = rnorm(30), x2 = rnorm(30))
  fit <- lm(y ~ x1 + x2, data = df)
  ps  <- extract_priors(fit, multiplier = 2, min_sd = 0.1)

  expect_s3_class(ps, "et_prior_spec")
  expect_equal(ps$method, "lm")
  expect_equal(sort(ps$pred_names), sort(c("x1", "x2")))
  expect_named(ps$coefs, ps$pred_names)
  expect_true(all(names(ps$coefs) %in% c("x1", "x2")))
  expect_equal(ps$multiplier, 2)
  expect_equal(ps$min_sd, 0.1)
  expect_s3_class(ps$prior, "brmsprior")
})

test_that("extract_priors.lm min_sd floor is enforced", {
  # Near-zero coefficient — prior SD should be min_sd, not multiplier * |coef|
  set.seed(42)
  df  <- data.frame(y = rnorm(50), x1 = rnorm(50), x2 = rnorm(50))
  # Force x2 coefficient close to zero by making it weakly related to y
  df$y <- df$x1 * 2 + rnorm(50, 0, 5)
  fit <- lm(y ~ x1 + x2, data = df)
  ps  <- extract_priors(fit, multiplier = 2, min_sd = 0.5)

  # Extract SDs from brms prior object and check floor
  prior_df <- as.data.frame(ps$prior)
  beta_rows <- prior_df[prior_df$class == "b" & prior_df$coef != "", ]

  # All SDs should be >= min_sd
  extract_sd <- function(prior_str) {
    m <- regmatches(prior_str, regexpr("normal\\([^,]+,\\s*([0-9.]+)\\)", prior_str))
    if (length(m) == 0) return(NA)
    as.numeric(sub(".*,\\s*([0-9.]+)\\).*", "\\1", m))
  }
  sds <- sapply(beta_rows$prior, extract_sd)
  expect_true(all(sds >= 0.5, na.rm = TRUE))
})

test_that("extract_priors.glm works for gaussian family", {
  set.seed(2)
  df  <- data.frame(y = rnorm(40), x1 = rnorm(40), x2 = rnorm(40))
  fit <- glm(y ~ x1 + x2, data = df, family = gaussian())
  ps  <- extract_priors(fit)

  expect_s3_class(ps, "et_prior_spec")
  expect_equal(ps$method, "glm")
  expect_true(all(c("x1", "x2") %in% ps$pred_names))
})

test_that("extract_priors.lm print method runs without error", {
  set.seed(3)
  df  <- data.frame(y = rnorm(20), x1 = rnorm(20))
  fit <- lm(y ~ x1, data = df)
  ps  <- extract_priors(fit)
  expect_output(print(ps), "ErrorTracer prior specification")
  expect_output(print(ps), "lm")
})

test_that("extract_priors.lm errors with intercept-only model", {
  df  <- data.frame(y = rnorm(10))
  fit <- lm(y ~ 1, data = df)
  expect_error(extract_priors(fit), "No non-intercept predictors")
})

test_that("extract_priors dispatches correctly based on class", {
  set.seed(10)
  df  <- data.frame(y = rnorm(30), x = rnorm(30))
  lm_fit  <- lm(y ~ x, data = df)
  glm_fit <- glm(y ~ x, data = df, family = gaussian())

  ps_lm  <- extract_priors(lm_fit)
  ps_glm <- extract_priors(glm_fit)

  expect_equal(ps_lm$method, "lm")
  expect_equal(ps_glm$method, "glm")
})

test_that("extract_priors.glmnet works when glmnet is available", {
  skip_if_not_installed("glmnet")
  set.seed(5)
  x   <- matrix(rnorm(100 * 5), 100, 5,
                 dimnames = list(NULL, paste0("x", 1:5)))
  y   <- x[, 1] * 2 + x[, 3] * (-1) + rnorm(100)
  cvf <- glmnet::cv.glmnet(x, y, alpha = 0.5)
  ps  <- extract_priors(cvf)

  expect_s3_class(ps, "et_prior_spec")
  expect_equal(ps$method, "glmnet")
  expect_true(length(ps$pred_names) > 0)
  expect_true(all(ps$pred_names %in% paste0("x", 1:5)))
})

test_that("extract_priors.ranger works when ranger is available", {
  skip_if_not_installed("ranger")
  set.seed(7)
  df  <- data.frame(y = rnorm(80), x1 = rnorm(80), x2 = rnorm(80), x3 = rnorm(80))
  df$y <- df$x1 * 1.5 + df$x2 * (-0.5) + rnorm(80)
  rf  <- ranger::ranger(y ~ x1 + x2 + x3, data = df,
                        importance = "permutation")
  ps  <- extract_priors(rf, multiplier = 2, min_sd = 0.1)

  expect_s3_class(ps, "et_prior_spec")
  expect_equal(ps$method, "ranger")
  expect_null(ps$coefs)   # ranger has no signed coefficients
  expect_true(length(ps$pred_names) > 0)
})

# R/utils.R -- Internal utilities, logging, theming, numeric helpers

# ============================================================
# Package-level imports
# ============================================================

#' @importFrom stats coef lm sd setNames quantile median var rnorm
#' @importFrom utils head
#' @importFrom rlang .data
NULL

# ============================================================
# Internal logging (not exported)
# ============================================================

.et_log <- function(..., level = "INFO") {
  ts <- format(Sys.time(), "%Y-%m-%d %H:%M:%S")
  msg <- paste0("[", ts, "] [ErrorTracer/", level, "] ", paste0(..., collapse = ""))
  message(msg)
}

.et_info  <- function(...) .et_log(..., level = "INFO")
.et_warn  <- function(...) .et_log(..., level = "WARN")
.et_error <- function(...) .et_log(..., level = "ERROR")

# ============================================================
# Numeric helpers
# ============================================================

#' Standardize a numeric vector (mean-centre, unit-variance)
#'
#' @param x Numeric vector.
#' @return Standardized numeric vector. Returns zeros if variance is zero.
#' @export
standardize <- function(x) {
  s <- stats::sd(x, na.rm = TRUE)
  if (is.na(s) || s < 1e-12) return(rep(0, length(x)))
  (x - mean(x, na.rm = TRUE)) / s
}

#' Reverse standardization
#'
#' @param z Standardized values.
#' @param mu Original mean.
#' @param s  Original standard deviation.
#' @return Values on the original scale.
#' @export
unstandardize <- function(z, mu, s) {
  z * s + mu
}

# ============================================================
# ggplot2 theme
# ============================================================

#' Minimal ggplot2 theme for ErrorTracer plots
#'
#' @param base_size Base font size (default 12).
#' @return A \code{ggplot2} theme object.
#' @export
et_theme <- function(base_size = 12) {
  ggplot2::theme_bw(base_size = base_size) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(face = "bold", size = base_size + 2),
      strip.text = ggplot2::element_text(face = "bold"),
      panel.grid = ggplot2::element_blank(),
      legend.position  = "bottom"
    )
}

# ============================================================
# Internal helpers for et_predict / decomposition
# ============================================================

# Normalise env_noise to a named list of per-observation SD vectors.
# Each element of the returned list has length n_obs, so every row of
# newdata can have its own noise SD (enabling time-varying uncertainty).
#
# Accepted forms for env_noise:
#   NULL                  -- zero noise for all predictors
#   scalar numeric        -- fraction of each predictor's SD in newdata
#   named scalar list/vec -- fixed SD per predictor (constant across obs)
#   named vector list     -- per-row SDs; each entry must be length 1 or n_obs
.resolve_env_noise <- function(env_noise, pred_names, newdata) {
  n_obs <- nrow(newdata)
  zeros <- stats::setNames(lapply(pred_names, function(p) rep(0, n_obs)),
                           pred_names)

  if (is.null(env_noise)) return(zeros)

  if (is.numeric(env_noise) && length(env_noise) == 1) {
    # Scalar fraction: scale by each predictor's empirical SD
    result <- lapply(pred_names, function(p) {
      sd_p <- if (p %in% colnames(newdata)) {
        s <- stats::sd(newdata[[p]], na.rm = TRUE)
        if (is.na(s) || s < 1e-12) 1 else s
      } else 1
      rep(env_noise * sd_p, n_obs)
    })
    return(stats::setNames(result, pred_names))
  }

  # Named list or named numeric vector (possibly with per-row vectors)
  env_noise_list <- if (is.list(env_noise)) env_noise else as.list(env_noise)

  unknown <- setdiff(names(env_noise_list), pred_names)
  if (length(unknown) > 0) {
    .et_warn("env_noise contains predictor(s) not in model: ",
             paste(unknown, collapse = ", "), " -- ignored")
  }

  result <- zeros
  shared <- intersect(names(env_noise_list), pred_names)

  for (p in shared) {
    v <- as.numeric(env_noise_list[[p]])
    if (length(v) == 1L) {
      result[[p]] <- rep(v, n_obs)
    } else if (length(v) == n_obs) {
      result[[p]] <- v
    } else {
      stop(
        "env_noise[[\"", p, "\"]] has length ", length(v),
        " but newdata has ", n_obs, " row(s). ",
        "Supply a scalar (constant noise) or a vector of length n_obs ",
        "(time-varying noise)."
      )
    }
  }

  result
}

# Resolve env_cov into a p x p correlation matrix over pred_names.
# Accepted forms:
#   NULL          -- identity (independent noise; preserves legacy behavior)
#   "empirical"   -- correlation of pred_names columns in training_data
#   "newdata"     -- correlation of pred_names columns in newdata
#   numeric matrix -- user-supplied correlation or covariance; if the diagonal
#                     is not ~1, it is rescaled to a correlation matrix.
#                     rownames/colnames must contain all pred_names.
#
# The returned matrix is symmetric PD (eigenvalue-clipped if necessary) and has
# dimnames = pred_names in the order of pred_names.
.resolve_env_cor <- function(env_cov, pred_names, training_data, newdata) {
  p   <- length(pred_names)
  I_p <- diag(p)
  dimnames(I_p) <- list(pred_names, pred_names)

  if (is.null(env_cov) || isFALSE(env_cov)) return(I_p)

  if (is.character(env_cov) && length(env_cov) == 1L) {
    src <- switch(
      env_cov,
      empirical = training_data,
      newdata   = newdata,
      stop("env_cov must be NULL, 'empirical', 'newdata', or a matrix.")
    )
    keep <- intersect(pred_names, colnames(src))
    if (length(keep) < 2L) {
      .et_warn("env_cov = '", env_cov, "' needs >= 2 predictors present in ",
               "the source data; falling back to independent noise.")
      return(I_p)
    }
    X <- as.matrix(src[, keep, drop = FALSE])
    R <- suppressWarnings(stats::cor(X, use = "pairwise.complete.obs"))
    R[is.na(R)] <- 0
    full <- I_p
    full[keep, keep] <- R
    return(.nearest_pd(full))
  }

  if (is.matrix(env_cov) || is.data.frame(env_cov)) {
    M <- as.matrix(env_cov)
    if (!all(dim(M) == c(p, p))) {
      stop("env_cov matrix must be ", p, " x ", p,
           " (one entry per predictor).")
    }
    if (is.null(dimnames(M))) dimnames(M) <- list(pred_names, pred_names)
    if (!all(pred_names %in% rownames(M)) ||
        !all(pred_names %in% colnames(M))) {
      stop("env_cov matrix dimnames must include all predictors: ",
           paste(pred_names, collapse = ", "))
    }
    M <- M[pred_names, pred_names, drop = FALSE]
    if (max(abs(diag(M) - 1)) > 1e-6) {
      d <- sqrt(pmax(diag(M), 1e-12))
      M <- M / tcrossprod(d)
    }
    return(.nearest_pd(M))
  }

  stop("Unrecognized env_cov argument (class: ",
       paste(class(env_cov), collapse = "/"), ").")
}

# Resolve env_dist into a named character vector (pred_names -> distribution).
# Accepted forms:
#   NULL          -- "gaussian" for every predictor (additive Gaussian noise;
#                    preserves ErrorTracer's legacy behaviour)
#   single string -- apply that distribution to every predictor
#   named list / named character vector -- per-predictor distribution; any
#                    predictor not named defaults to "gaussian"
#
# Supported distributions: "gaussian", "lognormal", "gamma", "beta".
.resolve_env_dist <- function(env_dist, pred_names) {
  valid   <- c("gaussian", "lognormal", "gamma", "beta")
  default <- stats::setNames(rep("gaussian", length(pred_names)), pred_names)

  if (is.null(env_dist)) return(default)

  if (is.character(env_dist) && length(env_dist) == 1L && is.null(names(env_dist))) {
    if (!env_dist %in% valid) {
      stop("env_dist = '", env_dist, "' not in ",
           paste(valid, collapse = ", "))
    }
    return(stats::setNames(rep(env_dist, length(pred_names)), pred_names))
  }

  if (!is.null(names(env_dist))) {
    unknown <- setdiff(names(env_dist), pred_names)
    if (length(unknown) > 0) {
      .et_warn("env_dist contains predictor(s) not in model: ",
               paste(unknown, collapse = ", "), " -- ignored")
    }
    result <- default
    for (p in intersect(names(env_dist), pred_names)) {
      d <- as.character(env_dist[[p]])
      if (!d %in% valid) {
        stop("env_dist[['", p, "']] = '", d, "' not in ",
             paste(valid, collapse = ", "))
      }
      result[[p]] <- d
    }
    return(result)
  }

  stop("env_dist must be NULL, a single distribution name, ",
       "or a named list / named character vector.")
}

# Draw a distribution-aware perturbation of predictor x using latent
# standard-normal draws Z_std (already correlation-scaled). The returned
# vector is the perturbed predictor on its original scale, calibrated so
# that (approximately) E[perturbed] = x and Var[perturbed] = sigma^2, using
# the following parameterisations:
#
#   gaussian  : perturbed = x + Z_std * sigma                      (additive)
#   lognormal : log(perturbed) ~ N(log(x) - s^2/2, s^2),
#               s^2 = log(1 + (sigma/x)^2)                     (multiplicative)
#   gamma     : perturbed ~ Gamma(shape = (x/sigma)^2,
#                                 rate  =  x /sigma^2)
#   beta      : perturbed ~ Beta(alpha, beta) with
#               alpha + beta = x*(1-x)/sigma^2 - 1
#
# Correlation across predictors is induced by the latent Z_std (Gaussian
# copula); the marginal distribution is supplied here. Rows that cannot be
# parameterised for the requested distribution (e.g. gamma/lognormal with
# x <= 0; beta with sigma^2 >= x*(1-x)) are returned unperturbed and a
# warning is emitted at most once per call.
.perturb_predictor <- function(x, sigma, Z_std, dist) {
  if (length(sigma) == 1L) sigma <- rep(sigma, length(x))
  if (all(is.na(sigma) | sigma == 0)) return(x)

  if (dist == "gaussian") {
    return(x + Z_std * sigma)
  }

  u <- pmin(pmax(stats::pnorm(Z_std), 1e-10), 1 - 1e-10)
  out <- x

  if (dist == "lognormal") {
    safe <- !is.na(sigma) & sigma > 0 & !is.na(x) & x > 0
    if (any(safe)) {
      s2 <- log1p((sigma[safe] / x[safe])^2)
      mu <- log(x[safe]) - s2 / 2
      out[safe] <- stats::qlnorm(u[safe], meanlog = mu, sdlog = sqrt(s2))
    }
    if (any(!safe & sigma > 0 & !is.na(sigma))) {
      .et_warn("lognormal perturbation skipped for ", sum(!safe),
               " row(s) with non-positive x.")
    }
    return(out)
  }

  if (dist == "gamma") {
    safe <- !is.na(sigma) & sigma > 0 & !is.na(x) & x > 0
    if (any(safe)) {
      shape <- (x[safe] / sigma[safe])^2
      rate  <-  x[safe] / sigma[safe]^2
      out[safe] <- stats::qgamma(u[safe], shape = shape, rate = rate)
    }
    if (any(!safe & sigma > 0 & !is.na(sigma))) {
      .et_warn("gamma perturbation skipped for ", sum(!safe),
               " row(s) with non-positive x.")
    }
    return(out)
  }

  if (dist == "beta") {
    safe <- !is.na(sigma) & sigma > 0 & !is.na(x) & x > 0 & x < 1 &
            sigma^2 < x * (1 - x) - 1e-12
    if (any(safe)) {
      mu <- x[safe]
      nu <- mu * (1 - mu) / sigma[safe]^2 - 1
      out[safe] <- stats::qbeta(u[safe], shape1 = mu * nu,
                                         shape2 = (1 - mu) * nu)
    }
    n_skip <- sum(!safe & sigma > 0 & !is.na(sigma))
    if (n_skip > 0) {
      .et_warn("beta perturbation skipped for ", n_skip,
               " row(s) outside (0,1) or with sigma^2 >= x*(1-x).")
    }
    return(out)
  }

  stop("Unknown distribution: ", dist)
}

# Project a (possibly non-PD) symmetric matrix onto the PD cone by clipping
# eigenvalues. Used to defend against numerical noise and pairwise-complete
# correlations that are not strictly PSD.
.nearest_pd <- function(M, eps = 1e-8) {
  M  <- (M + t(M)) / 2
  ev <- eigen(M, symmetric = TRUE)
  lam <- pmax(ev$values, eps)
  out <- ev$vectors %*% diag(lam, length(lam)) %*% t(ev$vectors)
  dimnames(out) <- dimnames(M)
  out
}

# Extract the names of fixed-effect predictors (excludes Intercept) from a
# brmsfit object.
.brms_pred_names <- function(fit) {
  fe <- rownames(brms::fixef(fit))
  setdiff(fe, "Intercept")
}

# Extract posterior draws matrix from a brmsfit (rows = draws, cols = params).
# Optionally thin to at most max_draws rows.
.brms_draws_matrix <- function(fit, max_draws = NULL) {
  mat <- as.matrix(fit)
  if (!is.null(max_draws) && nrow(mat) > max_draws) {
    mat <- mat[seq_len(max_draws), , drop = FALSE]
  }
  mat
}

# Detect whether a brms formula (or fitted brmsfit's formula) carries an
# autocorrelation term -- ar(), ma(), arma(), cosy(), unstr(), sar(), car().
# Accepts a formula, brmsformula, mvbrmsformula, or brmsfit. Returns TRUE
# if any response component carries an autocor structure.
#
# Primary path: ask brms::brmsterms() and look at $dpars$mu$ac (covers both
# inline ar()/ma() syntax and the legacy autocor= slot, since brmsterms()
# normalises both onto $ac). For mvbrmsformula objects, scan each response.
# Fallback: regex-scan the deparsed formula for an inline autocor call.
.formula_has_autocor <- function(formula) {
  if (is.null(formula)) return(FALSE)
  if (inherits(formula, "brmsfit")) formula <- formula$formula

  has_ac <- function(bt) {
    if (is.null(bt)) return(FALSE)
    if (!is.null(bt$dpars$mu$ac)) return(TRUE)
    if (!is.null(bt$adforms$ac)) return(TRUE)
    FALSE
  }

  ok <- tryCatch({
    bt <- brms::brmsterms(formula)
    if (inherits(bt, "mvbrmsterms")) {
      any(vapply(bt$terms, has_ac, logical(1)))
    } else {
      has_ac(bt)
    }
  }, error = function(e) NA)
  if (isTRUE(ok) || isFALSE(ok)) return(ok)

  txt <- paste(deparse(formula), collapse = " ")
  grepl("\\b(ar|ma|arma|cosy|unstr|sar|car)\\s*\\(", txt)
}

# tests/testthat/test-utils.R
# Tests for exported and internal utilities in utils.R

# ── standardize / unstandardize ─────────────────────────────────────────────

test_that("standardize produces zero-mean unit-variance output", {
  x   <- c(1, 2, 3, 4, 5)
  z   <- standardize(x)
  expect_equal(mean(z), 0, tolerance = 1e-10)
  expect_equal(sd(z),   1, tolerance = 1e-10)
})

test_that("standardize returns zeros for constant input", {
  x <- rep(3, 5)
  expect_equal(standardize(x), rep(0, 5))
})

test_that("unstandardize inverts standardize", {
  x    <- rnorm(20, mean = 10, sd = 3)
  mu   <- mean(x)
  s    <- sd(x)
  z    <- (x - mu) / s
  back <- unstandardize(z, mu, s)
  expect_equal(back, x, tolerance = 1e-10)
})

# ── et_theme ────────────────────────────────────────────────────────────────

test_that("et_theme returns a ggplot2 theme object", {
  th <- et_theme()
  expect_s3_class(th, "theme")
})

test_that("et_theme respects base_size", {
  th <- et_theme(base_size = 14)
  expect_s3_class(th, "theme")
})

# ── .resolve_env_noise edge cases ───────────────────────────────────────────

test_that("resolve_env_noise scalar: missing predictor uses sd = 1 fallback", {
  nd  <- data.frame(Tmean = rnorm(4), PPT = rnorm(4))
  # SWE is in pred_names but NOT in newdata — fallback sd = 1, noise = 0.1 * 1
  res <- ErrorTracer:::.resolve_env_noise(0.1, c("Tmean", "PPT", "SWE"), nd)
  expect_equal(res[["SWE"]], rep(0.1, 4), tolerance = 1e-10)
})

test_that("resolve_env_noise named list: omitted predictor stays zero", {
  nd  <- data.frame(Tmean = rnorm(4), PPT = rnorm(4), SWE = rnorm(4))
  # Only Tmean and PPT supplied — SWE defaults to zero
  res <- ErrorTracer:::.resolve_env_noise(
    list(Tmean = 0.5, PPT = 0.2), c("Tmean", "PPT", "SWE"), nd
  )
  expect_true(all(res[["SWE"]] == 0))
})

test_that("resolve_env_noise scalar SD near-zero becomes 1", {
  # If sd of predictor is near zero, fall back to sd = 1
  nd  <- data.frame(X = rep(5, 6))  # constant predictor
  res <- ErrorTracer:::.resolve_env_noise(0.2, "X", nd)
  # sd(rep(5,6)) == 0 => fallback sd = 1, so noise = 0.2 * 1 = 0.2
  expect_equal(res[["X"]], rep(0.2, 6), tolerance = 1e-10)
})

# ── .compute_lp_perturbed edge case: no beta columns ────────────────────────

test_that("compute_lp_perturbed returns zero matrix when no beta columns found", {
  n_p  <- 50L
  n_obs <- 4L
  draws <- matrix(0.5, nrow = n_p, ncol = 1L,
                  dimnames = list(NULL, "b_Intercept"))
  nd    <- data.frame(X = rnorm(n_obs))
  noise <- list(X = rep(0.1, n_obs))
  # "X" maps to "b_X" which is not in draws — triggers the no-beta-columns branch
  result <- ErrorTracer:::.compute_lp_perturbed(
    draws_mat  = draws,
    newdata    = nd,
    pred_names = "X",
    noise_sds  = noise
  )
  expect_equal(dim(result), c(n_p, n_obs))
  expect_true(all(result == 0))
})

# ── print methods for et_prediction ─────────────────────────────────────────

test_that("print.et_prediction runs without error", {
  set.seed(1)
  n_d <- 50L; n_o <- 3L
  pp  <- matrix(rnorm(n_d * n_o), n_d, n_o)
  lp  <- pp
  lp_p <- pp
  sigma <- abs(rnorm(n_d, 0.5, 0.1))
  gauss_family <- list(family = "gaussian", link = "identity",
                       linkinv = identity)
  disp_draws   <- list(sigma = sigma, phi = NULL, shape = NULL, nu = NULL)
  decomp <- ErrorTracer:::.decompose_from_arrays(
    pp = pp, mu_draws = lp, mu_perturbed = lp_p, mu_draws_sub = lp,
    family = gauss_family, disp_draws = disp_draws
  )
  ci_df  <- ErrorTracer:::.compute_ci(pp, c(0.9))
  pred <- structure(
    list(posterior_predict = pp, posterior_linpred = lp,
         lp_perturbed = lp_p, sigma_draws = sigma,
         credible_intervals = ci_df, decomposition = decomp,
         newdata = data.frame(x = seq_len(n_o)),
         model = NULL, env_noise = NULL, n_draws = n_d),
    class = "et_prediction"
  )
  expect_output(print(pred), "ErrorTracer prediction")
})

test_that("print.et_prediction_list runs without error", {
  pred_list <- structure(
    list(predictions = list(A = NULL, B = NULL),
         grouping = "grp", newdata = data.frame()),
    class = "et_prediction_list"
  )
  expect_output(print(pred_list), "grouped predictions")
})

# tests/testthat/test-env-noise.R
# Unit tests for .resolve_env_noise() — covers NULL, scalar, named scalar,
# and the new per-row (time-varying) vector form.

n_obs      <- 6L
pred_names <- c("Tmean", "PPT", "SWE")
newdata    <- data.frame(
  Tmean = rnorm(n_obs),
  PPT   = rnorm(n_obs),
  SWE   = rnorm(n_obs)
)

# ── NULL ────────────────────────────────────────────────────────────────────

test_that("resolve_env_noise(NULL) returns all-zero list", {
  res <- ErrorTracer:::.resolve_env_noise(NULL, pred_names, newdata)
  expect_type(res, "list")
  expect_named(res, pred_names)
  for (p in pred_names) expect_true(all(res[[p]] == 0))
  for (p in pred_names) expect_length(res[[p]], n_obs)
})

# ── Scalar fraction ──────────────────────────────────────────────────────────

test_that("resolve_env_noise(scalar) scales by predictor SD", {
  res <- ErrorTracer:::.resolve_env_noise(0.1, pred_names, newdata)
  expect_type(res, "list")
  for (p in pred_names) {
    expect_length(res[[p]], n_obs)
    expected_sd <- 0.1 * sd(newdata[[p]])
    expect_equal(res[[p]], rep(expected_sd, n_obs), tolerance = 1e-10)
  }
})

# ── Named constant list ──────────────────────────────────────────────────────

test_that("resolve_env_noise(named list of scalars) replicates to n_obs", {
  en  <- list(Tmean = 0.5, PPT = 0.2)
  res <- ErrorTracer:::.resolve_env_noise(en, pred_names, newdata)
  expect_length(res[["Tmean"]], n_obs)
  expect_true(all(res[["Tmean"]] == 0.5))
  expect_true(all(res[["PPT"]]   == 0.2))
  expect_true(all(res[["SWE"]]   == 0.0))  # not supplied => zero
})

test_that("resolve_env_noise silently ignores unknown predictor names", {
  # .et_warn() uses an internal logger, not warning() — unknown names are
  # logged but the call still completes, returning zero noise for that pred.
  res <- ErrorTracer:::.resolve_env_noise(
    list(Tmean = 0.5, WIND = 0.1), pred_names, newdata
  )
  # WIND is not a model predictor; result should still have all pred_names
  expect_named(res, pred_names)
  expect_true(all(res[["Tmean"]] == 0.5))
})

# ── Per-row vector (time-varying noise) ─────────────────────────────────────

test_that("resolve_env_noise accepts per-row vector and preserves it", {
  growing <- seq(0.30, 0.80, length.out = n_obs)
  en  <- list(Tmean = growing, PPT = 0.20)
  res <- ErrorTracer:::.resolve_env_noise(en, pred_names, newdata)
  expect_equal(res[["Tmean"]], growing, tolerance = 1e-10)
  expect_true(all(res[["PPT"]] == 0.20))
})

test_that("resolve_env_noise errors when vector length != 1 and != n_obs", {
  en <- list(Tmean = c(0.3, 0.4))   # length 2, but n_obs = 6
  expect_error(
    ErrorTracer:::.resolve_env_noise(en, pred_names, newdata),
    "length"
  )
})

# ── Integration: per-row noise produces larger env_var for noisier obs ───────

test_that("time-varying noise inflates env_var for high-noise observations", {
  # Build a tiny mock draws matrix
  set.seed(7L)
  n_perturb  <- 300L
  draws_mat  <- matrix(
    c(rep(0.5, n_perturb),     # b_Intercept
      rep(1.0, n_perturb),     # b_Tmean
      rep(0.0, n_perturb),     # b_PPT
      rep(0.0, n_perturb)),    # b_SWE
    nrow = n_perturb,
    dimnames = list(NULL, c("b_Intercept", "b_Tmean", "b_PPT", "b_SWE"))
  )

  # Noise grows: first half of obs have low noise, second half high noise
  noise_sds <- list(
    Tmean = c(rep(0.01, n_obs %/% 2), rep(1.00, n_obs - n_obs %/% 2)),
    PPT   = rep(0, n_obs),
    SWE   = rep(0, n_obs)
  )

  lp_p <- ErrorTracer:::.compute_lp_perturbed(
    draws_mat  = draws_mat,
    newdata    = newdata,
    pred_names = pred_names,
    noise_sds  = noise_sds
  )

  # Variance across perturbation draws should be higher for high-noise obs
  lp_var <- apply(lp_p, 2, var)
  low_noise_mean  <- mean(lp_var[seq_len(n_obs %/% 2)])
  high_noise_mean <- mean(lp_var[(n_obs %/% 2 + 1):n_obs])
  expect_gt(high_noise_mean, low_noise_mean * 10)
})

# R/shelf_life.R -- shelf_life(): forecast horizon / shelf life analysis

#' Compute the forecast shelf life
#'
#' Quantifies \emph{when} a forecast becomes uninformative by comparing the
#' width of credible intervals to a response scale.  A forecast is
#' uninformative when its CI width exceeds \code{threshold * response_scale}.
#'
#' The function operates in three modes depending on the available data and
#' whether the uninformative threshold is crossed within the forecast window:
#'
#' \describe{
#'   \item{\strong{Observed}}{The threshold is crossed within the
#'     forecast/validation window.  The shelf life is the first time point
#'     at which \code{ratio >= threshold}.}
#'   \item{\strong{Projected}}{All forecast periods remain informative but
#'     the CI/range ratio is trending upward.  A linear trend is fitted to
#'     the ratios and extrapolated to estimate when the threshold would be
#'     reached.  The projected crossing time \eqn{t^* = (\tau - a) / b}
#'     (where \eqn{\tau} is the threshold, \eqn{a} the fitted intercept,
#'     \eqn{b} the fitted slope) is reported together with a Monte Carlo
#'     standard error \code{se_t_star} derived via the delta method.}
#'   \item{\strong{Lower bound}}{All forecast periods are informative with
#'     no upward trend in the ratio.  The shelf life is reported as a lower
#'     bound: \code{> last observed time}.}
#' }
#'
#' The intended workflow is:
#' \enumerate{
#'   \item Fit the model (\code{\link{et_fit}}).
#'   \item Predict on held-out data or a future time window
#'     (\code{\link{et_predict}}).
#'   \item Call \code{shelf_life()} on those predictions.
#'   \item If the threshold is not crossed in the held-out window, the
#'     \emph{projected} mode extrapolates the horizon automatically.
#' }
#'
#' @param predictions An \code{et_prediction} or \code{et_prediction_list}.
#' @param response_scale Numeric vector of length 2 (\code{c(min, max)})
#'   giving the response scale used as the denominator in the
#'   CI-width / range ratio.  For unbounded responses use
#'   \code{range(training_data$response)} as a conservative default, or
#'   supply a biologically motivated interval.
#'   The effective range is \code{diff(response_scale)}.
#' @param ci_level Numeric.  The credible interval level to use (default
#'   0.90).  Must be present in the \code{et_prediction} object.
#' @param threshold Numeric.  CI width / response scale above which the
#'   forecast is uninformative (default 1.0).
#' @param time_col Character.  Optional column in
#'   \code{predictions$newdata} to use as the time axis.  If \code{NULL},
#'   observation index is used.
#' @param min_slope_for_projection Numeric.  Minimum linear slope (of
#'   ratio vs.\ time) required to attempt extrapolation when all periods
#'   are informative.  Below this value the shelf life is reported as a
#'   lower bound only.  Default \code{1e-4}.
#' @param max_extrapolation_factor Numeric.  Cap on how far the linear
#'   projection may reach beyond the observed window.  If the projected
#'   crossing time exceeds
#'   \code{max(time) + max_extrapolation_factor * (max(time) - min(time))},
#'   the result is reported as a lower bound instead of a projection.
#'   Set to \code{Inf} to disable the cap.  Default \code{10}.
#' @param ... Unused.
#' @param plausible_range Deprecated.  Use \code{response_scale} instead.
#'
#' @return An \code{et_shelf_life} object (a \code{data.frame}) with columns:
#'   \describe{
#'     \item{obs_id}{Observation index.}
#'     \item{time}{Time axis value.}
#'     \item{ci_width}{Width of the credible interval at \code{ci_level}.}
#'     \item{plausible_range}{Effective response scale (scalar diff).}
#'     \item{ratio}{CI width / response scale.}
#'     \item{informative}{Logical; \code{TRUE} when ratio < threshold.}
#'   }
#'   For grouped predictions a \code{group} column is prepended.
#'   The object carries three attributes that drive \code{print()}:
#'   \describe{
#'     \item{\code{horizon}}{Named list with elements \code{value},
#'       \code{type} (\code{"observed"}, \code{"projected"}, or
#'       \code{"lower_bound"}), \code{last_informative},
#'       \code{description}, and (for projected) \code{se_t_star}.}
#'     \item{\code{horizon_by_group}}{For grouped objects: named list of
#'       per-group horizon lists.}
#'     \item{\code{threshold}}{The threshold value used.}
#'   }
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' df  <- data.frame(y = rnorm(20), year = 2001:2020, x1 = rnorm(20))
#' fit <- et_fit(y ~ x1, data = df,
#'               chains = 1, iter = 500, warmup = 250,
#'               cores = 1, refresh = 0)
#' new_df <- data.frame(x1 = rnorm(10), year = 2021:2030)
#' pred   <- et_predict(fit, newdata = new_df,
#'                      n_draws = 200, n_perturb = 50)
#' sl <- shelf_life(pred,
#'                  response_scale            = range(df$y),
#'                  ci_level                  = 0.90,
#'                  threshold                 = 1.0,
#'                  time_col                  = "year",
#'                  max_extrapolation_factor  = 10)
#' print(sl)
#' }
#' @export
shelf_life <- function(predictions,
                       response_scale,
                       ci_level = 0.90,
                       threshold = 1.0,
                       time_col = NULL,
                       min_slope_for_projection = 1e-4,
                       max_extrapolation_factor = 10,
                       ...,
                       plausible_range = NULL) {
  # Deprecated alias: plausible_range -> response_scale
  if (!is.null(plausible_range)) {
    if (!missing(response_scale)) {
      stop("Cannot specify both 'response_scale' and 'plausible_range'.")
    }
    warning(
      "'plausible_range' is deprecated; use 'response_scale' instead.",
      call. = FALSE
    )
    return(shelf_life(
      predictions              = predictions,
      response_scale           = plausible_range,
      ci_level                 = ci_level,
      threshold                = threshold,
      time_col                 = time_col,
      min_slope_for_projection = min_slope_for_projection,
      max_extrapolation_factor = max_extrapolation_factor,
      ...
    ))
  }
  UseMethod("shelf_life")
}

#' @export
shelf_life.et_prediction <- function(predictions,
                                     response_scale,
                                     ci_level = 0.90,
                                     threshold = 1.0,
                                     time_col = NULL,
                                     min_slope_for_projection = 1e-4,
                                     max_extrapolation_factor = 10,
                                     ...,
                                     plausible_range = NULL) {
  .compute_shelf_life_single(
    predictions = predictions,
    response_scale = response_scale,
    ci_level = ci_level,
    threshold = threshold,
    time_col = time_col,
    min_slope_for_projection = min_slope_for_projection,
    max_extrapolation_factor = max_extrapolation_factor
  )
}

#' @export
shelf_life.et_prediction_list <- function(predictions,
                                          response_scale,
                                          ci_level = 0.90,
                                          threshold = 1.0,
                                          time_col = NULL,
                                          min_slope_for_projection = 1e-4,
                                          max_extrapolation_factor = 10,
                                          ...,
                                          plausible_range = NULL) {
  horizon_by_group <- list()

  parts <- lapply(names(predictions$predictions), function(g) {
    pred <- predictions$predictions[[g]]
    if (is.null(pred)) return(NULL)

    sl <- .compute_shelf_life_single(
      pred, response_scale, ci_level, threshold,
      time_col, min_slope_for_projection, max_extrapolation_factor
    )
    horizon_by_group[[g]] <<- attr(sl, "horizon")
    cbind(data.frame(group = g, stringsAsFactors = FALSE), sl)
  })

  result <- do.call(rbind, Filter(Negate(is.null), parts))
  result <- structure(result, class = c("et_shelf_life", "data.frame"))
  attr(result, "horizon_by_group") <- horizon_by_group
  attr(result, "threshold") <- threshold
  result
}

#' @export
shelf_life.default <- function(predictions, ...) {
  stop("shelf_life() expects an et_prediction or et_prediction_list object.")
}

# ******************************************************************************
# Internal: compute shelf life for a single et_prediction
# ______________________________________________________________________________

.compute_shelf_life_single <- function(predictions, response_scale,
                                        ci_level, threshold, time_col,
                                        min_slope_for_projection,
                                        max_extrapolation_factor = 10) {
  ci_df <- predictions$credible_intervals
  avail_levels <- unique(ci_df$ci_level)

  if (!ci_level %in% avail_levels) {
    stop(
      "ci_level ", ci_level, " not found in predictions. ",
      "Available: ", paste(avail_levels, collapse = ", "),
      ". Re-run et_predict() with the desired level in ci_levels."
    )
  }

  ci_sub <- ci_df[ci_df$ci_level == ci_level, ]

  if (length(response_scale) != 2) {
    stop("response_scale must be a numeric vector of length 2: c(min, max).")
  }
  pr <- abs(diff(response_scale))
  if (pr < 1e-12) stop("response_scale min and max are equal.")

  time_vals <- if (!is.null(time_col) &&
                   time_col %in% colnames(predictions$newdata)) {
    predictions$newdata[[time_col]]
  } else {
    ci_sub$row_id
  }

  result_df <- data.frame(
    obs_id = ci_sub$row_id,
    time = time_vals,
    ci_width = ci_sub$width,
    plausible_range = pr,
    ratio = ci_sub$width / pr,
    informative = ci_sub$width < threshold * pr,
    stringsAsFactors = FALSE
  )

  # Compute shelf life horizon
  horizon <- .compute_horizon(result_df, threshold, min_slope_for_projection,
                              max_extrapolation_factor)

  result <- structure(result_df, class = c("et_shelf_life", "data.frame"))
  attr(result, "horizon") <- horizon
  attr(result, "threshold") <- threshold
  result
}

# ******************************************************************************
# Internal: derive horizon from a single-group shelf-life table
# ______________________________________________________________________________

.compute_horizon <- function(sl_df, threshold, min_slope,
                             max_extrapolation_factor = 10) {
  times <- sl_df$time
  ratios <- sl_df$ratio
  informative <- sl_df$informative

  # --- Mode 1: threshold already crossed ---
  if (any(!informative)) {
    first_cross  <- min(times[!informative])
    last_ok      <- if (any(informative)) max(times[informative]) else NA_real_
    return(list(
      value = first_cross,
      type = "observed",
      last_informative = last_ok,
      description = paste0(
        "Threshold (ratio >= ", threshold, ") first exceeded at ",
        first_cross, "."
      )
    ))
  }

  # --- Mode 2: all informative -- try linear extrapolation ---
  last_ok <- max(times)

  if (is.numeric(times) && length(times) >= 3) {
    lm_fit <- suppressWarnings(lm(ratios ~ times))
    slope <- coef(lm_fit)[["times"]]
    b0 <- coef(lm_fit)[["(Intercept)"]]

    if (!is.na(slope) && slope > min_slope) {
      proj <- (threshold - b0) / slope

      # SE of t* via delta method: t* = (tau - a) / b
      # d(t*)/d(a) = -1/b,  d(t*)/d(b) = -(tau - a)/b^2 = -t*/b
      vc   <- stats::vcov(lm_fit)
      va   <- vc["(Intercept)", "(Intercept)"]
      vb   <- vc["times", "times"]
      cab  <- vc["(Intercept)", "times"]
      se_t_star <- sqrt(max(0, (va + proj^2 * vb + 2 * proj * cab) / slope^2))

      # Cap extrapolation: if projection exceeds last_ok + factor * window,
      # the trend is too flat to give a meaningful estimate -- report lower bound.
      window   <- max(times) - min(times)
      cap_time <- max(times) + max_extrapolation_factor * window

      if (is.finite(max_extrapolation_factor) && proj > cap_time) {
        return(list(
          value = NA_real_,
          type = "lower_bound",
          last_informative = last_ok,
          description = paste0(
            "All ", nrow(sl_df), " forecast periods informative. ",
            "Linear trend (slope = ", round(slope, 5), " per time unit) ",
            "projects threshold crossing at ~", round(proj, 1),
            ", but this exceeds the extrapolation cap (",
            round(cap_time, 1), "). Shelf life > ", last_ok, "."
          )
        ))
      }

      return(list(
        value = proj,
        type = "projected",
        last_informative = last_ok,
        se_t_star = se_t_star,
        slope = slope,
        description = paste0(
          "All ", nrow(sl_df), " forecast periods informative. ",
          "Linear trend (slope = ", round(slope, 5), " per time unit) ",
          "projects threshold crossing at ~", round(proj, 1),
          " (SE = ", round(se_t_star, 2), ")."
        )
      ))
    }
  }

  # --- Mode 3: no upward trend -- lower bound only ---
  list(
    value = NA_real_,
    type = "lower_bound",
    last_informative = last_ok,
    description = paste0(
      "All ", nrow(sl_df), " forecast periods informative with no ",
      "upward trend in CI/range ratio. Shelf life > ", last_ok, "."
    )
  )
}

# ******************************************************************************
# Internal: format a horizon for printing
# ______________________________________________________________________________

.format_horizon <- function(hor) {
  if (is.null(hor)) return(NULL)
  switch(hor$type,
    observed  = paste0("~", hor$value,
                         " (observed -- threshold first exceeded)"),
    projected = paste0("~", round(hor$value, 1),
                         " (projected -- extrapolated beyond forecast window)"),
    lower_bound = paste0("> ", hor$last_informative,
                         " (lower bound -- all periods informative, no trend)")
  )
}

# ******************************************************************************
# S3 methods for et_shelf_life
# ______________________________________________________________________________

#' @export
print.et_shelf_life <- function(x, ...) {
  thr <- attr(x, "threshold")
  cat("ErrorTracer shelf life analysis\n")
  cat("  Observations     :", nrow(x), "\n")
  cat("  Plausible range  :", unique(x$plausible_range), "\n")
  cat("  Threshold        :", if (is.null(thr)) 1.0 else thr, "\n")

  if ("group" %in% colnames(x)) {
    groups  <- unique(x$group)
    hor_grp <- attr(x, "horizon_by_group")

    for (g in groups) {
      sub <- x[x$group == g, ]
      pct <- round(100 * mean(sub$informative), 1)
      hor <- if (!is.null(hor_grp)) hor_grp[[g]] else NULL
      hor_str <- if (!is.null(hor)) .format_horizon(hor) else ""
      cat(sprintf(
        "  [%s]  %.1f%% informative  mean ratio = %.3f  shelf life %s\n",
        g, pct, mean(sub$ratio), hor_str
      ))
    }
  } else {
    cat("  Informative      :", sum(x$informative), "/", nrow(x), "\n")
    cat("  Mean CI/range    :", round(mean(x$ratio), 3), "\n")
    cat("  Max CI/range     :", round(max(x$ratio), 3), "\n")

    hor <- attr(x, "horizon")
    if (!is.null(hor)) {
      cat("  Shelf life       :", .format_horizon(hor), "\n")
      cat("  Detail           :", hor$description, "\n")
    }
  }
  invisible(x)
}

#' @export
summary.et_shelf_life <- function(object, ...) {
  print(object)
  cat("\n--- Full table ---\n")
  print(as.data.frame(object))
  invisible(object)
}

# R/calibrate.R -- et_calibrate() and et_diagnose()

# ******************************************************************************
# et_calibrate(): coverage probability assessment
# ______________________________________________________________________________

#' Assess calibration of posterior predictive intervals
#'
#' Computes observed coverage probability at multiple nominal CI levels.
#' A well-calibrated Bayesian model should produce 90\% CIs that contain
#' the true value 90\% of the time, etc.
#'
#' @param predictions An \code{et_prediction} or \code{et_prediction_list}.
#' @param observed A \code{data.frame} with true response values.  Must have
#'   the same number of rows as \code{predictions$newdata} (rows are matched
#'   positionally) and a column with the true response values.
#' @param response_col Character.  Name of the response column in
#'   \code{observed}.  Defaults to the left-hand side of the model formula
#'   if it can be inferred, otherwise must be specified.
#' @param ci_levels Numeric vector.  CI levels to assess.  Defaults to all
#'   levels present in the \code{et_prediction} object.
#' @param ... Unused.
#' @return A \code{data.frame} with columns:
#'   \describe{
#'     \item{ci_level}{Nominal CI level.}
#'     \item{nominal}{Same as ci_level.}
#'     \item{observed_coverage}{Fraction of true values falling inside the CI.}
#'     \item{n_obs}{Number of observations used.}
#'     \item{calibration_error}{Signed difference: observed - nominal.
#'       Positive = over-coverage (CIs too wide / conservative).
#'       Negative = under-coverage (CIs too narrow / overconfident).}
#'     \item{sharpness}{Mean CI width across observations.  Sharpness and
#'       calibration are complementary: a model can be calibrated but useless
#'       if sharpness is poor (very wide CIs).}
#'   }
#'   For grouped predictions, a \code{group} column is prepended.
#' @examples
#' \donttest{
#' set.seed(1)
#' df  <- data.frame(y = rnorm(20), x1 = rnorm(20))
#' fit <- et_fit(y ~ x1, data = df,
#'               chains = 1, iter = 500, warmup = 250,
#'               cores = 1, refresh = 0)
#' valid_df <- data.frame(y = rnorm(5), x1 = rnorm(5))
#' pred <- et_predict(fit, newdata = valid_df,
#'                    n_draws = 200, n_perturb = 50)
#' cal <- et_calibrate(pred, observed = valid_df, response_col = "y")
#' print(cal)
#' }
#' @export
et_calibrate <- function(predictions, observed, response_col = NULL,
                          ci_levels = NULL, ...) {
  UseMethod("et_calibrate")
}

#' @export
et_calibrate.et_prediction <- function(predictions, observed,
                                         response_col = NULL,
                                         ci_levels = NULL, ...) {
  response_col <- .resolve_response_col(response_col, predictions$model)
  if (nrow(observed) != nrow(predictions$newdata)) {
    stop("observed must have the same number of rows as newdata (",
         nrow(predictions$newdata), "), got ", nrow(observed), ".")
  }
  y_true <- observed[[response_col]]
  if (is.null(y_true)) {
    stop("Column '", response_col, "' not found in observed data.frame.")
  }
  .compute_calibration(predictions$credible_intervals, y_true, ci_levels)
}

#' @export
et_calibrate.et_prediction_list <- function(predictions, observed,
                                              response_col = NULL,
                                              ci_levels = NULL, ...) {
  grouping <- predictions$grouping
  if (!grouping %in% colnames(observed)) {
    stop("Grouping column '", grouping, "' not found in observed.")
  }

  parts <- lapply(names(predictions$predictions), function(g) {
    pred <- predictions$predictions[[g]]
    if (is.null(pred)) return(NULL)
    obs_g <- observed[observed[[grouping]] == g, , drop = FALSE]
    if (nrow(obs_g) == 0) return(NULL)

    rc <- .resolve_response_col(response_col, pred$model)
    y_true <- obs_g[[rc]]
    if (is.null(y_true) || length(y_true) != nrow(pred$newdata)) return(NULL)

    cal <- .compute_calibration(pred$credible_intervals, y_true, ci_levels)
    cbind(data.frame(group = g, stringsAsFactors = FALSE), cal)
  })

  do.call(rbind, Filter(Negate(is.null), parts))
}

.compute_calibration <- function(ci_df, y_true, ci_levels) {
  avail <- unique(ci_df$ci_level)
  levels_use <- if (is.null(ci_levels)) avail else intersect(ci_levels, avail)

  rows <- lapply(levels_use, function(lv) {
    sub    <- ci_df[ci_df$ci_level == lv, ]
    inside <- y_true >= sub$lower & y_true <= sub$upper
    cov    <- mean(inside, na.rm = TRUE)
    data.frame(
      ci_level           = lv,
      nominal            = lv,
      observed_coverage  = cov,
      n_obs              = sum(!is.na(inside)),
      calibration_error  = cov - lv,   # signed: + over-coverage, - under-coverage
      sharpness          = mean(sub$width, na.rm = TRUE),
      stringsAsFactors   = FALSE
    )
  })
  do.call(rbind, rows)
}

.resolve_response_col <- function(response_col, model) {
  if (!is.null(response_col)) return(response_col)
  if (!is.null(model) && !is.null(model$formula)) {
    lhs <- all.vars(stats::update(model$formula, . ~ 0))
    if (length(lhs) > 0) return(lhs[1])
  }
  stop("Could not infer response column name. ",
       "Please supply response_col explicitly.")
}

# ******************************************************************************
# et_diagnose(): MCMC convergence + LOO diagnostics
# ______________________________________________________________________________

#' Diagnose a fitted ErrorTracer model
#'
#' Computes Rhat, effective sample size ratios, divergent transitions, and
#' leave-one-out cross-validation (LOO-CV) for a fitted \code{et_model}.
#'
#' @param model An \code{et_model} or \code{et_model_list}.
#' @param loo Logical.  Whether to run LOO-CV (can be slow; default
#'   \code{TRUE}).
#' @param ... Unused.
#' @return A list with elements:
#'   \describe{
#'     \item{convergence}{List: \code{rhat_max}, \code{rhat_all_ok},
#'       \code{neff_min}, \code{neff_all_ok}, \code{n_divergences}.}
#'     \item{loo}{List or \code{NULL}: \code{elpd_loo}, \code{p_loo},
#'       \code{looic}, \code{n_bad_pareto_k}, \code{loo_object}.}
#'     \item{summary}{Printed summary from \code{brms::summary()}.}
#'   }
#'   For \code{et_model_list}, a named list of per-group diagnostic lists
#'   plus an aggregated \code{summary} data.frame.
#' @export
et_diagnose <- function(model, loo = TRUE, ...) {
  UseMethod("et_diagnose")
}

#' @export
et_diagnose.et_model <- function(model, loo = TRUE, ...) {
  .diagnose_single(model$fit, loo = loo)
}

#' @export
et_diagnose.et_model_list <- function(model, loo = TRUE, ...) {
  diag_list <- lapply(names(model$models), function(g) {
    m <- model$models[[g]]
    if (is.null(m)) return(NULL)
    .diagnose_single(m$fit, loo = loo)
  })
  names(diag_list) <- names(model$models)

  # Build summary data.frame
  summary_df <- data.frame(
    group          = names(diag_list),
    rhat_ok        = vapply(diag_list, function(d) if (is.null(d)) NA else d$convergence$rhat_all_ok, logical(1)),
    neff_ok        = vapply(diag_list, function(d) if (is.null(d)) NA else d$convergence$neff_all_ok, logical(1)),
    n_divergences  = vapply(diag_list, function(d) if (is.null(d)) NA_real_ else d$convergence$n_divergences, numeric(1)),
    elpd_loo       = vapply(diag_list, function(d) if (is.null(d) || is.null(d$loo)) NA_real_ else d$loo$elpd_loo, numeric(1)),
    n_bad_pareto_k = vapply(diag_list, function(d) if (is.null(d) || is.null(d$loo)) NA_real_ else d$loo$n_bad_pareto_k, numeric(1)),
    stringsAsFactors = FALSE
  )

  list(per_group = diag_list, summary = summary_df)
}

.diagnose_single <- function(fit, loo) {
  rhat_vals  <- brms::rhat(fit)
  neff_vals  <- brms::neff_ratio(fit)
  nuts_df    <- brms::nuts_params(fit)
  divs       <- sum(nuts_df$Value[nuts_df$Parameter == "divergent__"], na.rm = TRUE)

  convergence <- list(
    rhat_max      = max(rhat_vals, na.rm = TRUE),
    rhat_all_ok   = all(rhat_vals < 1.05, na.rm = TRUE),
    neff_min      = min(neff_vals, na.rm = TRUE),
    neff_all_ok   = all(neff_vals > 0.1, na.rm = TRUE),
    n_divergences = divs
  )

  if (!convergence$rhat_all_ok) {
    .et_warn("Rhat > 1.05 for some parameters -- consider more iterations or ",
             "reparametrization.")
  }
  if (divs > 0) {
    .et_warn(divs, " divergent transitions -- consider increasing adapt_delta.")
  }

  loo_res <- NULL
  if (loo) {
    loo_res <- tryCatch({
      loo_obj <- brms::loo(fit)
      list(
        elpd_loo       = loo_obj$estimates["elpd_loo", "Estimate"],
        p_loo          = loo_obj$estimates["p_loo", "Estimate"],
        looic          = loo_obj$estimates["looic", "Estimate"],
        n_bad_pareto_k = sum(loo_obj$diagnostics$pareto_k > 0.7, na.rm = TRUE),
        loo_object     = loo_obj
      )
    }, error = function(e) {
      .et_warn("LOO-CV failed: ", e$message)
      NULL
    })
  }

  list(
    convergence = convergence,
    loo         = loo_res,
    summary     = summary(fit)
  )
}

# tests/testthat/test-integration.R
#
# Integration tests that exercise et_fit(), et_predict(), et_calibrate(), and
# et_diagnose() with a real (but minimal) Stan/brms fit.
#
# These tests are skipped on CRAN (Stan compilation is too slow) and require
# the rstan backend.  They are the primary source of coverage for:
#   R/fit.R, R/calibrate.R, and the brms-calling branches of R/predict.R
#
# Runtime: ~60-120 s on first call (C++ compilation); <10 s on reruns.

# ── Shared fixture: fit a minimal model once per file ───────────────────────

# testthat caches the environment per file, so we compute the fit at top level.
skip_if_not_installed("brms")
# covr 3.6.5 does not set NOT_CRAN, so skip_on_cran() would skip these tests
# during coverage measurement.  Detect the covr subprocess and opt in.
if (requireNamespace("covr", quietly = TRUE) && covr::in_covr()) {
  Sys.setenv(NOT_CRAN = "true")
}
skip_on_cran()

options(brms.backend = "rstan")

# Use a tiny synthetic dataset (n = 15) for speed
set.seed(42L)
.n   <- 15L
.int_df <- data.frame(
  y  = rnorm(.n, mean = 0.5),
  x1 = rnorm(.n),
  x2 = rnorm(.n)
)
.int_valid <- data.frame(
  y  = rnorm(5L, mean = 0.5),
  x1 = rnorm(5L),
  x2 = rnorm(5L)
)

# Extract priors via lm (no glmnet needed)
.lm_fit    <- lm(y ~ x1 + x2, data = .int_df)
.int_prior <- extract_priors(.lm_fit, multiplier = 2.0, min_sd = 0.1)

# Fit once — used by all tests in this file
suppressWarnings(
  .int_fit <- et_fit(
    formula = y ~ x1 + x2,
    data    = .int_df,
    priors  = .int_prior,
    chains  = 1L, iter = 400L, warmup = 200L,
    cores   = 1L, seed  = 42L, refresh = 0L,
    silent  = 2L
  )
)

# ── et_fit tests ─────────────────────────────────────────────────────────────

test_that("et_fit returns an et_model with the correct structure", {
  expect_s3_class(.int_fit, "et_model")
  expect_true(inherits(.int_fit$fit, "brmsfit"))
  expect_equal(nrow(.int_fit$data), .n)
  expect_true(!is.null(.int_fit$prior_spec))
})

test_that("print.et_model runs without error", {
  expect_output(print(.int_fit), "et_model")
})

test_that("summary.et_model runs without error", {
  expect_output(summary(.int_fit), "Fixed effects")
})

test_that("et_fit Rhat is below 1.10 for trivial model", {
  rhats <- brms::rhat(.int_fit$fit)
  expect_true(all(rhats < 1.10, na.rm = TRUE))
})

# ── et_predict tests ─────────────────────────────────────────────────────────

test_that("et_predict returns an et_prediction with correct dimensions", {
  pred <- et_predict(
    model     = .int_fit,
    newdata   = .int_valid,
    env_noise = list(x1 = 0.2, x2 = 0.1),
    n_draws   = 200L,
    ci_levels = c(0.50, 0.90),
    n_perturb = 50L
  )
  expect_s3_class(pred, "et_prediction")
  expect_equal(ncol(pred$posterior_predict), 5L)
  expect_equal(nrow(pred$posterior_predict), 200L)
  expect_true(all(c("row_id", "ci_level", "lower", "upper", "width") %in%
                    colnames(pred$credible_intervals)))
})

test_that("et_predict with no env_noise gives zero env_var", {
  pred_no_noise <- suppressWarnings(et_predict(
    model     = .int_fit,
    newdata   = .int_valid,
    env_noise = NULL,
    n_draws   = 100L,
    n_perturb = 50L
  ))
  expect_true(all(pred_no_noise$decomposition$env_var == 0))
})

test_that("et_predict with time-varying env_noise inflates later env_var", {
  noise_vec <- c(0.01, 0.01, 0.5, 0.5, 0.5)   # low for obs 1-2, high for 3-5
  pred_tv <- suppressWarnings(et_predict(
    model     = .int_fit,
    newdata   = .int_valid,
    env_noise = list(x1 = noise_vec),
    n_draws   = 200L,
    n_perturb = 200L
  ))
  env_var <- pred_tv$decomposition$env_var
  # High-noise obs (3-5) should have larger env_var than low-noise (1-2)
  expect_gt(mean(env_var[3:5]), mean(env_var[1:2]))
})

# ── et_diagnose tests ────────────────────────────────────────────────────────

test_that("et_diagnose returns convergence summary without LOO", {
  diag <- et_diagnose(.int_fit, loo = FALSE)
  expect_true(!is.null(diag$convergence))
  expect_true(is.logical(diag$convergence$rhat_all_ok))
  expect_true(is.logical(diag$convergence$neff_all_ok))
  expect_true(is.numeric(diag$convergence$n_divergences))
  expect_null(diag$loo)
})

test_that("et_diagnose with loo = TRUE returns LOO summary", {
  diag <- suppressWarnings(et_diagnose(.int_fit, loo = TRUE))
  expect_false(is.null(diag$loo))
  expect_true(is.numeric(diag$loo$elpd_loo))
})

# ── et_calibrate tests ───────────────────────────────────────────────────────

test_that("et_calibrate returns coverage data.frame with correct structure", {
  pred <- suppressWarnings(et_predict(
    .int_fit, .int_valid, n_draws = 200L, n_perturb = 50L
  ))
  cal <- et_calibrate(pred, observed = .int_valid,
                      response_col = "y", ci_levels = c(0.50, 0.90))
  expect_s3_class(cal, "data.frame")
  expect_true(all(c("ci_level", "nominal", "observed_coverage",
                    "n_obs", "calibration_error") %in% colnames(cal)))
  expect_equal(nrow(cal), 2L)
  expect_true(all(cal$observed_coverage >= 0 & cal$observed_coverage <= 1))
})

test_that("et_calibrate print method runs without error", {
  pred <- suppressWarnings(et_predict(
    .int_fit, .int_valid, n_draws = 200L, n_perturb = 50L
  ))
  cal <- et_calibrate(pred, observed = .int_valid, response_col = "y")
  expect_output(print(cal), "calibration")
})

# ── et_fit with lm prior (default method path) ───────────────────────────────

test_that("et_fit works with flat (NULL) priors", {
  suppressWarnings(
    fit_flat <- et_fit(
      formula = y ~ x1 + x2,
      data    = .int_df,
      priors  = NULL,
      chains  = 1L, iter = 300L, warmup = 150L,
      cores   = 1L, seed = 7L, refresh = 0L, silent = 2L
    )
  )
  expect_s3_class(fit_flat, "et_model")
})

# R/data.R — Documentation for bundled datasets

#' Simulated allele-frequency-change dataset for ErrorTracer tutorials
#'
#' A named list containing training data, forecast predictors, validation
#' responses, and ground-truth parameters for two synthetic SNP clusters (A
#' and B) at a hypothetical mountain plant site.  The dataset is designed to
#' exercise every step of the ErrorTracer pipeline and to support
#' parameter-recovery validation (true coefficients are known).
#'
#' @format A named \code{list} with seven elements:
#' \describe{
#'   \item{\code{train}}{
#'     \code{data.frame} with 100 rows and 6 columns (2 clusters \eqn{\times}
#'     50 training years, 1970–2019):
#'     \describe{
#'       \item{year}{Integer. Calendar year.}
#'       \item{cluster_id}{Character. SNP cluster identifier: \code{"A"} or
#'         \code{"B"}.}
#'       \item{Tmean}{Numeric. Standardised mean growing-season temperature
#'         (original unit: °C; standardised using training-period mean and SD).}
#'       \item{PPT}{Numeric. Standardised total growing-season precipitation
#'         (original unit: mm).}
#'       \item{SWE}{Numeric. Standardised peak snow water equivalent
#'         (original unit: mm).}
#'       \item{z_diff}{Numeric. Simulated allele-frequency change on the
#'         arcsin-sqrt scale (\eqn{\arcsin(\sqrt{f})} transformation).
#'         This transformation is unbounded and has no fixed
#'         [\eqn{-1}, 1] constraint; the plausible range should be derived
#'         from the training data or from the theoretical bounds
#'         (\eqn{[0,\, \pi/2]}\ on the arcsin scale).}
#'     }
#'   }
#'   \item{\code{forecast}}{
#'     \code{data.frame} with 30 rows and 5 columns (2 clusters \eqn{\times}
#'     15 forecast years, 2020–2034).  Same columns as \code{train} except
#'     \code{z_diff} is absent — these are the prediction targets.  Predictors
#'     are standardised using the training-period statistics stored in
#'     \code{standardization}.
#'   }
#'   \item{\code{validation}}{
#'     \code{data.frame} with 30 rows and 3 columns (\code{year},
#'     \code{cluster_id}, \code{z_diff}).  True response values for the
#'     forecast period; used with \code{\link{et_calibrate}} to assess
#'     posterior predictive coverage.
#'   }
#'   \item{\code{true_params}}{
#'     Named \code{list} with one element per cluster.  Each element is a named
#'     numeric vector of the true data-generating parameters:
#'     \code{intercept}, \code{Tmean}, \code{PPT}, \code{SWE}, and
#'     \code{sigma} (residual SD).
#'     \describe{
#'       \item{Cluster A}{\code{intercept = 0.00}, \code{Tmean = 0.50},
#'         \code{PPT = -0.30}, \code{SWE = 0.20}, \code{sigma = 0.30}}
#'       \item{Cluster B}{\code{intercept = 0.10}, \code{Tmean = 0.30},
#'         \code{PPT = -0.20}, \code{SWE = -0.25}, \code{sigma = 0.35}}
#'     }
#'   }
#'   \item{\code{env_noise}}{
#'     Named \code{list}.  Suggested per-predictor environmental noise SDs (on
#'     the standardised scale) for use with the \code{env_noise} argument of
#'     \code{\link{et_predict}}: \code{Tmean = 0.30}, \code{PPT = 0.20},
#'     \code{SWE = 0.15}.
#'   }
#'   \item{\code{standardization}}{
#'     Named \code{list} with one element per predictor.  Each element is a
#'     named numeric vector \code{c(mean = ..., sd = ...)} giving the
#'     training-period statistics used to standardise the data.  Needed if you
#'     wish to back-transform predictions to the original (unstandardised)
#'     scale with \code{\link{unstandardize}}.
#'   }
#'   \item{\code{description}}{
#'     Character string.  Human-readable description of the dataset and how it
#'     was generated.
#'   }
#' }
#'
#' @details
#' The climate time series are simulated as AR(1) processes with a warming
#' trend in \code{Tmean} (+0.015 °C yr\eqn{^{-1}}, cumulating to ~0.30 SD
#' above the training mean over the 15-year forecast window).  \code{SWE}
#' is generated with a weak dependence on \code{Tmean} (negative coupling)
#' and \code{PPT} (positive coupling) plus a dominant independent noise
#' component, so that the three predictors have only mild pairwise
#' correlation (|R| < 0.2 on the training subset).  This makes all
#' regression coefficients reliably identifiable in a single replicate.
#' All predictors are standardised using training-period statistics only.
#'
#' In addition to the three real climate predictors, the dataset includes
#' ten independent standardised nuisance predictors \code{d1}, ..., \code{d10}
#' with zero true effect on the response.  They give the regularized-prior
#' extraction step a real selection job: cv.glmnet (alpha = 0.5) is expected
#' to identify the three real predictors and shrink the ten dummies toward
#' zero.  Without them, regularization on three well-identified predictors
#' merely biases the true coefficients without selecting anything.
#'
#' Responses are generated from a linear model with Gaussian noise:
#' \deqn{z\_diff_t = \alpha + \beta_1 Tmean_t + \beta_2 PPT_t +
#'       \beta_3 SWE_t + \varepsilon_t, \quad \varepsilon_t \sim
#'       \mathcal{N}(0, \sigma^2)}
#' The true parameters are stored in \code{et_sim$true_params} for
#' parameter-recovery validation.
#'
#' The dataset was generated with \code{set.seed(111)}.  The full generation
#' script is at \code{data-raw/generate_et_sim.R}.
#'
#' @examples
#' data(et_sim)
#'
#' # Inspect structure
#' str(et_sim, max.level = 2)
#'
#' # Training data
#' head(et_sim$train)
#'
#' # True parameters for cluster A
#' et_sim$true_params$A
#'
#' # Suggested noise SDs for et_predict()
#' et_sim$env_noise
#'
#' @source Generated by \code{data-raw/generate_et_sim.R} with
#'   \code{set.seed(111)}.
"et_sim"

# R/fit.R — et_fit(): Bayesian model fitting via brms

# ******************************************************************************
# Main fitting function
# ______________________________________________________________________________

#' Fit a Bayesian regression model with informed priors
#'
#' Wraps \code{brms::brm()} and attaches the prior specification,
#' training data reference, and configuration for downstream uncertainty
#' decomposition.  Pass \code{priors} from \code{\link{extract_priors}} to
#' use regularized-model coefficients as prior means; omit it for
#' default (weakly informative) priors.
#'
#' @param formula An R formula, e.g.\ \code{response ~ .} or
#'   \code{y ~ x1 + x2}.
#' @param data A \code{data.frame} with all predictors and the response.
#' @param priors An \code{et_prior_spec} object from
#'   \code{\link{extract_priors}}, or a \code{brmsprior} object, or
#'   \code{NULL} for brms defaults.
#' @param chains Integer.  Number of MCMC chains (default 4).
#' @param iter Integer.  Total iterations per chain, including warmup
#'   (default 2000).
#' @param warmup Integer.  Warmup iterations per chain (default
#'   \code{floor(iter / 2)}).
#' @param cores Integer.  Parallel cores (default
#'   \code{min(chains, parallel::detectCores())}).
#' @param seed Integer.  Random seed for reproducibility (default 42).
#' @param adapt_delta Numeric.  Target acceptance probability for HMC
#'   (default 0.95).
#' @param max_treedepth Integer.  Maximum tree depth (default 12).
#' @param grouping Character.  Name of a column in \code{data} to use for
#'   grouping.  If non-\code{NULL}, one model is fitted per unique group
#'   value and an \code{et_model_list} is returned.
#' @param eiv Optional errors-in-variables specification.  A named list /
#'   vector mapping predictor names to either a scalar SD or a vector of
#'   per-row SDs (length \code{nrow(data)}).  For each entry, the formula
#'   term for that predictor is rewritten as \code{brms::me(pred, se_pred)}
#'   (an auxiliary \code{se_<pred>} column is appended to \code{data}), so
#'   the posterior reflects measurement error in the predictor as well as
#'   coefficient uncertainty.  The beta posteriors widen accordingly, which
#'   partially absorbs what ErrorTracer's downstream \code{env_var}
#'   component would otherwise report.  When \code{eiv} is supplied together
#'   with an \code{et_prior_spec} from \code{\link{extract_priors}}, the
#'   informed priors are \emph{dropped} because they target \code{class = "b"}
#'   terms and \code{me()} terms live under \code{class = "bsp"}; brms
#'   defaults are used instead (and a warning is logged).
#' @param silent Integer passed to \code{brms::brm()} (default 2, no Stan
#'   output).
#' @param ... Additional arguments passed to \code{brms::brm()}.
#' @return An \code{et_model} object (or an \code{et_model_list} if
#'   \code{grouping} is specified).
#' @examples
#' \donttest{
#' set.seed(1)
#' df  <- data.frame(y = rnorm(20), x1 = rnorm(20), x2 = rnorm(20))
#' ps  <- extract_priors(lm(y ~ x1 + x2, data = df))
#' fit <- et_fit(y ~ x1 + x2, data = df, priors = ps,
#'               chains = 1, iter = 500, warmup = 250,
#'               cores = 1, refresh = 0)
#' print(fit)
#' }
#' @export
et_fit <- function(formula,
                   data,
                   priors = NULL,
                   chains = 4L,
                   iter = 2000L,
                   warmup = floor(iter / 2),
                   cores = min(chains, parallel::detectCores()),
                   seed = 42L,
                   adapt_delta = 0.95,
                   max_treedepth = 12L,
                   grouping = NULL,
                   eiv = NULL,
                   silent = 2L,
                   ...) {

  # Rewrite formula + augment data for measurement-error predictors if eiv
  # was supplied. Strips any et_prior_spec priors because their (class="b",
  # coef=pred_name) entries don't apply to me() (class="bsp") terms.
  eiv_spec <- NULL
  if (!is.null(eiv)) {
    rewrite  <- .apply_eiv(formula, data, eiv)
    formula  <- rewrite$formula
    data     <- rewrite$data
    eiv_spec <- rewrite$eiv_spec
    # brms parses me() via eval() against the search path, so brms must be
    # attached (not just imported) for me() to resolve. Attach it once.
    if (!"package:brms" %in% search()) {
      suppressPackageStartupMessages(attachNamespace("brms"))
    }
    if (inherits(priors, "et_prior_spec")) {
      .et_warn("eiv specified; dropping informed priors (they target ",
               "class='b' terms but me() coefficients are class='bsp'). ",
               "Using brms defaults.")
      priors <- NULL
    }
  }

  if (!is.null(grouping)) {
    result <- .et_fit_grouped(
      formula = formula, data = data, priors = priors,
      chains = chains, iter = iter, warmup = warmup, cores = cores,
      seed = seed, adapt_delta = adapt_delta, max_treedepth = max_treedepth,
      grouping = grouping, silent = silent, ...
    )
    if (!is.null(eiv_spec)) result$eiv_spec <- eiv_spec
    return(result)
  }

  result <- .et_fit_single(
    formula = formula, data = data, priors = priors,
    chains = chains, iter = iter, warmup = warmup, cores = cores,
    seed = seed, adapt_delta = adapt_delta, max_treedepth = max_treedepth,
    silent = silent, ...
  )
  if (!is.null(eiv_spec)) result$eiv_spec <- eiv_spec
  result
}

# ******************************************************************************
# Internal: rewrite formula and augment data for errors-in-variables
# ______________________________________________________________________________

# Given (formula, data, eiv), return a list(formula = new, data = new).
# For each named predictor in eiv:
#   * append a column se_<pred> to data (scalar recycled or vector of length
#     nrow(data)),
#   * substitute pred -> me(pred, se_<pred>) in the formula RHS.
# The substitution preserves everything else in the formula (intercept,
# interactions, factor contrasts). Matches only on whole-token predictor
# names to avoid partial-string collisions.
.apply_eiv <- function(formula, data, eiv) {
  if (!is.list(eiv) && !is.numeric(eiv)) {
    stop("eiv must be a named list or named numeric vector.")
  }
  if (is.null(names(eiv)) || any(!nzchar(names(eiv)))) {
    stop("eiv must have names matching predictor columns.")
  }
  eiv <- as.list(eiv)

  term_labels <- attr(stats::terms(formula, data = data), "term.labels")
  missing_preds <- setdiff(names(eiv), colnames(data))
  if (length(missing_preds)) {
    stop("eiv references column(s) not in data: ",
         paste(missing_preds, collapse = ", "))
  }

  n_obs <- nrow(data)
  for (p in names(eiv)) {
    v <- as.numeric(eiv[[p]])
    if (length(v) == 1L) v <- rep(v, n_obs)
    if (length(v) != n_obs) {
      stop("eiv[['", p, "']] has length ", length(v),
           " but data has ", n_obs, " row(s).")
    }
    if (any(is.na(v) | v < 0)) {
      stop("eiv[['", p, "']] must be non-negative and finite.")
    }
    data[[paste0("se_", p)]] <- v
  }

  rhs <- deparse(formula[[3L]], width.cutoff = 500L)
  rhs <- paste(rhs, collapse = " ")
  for (p in names(eiv)) {
    pattern <- paste0("(?<![A-Za-z0-9_.])", p, "(?![A-Za-z0-9_.])")
    replacement <- paste0("me(", p, ", se_", p, ")")
    rhs <- gsub(pattern, replacement, rhs, perl = TRUE)
  }
  lhs <- deparse(formula[[2L]], width.cutoff = 500L)
  # brms evaluates `me()` during formula parsing via eval() on the formula's
  # own environment, so we bind `me` (and `mi`, for completeness) directly
  # in that environment rather than relying on brms being attached.
  env <- new.env(parent = environment(formula) %||% parent.frame())
  env$me <- brms::me
  if (exists("mi", envir = asNamespace("brms"))) env$mi <- brms::mi
  new_formula <- stats::as.formula(paste(lhs, "~", rhs), env = env)

  list(formula = new_formula, data = data, eiv_spec = eiv)
}

# ******************************************************************************
# Internal: fit a single model
# ______________________________________________________________________________

.et_fit_single <- function(formula, data, priors, chains, iter, warmup,
                            cores, seed, adapt_delta, max_treedepth,
                            silent, ...) {

  brms_prior <- if (inherits(priors, "et_prior_spec")) priors$prior
                else priors  # brmsprior or NULL

  n_pred <- length(attr(stats::terms(formula, data = data), "term.labels"))
  .et_info("Fitting Bayesian model: ", deparse(formula),
           " (", nrow(data), " obs, ~", n_pred, " predictors)")

  fit <- brms::brm(
    formula = brms::bf(formula),
    data = data,
    prior = brms_prior,
    chains = chains,
    iter = iter,
    warmup = warmup,
    cores = cores,
    seed = seed,
    control = list(adapt_delta = adapt_delta, max_treedepth = max_treedepth),
    silent  = silent,
    refresh = 0,
    ...
  )

  structure(
    list(
      fit = fit,
      prior_spec = if (inherits(priors, "et_prior_spec")) priors else NULL,
      formula = formula,
      data = data,
      config = list(
        chains = chains, iter = iter, warmup = warmup,
        cores = cores,  seed = seed,
        adapt_delta = adapt_delta, max_treedepth = max_treedepth
      )
    ),
    class = "et_model"
  )
}

# ******************************************************************************
# Internal: fit one model per group
# ______________________________________________________________________________

.et_fit_grouped <- function(formula, data, priors, chains, iter, warmup,
                             cores, seed, adapt_delta, max_treedepth,
                             grouping, silent, ...) {

  if (!grouping %in% colnames(data)) {
    stop("grouping column '", grouping, "' not found in data.")
  }

  groups <- unique(data[[grouping]])
  models <- vector("list", length(groups))
  names(models) <- as.character(groups)

  for (g in groups) {
    gname <- as.character(g)
    .et_info("Fitting model for group: ", gname)
    sub_data <- data[data[[grouping]] == g, , drop = FALSE]

    # Use group-specific prior if priors is a named list
    group_prior <- if (is.list(priors) && !inherits(priors, "et_prior_spec")) {
      priors[[gname]]
    } else {
      priors
    }

    models[[gname]] <- tryCatch(
      .et_fit_single(
        formula = formula, data = sub_data, priors = group_prior,
        chains = chains, iter = iter, warmup = warmup, cores = cores,
        seed = seed, adapt_delta = adapt_delta, max_treedepth = max_treedepth,
        silent = silent, ...
      ),
      error = function(e) {
        .et_error("Failed to fit model for group ", gname, ": ", e$message)
        NULL
      }
    )
  }

  structure(
    list(
      models   = models,
      grouping = grouping,
      formula  = formula
    ),
    class = "et_model_list"
  )
}

# ******************************************************************************
# S3 methods for et_model
# ______________________________________________________________________________

#' @export
print.et_model <- function(x, ...) {
  cat("ErrorTracer model (et_model)\n")
  cat("  Formula :", deparse(x$formula), "\n")
  cat("  n obs   :", nrow(x$data), "\n")
  cat("  Chains  :", x$config$chains,
      "  Iter:", x$config$iter,
      "  Warmup:", x$config$warmup, "\n")
  if (!is.null(x$prior_spec)) {
    cat("  Priors  : informed (", x$prior_spec$method, ", ",
        length(x$prior_spec$pred_names), " predictors)\n", sep = "")
  } else {
    cat("  Priors  : brms defaults\n")
  }
  rhat_max <- tryCatch(max(brms::rhat(x$fit), na.rm = TRUE), error = function(e) NA)
  cat("  Rhat max:", if (is.na(rhat_max)) "NA" else round(rhat_max, 3), "\n")
  invisible(x)
}

#' @export
summary.et_model <- function(object, ...) {
  cat("=== ErrorTracer model summary ===\n\n")
  print(object)
  cat("\n--- Fixed effects ---\n")
  print(brms::fixef(object$fit))
  invisible(object)
}

# ******************************************************************************
# S3 methods for et_model_list
# ______________________________________________________________________________

#' @export
print.et_model_list <- function(x, ...) {
  cat("ErrorTracer grouped model list (et_model_list)\n")
  cat("  Grouping :", x$grouping, "\n")
  cat("  Formula  :", deparse(x$formula), "\n")
  cat("  Groups   :", length(x$models), "\n")
  fitted <- sum(!vapply(x$models, is.null, logical(1)))
  cat("  Fitted   :", fitted, "/", length(x$models), "\n")
  invisible(x)
}

#' @export
summary.et_model_list <- function(object, ...) {
  print(object)
  cat("\n--- Per-group Rhat max ---\n")
  for (nm in names(object$models)) {
    m <- object$models[[nm]]
    if (is.null(m)) {
      cat("  ", nm, ": FAILED\n")
    } else {
      rhat_max <- tryCatch(max(brms::rhat(m$fit), na.rm = TRUE), error = function(e) NA)
      cat(sprintf("  %-20s  Rhat max = %.3f\n", nm,
                  if (is.na(rhat_max)) NA else rhat_max))
    }
  }
  invisible(object)
}

# R/plot.R — et_plot_*(): ggplot2 visualizations (all return ggplot objects)

# ******************************************************************************
# et_plot_decomposition(): stacked bar of variance components
# ______________________________________________________________________________

#' Plot uncertainty decomposition
#'
#' Produces a stacked bar chart showing the relative contributions of
#' parameter, environmental, and residual variance for each observation,
#' plus a fourth temporal-autocorrelation component when present in
#' \code{decomp} (see \code{\link{decompose_uncertainty}}).
#'
#' @param decomp A \code{data.frame} from \code{\link{decompose_uncertainty}}
#'   or directly the \code{$decomposition} slot of an \code{et_prediction}.
#' @param proportional Logical.  If \code{TRUE} (default), bars are scaled to
#'   sum to 1 (proportional contribution).  If \code{FALSE}, raw variances
#'   are shown.
#' @param group_col Character.  Optional name of a grouping column in
#'   \code{decomp} (present for grouped predictions).
#' @return A \code{ggplot2} object.
#' @export
et_plot_decomposition <- function(decomp, proportional = TRUE,
                                   group_col = NULL) {
  components <- c("param_var", "env_var", "residual_var")
  if ("temporal_var" %in% colnames(decomp)) {
    components <- c(components, "temporal_var")
  }
  long <- .pivot_decomp(decomp, components)

  x_col <- if (!is.null(group_col) && group_col %in% colnames(decomp)) {
    group_col
  } else {
    "obs_id"
  }

  fill_values <- c(
    "Parameter"     = "#2166AC",
    "Environmental" = "#4DAC26",
    "Residual"      = "#D1E5F0",
    "Temporal"      = "#B2182B"
  )

  p <- ggplot2::ggplot(long, ggplot2::aes(
    x    = .data[[x_col]],
    y    = .data[["value"]],
    fill = .data[["component"]]
  )) +
    ggplot2::geom_col(
      position = if (proportional) "fill" else "stack"
    ) +
    ggplot2::scale_fill_manual(values = fill_values, name = "Component") +
    ggplot2::labs(
      title = "Uncertainty Decomposition",
      x     = if (x_col == "obs_id") "Observation" else x_col,
      y     = if (proportional) "Proportion of variance" else "Variance"
    ) +
    et_theme() +
    ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))

  p
}

# ******************************************************************************
# et_plot_shelf_life(): CI width vs time
# ______________________________________________________________________________

#' Plot forecast shelf life
#'
#' Shows how credible interval width grows over the forecast horizon and
#' marks the threshold beyond which the forecast is uninformative.
#'
#' @param sl An \code{et_shelf_life} object from \code{\link{shelf_life}}.
#' @param show_ratio Logical.  If \code{TRUE} (default), plots the
#'   ratio (CI width / plausible range) rather than raw CI width.
#' @return A \code{ggplot2} object.
#' @export
et_plot_shelf_life <- function(sl, show_ratio = TRUE) {
  y_col  <- if (show_ratio) "ratio"    else "ci_width"
  hline  <- if (show_ratio) 1.0        else unique(sl$plausible_range)[1]
  y_lab  <- if (show_ratio) "CI width / plausible range" else "CI width"
  hline_lab <- if (show_ratio) "Uninformative threshold (ratio = 1)"
               else "Plausible range"

  has_group <- "group" %in% colnames(sl)

  p <- ggplot2::ggplot(sl, ggplot2::aes(
    x     = .data[["time"]],
    y     = .data[[y_col]],
    color = if (has_group) .data[["group"]] else NULL,
    group = if (has_group) .data[["group"]] else NULL
  )) +
    ggplot2::geom_line(linewidth = 0.9) +
    ggplot2::geom_point(size = 2.5) +
    ggplot2::geom_hline(yintercept = hline, linetype = "dashed", color = "grey40") +
    ggplot2::annotate("text", x = min(sl$time), y = hline * 1.04,
                      label = hline_lab, hjust = 0, color = "grey40", size = 3) +
    ggplot2::labs(
      title = "Forecast Shelf Life",
      x     = "Time",
      y     = y_lab,
      color = if (has_group) "Group" else NULL
    ) +
    et_theme()

  if (has_group) p <- p + ggplot2::scale_color_viridis_d()

  p
}

# ******************************************************************************
# et_plot_calibration(): observed vs nominal coverage
# ______________________________________________________________________________

#' Plot calibration: observed vs nominal coverage
#'
#' A well-calibrated model produces points along the 1:1 diagonal.
#' Points above the diagonal indicate over-coverage (conservative);
#' below indicates under-coverage (anti-conservative).
#'
#' @param cal A \code{data.frame} from \code{\link{et_calibrate}}.
#' @param group_col Optional character.  Name of a column in \code{cal} that
#'   identifies sub-groups (e.g.\ \code{"species"}, \code{"cluster_id"}).
#'   When \code{NULL} (default), \code{et_plot_calibration} uses the
#'   \code{group} column if present; otherwise it auto-detects any single
#'   non-canonical column with more than one unique value and treats it as
#'   the grouping.  Set to \code{NA} to force a single un-grouped series.
#' @return A \code{ggplot2} object.
#' @export
et_plot_calibration <- function(cal, group_col = NULL) {
  canonical <- c("ci_level", "nominal", "observed_coverage",
                 "n_obs", "calibration_error", "sharpness")

  if (is.null(group_col)) {
    if ("group" %in% colnames(cal)) {
      group_col <- "group"
    } else {
      extras <- setdiff(colnames(cal), canonical)
      extras <- extras[vapply(extras,
                              function(x) length(unique(cal[[x]])) > 1,
                              logical(1))]
      if (length(extras) == 1L) group_col <- extras
    }
  } else if (is.na(group_col)) {
    group_col <- NULL
  }

  has_group <- !is.null(group_col) && group_col %in% colnames(cal)

  p <- ggplot2::ggplot(cal, ggplot2::aes(
    x = .data[["nominal"]],
    y = .data[["observed_coverage"]]
  )) +
    ggplot2::geom_abline(slope = 1, intercept = 0,
                         linetype = "dashed", color = "grey50") +
    ggplot2::labs(
      title = "Calibration: Observed vs Nominal Coverage",
      x     = "Nominal CI level",
      y     = "Observed coverage"
    ) +
    ggplot2::coord_equal(xlim = c(0, 1), ylim = c(0, 1)) +
    et_theme()

  if (has_group) {
    p <- p +
      ggplot2::geom_line(ggplot2::aes(
        color = .data[[group_col]], group = .data[[group_col]]
      ), alpha = 0.6) +
      ggplot2::geom_point(ggplot2::aes(color = .data[[group_col]]), size = 3) +
      ggplot2::scale_color_viridis_d(name = group_col)
  } else {
    p <- p +
      ggplot2::geom_line(alpha = 0.7) +
      ggplot2::geom_point(size = 3, color = "steelblue")
  }

  p
}

# ******************************************************************************
# et_plot_forecast(): fan chart of posterior predictive intervals
# ______________________________________________________________________________

#' Plot posterior predictive fan chart
#'
#' Overlays nested credible interval ribbons on the median forecast, with
#' optional observed values for calibration assessment.
#'
#' @param predictions An \code{et_prediction} object.
#' @param observed Optional \code{data.frame} with true response values.
#'   If provided, points are overlaid.
#' @param response_col Character.  Name of the response column in
#'   \code{observed}.
#' @param time_col Character.  Column in \code{predictions$newdata} used
#'   as the x-axis.  Defaults to observation index.
#' @return A \code{ggplot2} object.
#' @export
et_plot_forecast <- function(predictions, observed = NULL,
                              response_col = NULL, time_col = NULL) {
  ci_df <- predictions$credible_intervals

  time_vals <- if (!is.null(time_col) &&
                   time_col %in% colnames(predictions$newdata)) {
    rep(predictions$newdata[[time_col]],
        times = length(unique(ci_df$ci_level)))
  } else {
    ci_df$row_id
  }
  ci_df$time <- time_vals

  p <- ggplot2::ggplot(ci_df, ggplot2::aes(x = .data[["time"]])) +
    ggplot2::geom_ribbon(ggplot2::aes(
      ymin  = .data[["lower"]],
      ymax  = .data[["upper"]],
      alpha = .data[["ci_level"]],
      group = .data[["ci_level"]]
    ), fill = "steelblue") +
    ggplot2::geom_line(
      data = ci_df[ci_df$ci_level == max(ci_df$ci_level), ],
      ggplot2::aes(y = .data[["median"]]),
      color = "darkblue", linewidth = 0.9
    ) +
    ggplot2::scale_alpha_continuous(range = c(0.12, 0.45), guide = "none") +
    ggplot2::labs(
      title = "Forecast Fan Chart",
      x     = if (!is.null(time_col)) time_col else "Observation",
      y     = "Response"
    ) +
    et_theme()

  if (!is.null(observed) && !is.null(response_col)) {
    obs_time <- if (!is.null(time_col) &&
                    time_col %in% colnames(observed)) {
      observed[[time_col]]
    } else {
      seq_len(nrow(observed))
    }
    obs_df <- data.frame(
      time     = obs_time,
      response = observed[[response_col]],
      stringsAsFactors = FALSE
    )
    p <- p + ggplot2::geom_point(
      data = obs_df,
      ggplot2::aes(x = .data[["time"]], y = .data[["response"]]),
      color = "firebrick", size = 2.5, inherit.aes = FALSE
    )
  }

  p
}

# ******************************************************************************
# et_plot_prior_posterior(): density overlay
# ______________________________________________________________________________

#' Plot prior vs posterior distributions for model coefficients
#'
#' Overlays prior and posterior density for each predictor coefficient,
#' visualising how much the data update the priors.
#'
#' @param model An \code{et_model} object.
#' @param max_preds Integer.  Maximum number of predictors to show
#'   (default 8).  Predictors are shown in the order they appear in the
#'   prior specification.
#' @param n_prior_draws Integer.  Number of random draws for the prior
#'   density (default 4000).
#' @return A \code{ggplot2} object.
#' @export
et_plot_prior_posterior <- function(model, max_preds = 8L,
                                     n_prior_draws = 4000L) {
  if (!inherits(model, "et_model")) {
    stop("model must be an et_model object.")
  }
  if (is.null(model$prior_spec)) {
    stop("No et_prior_spec attached to model. ",
         "Prior/posterior plot requires a model fitted via et_fit() with ",
         "priors from extract_priors().")
  }

  fit     <- model$fit
  spec    <- model$prior_spec
  preds   <- head(spec$pred_names, max_preds)

  post_draws <- as.data.frame(fit)
  beta_cols  <- paste0("b_", preds)
  avail      <- intersect(beta_cols, colnames(post_draws))
  if (length(avail) == 0) stop("No beta columns found in posterior draws.")

  post_long <- do.call(rbind, lapply(avail, function(bc) {
    pname <- sub("^b_", "", bc)
    data.frame(
      parameter    = pname,
      value        = post_draws[[bc]],
      distribution = "Posterior",
      stringsAsFactors = FALSE
    )
  }))

  prior_long <- do.call(rbind, lapply(avail, function(bc) {
    pname <- sub("^b_", "", bc)
    idx   <- match(pname, spec$pred_names)
    mu    <- if (!is.null(spec$coefs)) spec$coefs[idx] else 0
    sd    <- max(spec$multiplier * abs(mu), spec$min_sd)
    data.frame(
      parameter    = pname,
      value        = stats::rnorm(n_prior_draws, mu, sd),
      distribution = "Prior",
      stringsAsFactors = FALSE
    )
  }))

  combined <- rbind(post_long, prior_long)
  combined$distribution <- factor(combined$distribution, levels = c("Prior", "Posterior"))

  ggplot2::ggplot(combined,
                  ggplot2::aes(x = .data[["value"]], fill = .data[["distribution"]])) +
    ggplot2::geom_density(alpha = 0.45) +
    ggplot2::facet_wrap(~ .data[["parameter"]], scales = "free") +
    ggplot2::scale_fill_manual(
      values = c("Prior" = "grey60", "Posterior" = "steelblue"),
      name   = NULL
    ) +
    ggplot2::geom_vline(xintercept = 0, linetype = "dotted", color = "grey30") +
    ggplot2::labs(
      title = "Prior vs Posterior: Regression Coefficients",
      x     = "Value",
      y     = "Density"
    ) +
    et_theme()
}

# ******************************************************************************
# et_plot_coefficients(): forest plot, Bayesian vs regularized
# ______________________________________________________________________________

#' Forest plot of regression coefficients
#'
#' Compares Bayesian posterior estimates (95\% CI) with the regularized
#' coefficient values used as prior means (shown as crosses).
#'
#' @param model An \code{et_model} or \code{et_model_list}.
#' @return A \code{ggplot2} object.
#' @export
et_plot_coefficients <- function(model) {
  UseMethod("et_plot_coefficients")
}

#' @export
et_plot_coefficients.et_model <- function(model) {
  coef_df <- .extract_coef_df(model, group_label = NULL)
  .forest_plot(coef_df, facet_by_group = FALSE)
}

#' @export
et_plot_coefficients.et_model_list <- function(model) {
  parts <- lapply(names(model$models), function(g) {
    m <- model$models[[g]]
    if (is.null(m)) return(NULL)
    .extract_coef_df(m, group_label = g)
  })
  coef_df <- do.call(rbind, Filter(Negate(is.null), parts))
  .forest_plot(coef_df, facet_by_group = TRUE)
}

.extract_coef_df <- function(model, group_label) {
  fit  <- model$fit
  spec <- model$prior_spec
  post <- brms::fixef(fit)
  fe   <- rownames(post)
  fe   <- fe[fe != "Intercept"]

  rows <- lapply(fe, function(pname) {
    enet_val <- if (!is.null(spec) && !is.null(spec$coefs) &&
                    pname %in% names(spec$coefs)) {
      spec$coefs[pname]
    } else {
      NA_real_
    }
    df <- data.frame(
      predictor  = pname,
      post_mean  = post[pname, "Estimate"],
      post_lower = post[pname, "Q2.5"],
      post_upper = post[pname, "Q97.5"],
      enet_coef  = enet_val,
      stringsAsFactors = FALSE
    )
    if (!is.null(group_label)) df$group <- group_label
    df
  })
  do.call(rbind, rows)
}

.forest_plot <- function(coef_df, facet_by_group) {
  p <- ggplot2::ggplot(coef_df,
    ggplot2::aes(y = .data[["predictor"]])) +
    ggplot2::geom_pointrange(
      ggplot2::aes(
        x    = .data[["post_mean"]],
        xmin = .data[["post_lower"]],
        xmax = .data[["post_upper"]]
      ),
      color = "steelblue", linewidth = 0.6
    ) +
    ggplot2::geom_vline(xintercept = 0, linetype = "dashed", color = "grey40") +
    ggplot2::labs(
      title = "Coefficient Estimates: Posterior (blue) vs Regularized (red \u00d7)",
      x     = "Coefficient",
      y     = NULL
    ) +
    et_theme()

  # Add enet estimates if present
  has_enet <- any(!is.na(coef_df$enet_coef))
  if (has_enet) {
    p <- p + ggplot2::geom_point(
      ggplot2::aes(x = .data[["enet_coef"]]),
      color = "firebrick", shape = 4, size = 3, na.rm = TRUE
    )
  }

  if (facet_by_group && "group" %in% colnames(coef_df)) {
    p <- p + ggplot2::facet_wrap(~ .data[["group"]], scales = "free")
  }

  p
}

# ******************************************************************************
# et_plot_sensitivity(): noise fraction vs horizon / env_share
# ______________________________________________________________________________

#' Plot a sensitivity profile
#'
#' Visualises the output of \code{\link{et_sensitivity_profile}}: for each
#' noise grid point, shows how the \strong{environmental share} of total
#' variance and the \strong{forecast horizon} respond.  Observed horizons
#' are drawn as solid points; projected horizons are hollow points;
#' lower-bound rows are drawn as upward arrows at the last informative time.
#'
#' The x-axis is the noise fraction (when the grid was built from
#' \code{fraction_grid}) or the grid step label otherwise.  When both a
#' numeric fraction and a descriptive label exist, the fraction is preferred
#' for continuous x-positioning.
#'
#' @param sens An \code{et_sensitivity} object from
#'   \code{\link{et_sensitivity_profile}}.
#' @param show \code{"horizon"} (default) shows the shelf-life horizon;
#'   \code{"env_share"} shows \code{env_var / total_var}; \code{"ratio"}
#'   shows the mean CI width / plausible range.
#' @return A \code{ggplot2} object.
#' @seealso \code{\link{et_sensitivity_profile}}
#' @export
et_plot_sensitivity <- function(sens, show = c("horizon", "env_share", "ratio")) {
  if (!inherits(sens, "et_sensitivity")) {
    stop("et_plot_sensitivity() expects an et_sensitivity object.")
  }
  show <- match.arg(show)

  df <- as.data.frame(sens)
  x_is_numeric <- !all(is.na(df$fraction))
  df$x <- if (x_is_numeric) df$fraction else seq_len(nrow(df))
  x_lab <- if (x_is_numeric) "Noise fraction of predictor SD" else "Grid step"

  if (show == "horizon") {
    df$y <- df$horizon
    df$point_shape <- factor(
      ifelse(df$horizon_type == "observed", "observed",
      ifelse(df$horizon_type == "projected", "projected", "lower bound")),
      levels = c("observed", "projected", "lower bound")
    )
    p <- ggplot2::ggplot(df, ggplot2::aes(x = .data[["x"]], y = .data[["y"]])) +
      ggplot2::geom_line(na.rm = TRUE, colour = "#2166AC") +
      ggplot2::geom_point(ggplot2::aes(shape = .data[["point_shape"]]),
                          size = 3, na.rm = TRUE) +
      ggplot2::scale_shape_manual(
        values = c("observed" = 16, "projected" = 21, "lower bound" = 2),
        drop = FALSE, name = "Horizon type"
      ) +
      ggplot2::labs(x = x_lab, y = "Forecast horizon",
                    title = "Shelf life vs. assumed measurement noise")
  } else if (show == "env_share") {
    p <- ggplot2::ggplot(df, ggplot2::aes(x = .data[["x"]],
                                          y = .data[["env_share"]])) +
      ggplot2::geom_line(colour = "#4DAC26") +
      ggplot2::geom_point(colour = "#4DAC26", size = 3) +
      ggplot2::scale_y_continuous(labels = scales_percent_safe,
                                  limits = c(0, NA)) +
      ggplot2::labs(x = x_lab,
                    y = "Environmental share of total variance",
                    title = "How much of the forecast variance is environmental?")
  } else {
    thr <- attr(sens, "threshold")
    p <- ggplot2::ggplot(df, ggplot2::aes(x = .data[["x"]],
                                          y = .data[["ratio_mean"]])) +
      ggplot2::geom_line(colour = "#B2182B") +
      ggplot2::geom_point(colour = "#B2182B", size = 3) +
      ggplot2::geom_hline(yintercept = if (is.null(thr)) 1 else thr,
                          linetype = "dashed") +
      ggplot2::labs(x = x_lab,
                    y = "Mean CI width / plausible range",
                    title = "Forecast informativeness vs. noise")
  }

  p + et_theme()
}

# Permissive wrapper for scales::percent — fall back to identity if the
# scales package isn't installed so the plot still renders.
scales_percent_safe <- function(x) {
  if (requireNamespace("scales", quietly = TRUE)) scales::percent(x)
  else formatC(100 * x, format = "f", digits = 0, flag = "#")
}

# ******************************************************************************
# Internal helpers
# ______________________________________________________________________________

# Pivot variance decomposition to long format for ggplot stacking
.pivot_decomp <- function(decomp, cols) {
  rows <- lapply(cols, function(col) {
    label <- switch(col,
      param_var    = "Parameter",
      env_var      = "Environmental",
      residual_var = "Residual",
      temporal_var = "Temporal",
      col
    )
    df <- decomp[, c(setdiff(colnames(decomp), cols), col), drop = FALSE]
    df$component <- label
    df$value     <- decomp[[col]]
    df[, c(setdiff(colnames(decomp), cols), "component", "value")]
  })
  do.call(rbind, rows)
}

# tests/testthat/test-plot.R
# Smoke tests for et_plot_* functions — verify they return ggplot objects
# and do not error with mock data, without requiring a real brms fit.

library(ggplot2)

# ── Helper: minimal mock decomposition data.frame ───────────────────────────

.mock_decomp <- function(n_obs = 5) {
  data.frame(
    obs_id       = seq_len(n_obs),
    total_var    = runif(n_obs, 0.3, 0.5),
    param_var    = runif(n_obs, 0.05, 0.15),
    env_var      = runif(n_obs, 0.02, 0.08),
    residual_var = rep(0.22, n_obs)
  )
}

.mock_shelf_life <- function(n_obs = 5) {
  sl <- data.frame(
    obs_id          = seq_len(n_obs),
    time            = 2020 + seq_len(n_obs) - 1L,
    ci_width        = runif(n_obs, 0.8, 1.4),
    plausible_range = 3.5,
    ratio           = runif(n_obs, 0.2, 0.4),
    informative     = TRUE,
    stringsAsFactors = FALSE
  )
  attr(sl, "threshold") <- 1.0
  attr(sl, "horizon")   <- list(
    value            = NA_real_,
    type             = "lower_bound",
    last_informative = 2024,
    description      = "All periods informative."
  )
  structure(sl, class = c("et_shelf_life", "data.frame"))
}

.mock_calibration <- function() {
  data.frame(
    ci_level           = c(0.50, 0.80, 0.90, 0.95),
    nominal            = c(0.50, 0.80, 0.90, 0.95),
    observed_coverage  = c(0.60, 0.80, 1.00, 1.00),
    n_obs              = 5L,
    calibration_error  = c(0.10, 0.00, 0.10, 0.05),
    stringsAsFactors   = FALSE
  )
}

# ── et_plot_decomposition ────────────────────────────────────────────────────

test_that("et_plot_decomposition returns a ggplot object (proportional)", {
  p <- et_plot_decomposition(.mock_decomp(), proportional = TRUE)
  expect_s3_class(p, "ggplot")
})

test_that("et_plot_decomposition returns a ggplot object (absolute)", {
  p <- et_plot_decomposition(.mock_decomp(), proportional = FALSE)
  expect_s3_class(p, "ggplot")
})

test_that("et_plot_decomposition handles grouped decomp with group_col", {
  d <- rbind(
    cbind(group = "A", .mock_decomp(3)),
    cbind(group = "B", .mock_decomp(3))
  )
  p <- et_plot_decomposition(d, group_col = "group")
  expect_s3_class(p, "ggplot")
})

# ── et_plot_shelf_life ───────────────────────────────────────────────────────

test_that("et_plot_shelf_life returns a ggplot object", {
  p <- et_plot_shelf_life(.mock_shelf_life())
  expect_s3_class(p, "ggplot")
})

test_that("et_plot_shelf_life with show_ratio = FALSE works", {
  p <- et_plot_shelf_life(.mock_shelf_life(), show_ratio = FALSE)
  expect_s3_class(p, "ggplot")
})

test_that("et_plot_shelf_life with grouped et_shelf_life adds colour", {
  sl_g <- cbind(group = rep(c("A","B"), each = 3),
                .mock_shelf_life(6))
  attr(sl_g, "threshold") <- 1.0
  class(sl_g) <- c("et_shelf_life", "data.frame")
  p <- et_plot_shelf_life(sl_g)
  expect_s3_class(p, "ggplot")
})

# ── et_plot_calibration ──────────────────────────────────────────────────────

test_that("et_plot_calibration returns a ggplot object (single group)", {
  p <- et_plot_calibration(.mock_calibration())
  expect_s3_class(p, "ggplot")
})

test_that("et_plot_calibration returns a ggplot object (grouped)", {
  cal <- rbind(
    cbind(group = "A", .mock_calibration()),
    cbind(group = "B", .mock_calibration())
  )
  p <- et_plot_calibration(cal)
  expect_s3_class(p, "ggplot")
})

# R/decompose.R — decompose_uncertainty(): variance decomposition
# Reports a three-way split (parameter / environmental / residual) for iid
# models, and a four-way split adding a temporal-autocorrelation component
# when the model formula contains an ar() / ma() / arma() / cosy() / unstr()
# / sar() / car() term.

#' Extract or recompute uncertainty decomposition
#'
#' Returns a \code{data.frame} with the uncertainty decomposition stored
#' inside an \code{et_prediction} object:
#' \describe{
#'   \item{param_var}{Variance of the posterior linear predictor — captures
#'     uncertainty in fitted regression coefficients.}
#'   \item{env_var}{Additional variance arising from measurement or
#'     prediction uncertainty in the predictor values (estimated via
#'     perturbation in \code{\link{et_predict}}).  Zero when
#'     \code{env_noise = NULL}.}
#'   \item{residual_var}{Posterior mean of \eqn{\sigma^2} (or its
#'     family-specific analogue) — biological process noise, unmeasured
#'     drivers, and drift. For autocorrelation models this is the
#'     \emph{innovation} variance, not the stationary marginal variance;
#'     the autocorrelated accumulation is reported separately in
#'     \code{temporal_var}.}
#'   \item{temporal_var}{(Only present when the model formula contains an
#'     autocorrelation term such as \code{ar()}, \code{ma()},
#'     \code{arma()}, \code{cosy()}, \code{unstr()}, \code{sar()}, or
#'     \code{car()}.) Variance attributable to residual temporal or
#'     spatial dependence beyond the iid \code{param + residual} sum,
#'     computed as
#'     \code{pmax(0, total_var - (param_var + residual_var))}.
#'     \code{env_var} is deliberately excluded from this gap because it
#'     is an additive perturbation-based augmentation measured outside
#'     of \code{posterior_predict}.}
#'   \item{total_var}{Variance of the full posterior predictive draws,
#'     including any autocorrelation structure modelled by brms.}
#' }
#'
#' All variance components are guaranteed non-negative. When
#' \code{temporal_var} is present, \code{param_var + residual_var +
#' temporal_var} reconstructs \code{total_var} (modulo Monte Carlo error);
#' when it is absent, \code{param_var + residual_var} does.
#' \code{env_var} is always additive on top, representing the extra
#' variance that would be contributed by perturbing predictors with
#' \code{env_noise}.
#'
#' @param predictions An \code{et_prediction} object from
#'   \code{\link{et_predict}}, or an \code{et_prediction_list} (grouped).
#' @param ... Unused.
#' @return A \code{data.frame} with columns
#'   \code{obs_id, param_var, env_var, residual_var, total_var}
#'   (plus \code{temporal_var} for autocorrelation models, and a leading
#'   \code{group} column for grouped predictions).
#' @examples
#' \donttest{
#' set.seed(1)
#' df  <- data.frame(y = rnorm(20), x1 = rnorm(20))
#' fit <- et_fit(y ~ x1, data = df,
#'               chains = 1, iter = 500, warmup = 250,
#'               cores = 1, refresh = 0)
#' new_df <- data.frame(x1 = rnorm(5))
#' pred   <- et_predict(fit, newdata = new_df,
#'                      env_noise = list(x1 = 0.2),
#'                      n_draws = 200, n_perturb = 50)
#' decomp <- decompose_uncertainty(pred)
#' head(decomp)
#' }
#' @export
decompose_uncertainty <- function(predictions, ...) {
  UseMethod("decompose_uncertainty")
}

#' @export
decompose_uncertainty.et_prediction <- function(predictions, ...) {
  predictions$decomposition
}

#' @export
decompose_uncertainty.et_prediction_list <- function(predictions, ...) {
  parts <- lapply(names(predictions$predictions), function(g) {
    pred <- predictions$predictions[[g]]
    if (is.null(pred)) return(NULL)
    d <- pred$decomposition
    cbind(data.frame(group = g, stringsAsFactors = FALSE), d)
  })
  do.call(rbind, Filter(Negate(is.null), parts))
}

#' @export
decompose_uncertainty.default <- function(predictions, ...) {
  stop("decompose_uncertainty() expects an et_prediction or et_prediction_list object.")
}

---
title: "A one-predictor worked example"
subtitle: "Walking step by step through the ErrorTracer pipeline"
author: "Luis Javier Madrigal-Roca"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 3
    number_sections: true
vignette: >
  %\VignetteIndexEntry{A one-predictor worked example}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = FALSE,   # Stan/brms chunks are too slow for automated builds
  fig.width = 7,
  fig.height = 4.5
)
```

# What this vignette is for

The methods paper for `ErrorTracer` contains a one-page **Box 1**
that walks through every stage of the pipeline using a single
predictor and a single response. The box gives the *intuition* in a
compressed form. This vignette is the long-form companion: same
example, same numbers, but with the code, the math, and the diagnostic
plots laid out one stage at a time so a reader new to Bayesian
modelling can follow along, run it, and modify it.

The whole example fits one model: phenology shift (days) as a
function of a single standardised temperature anomaly. We *know* the
data-generating coefficients because we simulate the data ourselves
— so you can see how well each stage recovers what was put in.

```{r}
library(ErrorTracer)
library(ggplot2)
library(patchwork)
```

# Setup: simulate one predictor, one response — over years

Shelf life is a *temporal* concept ("for how many years is my forecast
still useful?"), so the toy data is generated as a year-indexed time
series. The temperature predictor `x` drifts upward at a fixed warming
rate, the response `y` follows it through a linear coefficient
$\beta$, and the training window is the past while the forecast
window is the future.

```{r}
set.seed(20260523L)
n_train       <- 12L      # 12 training years
true_beta     <- 4.0      # days of phenology shift per SD of temperature
true_sigma    <- 1.5      # residual SD (days)
true_alpha    <- 0.0      # intercept
warming_slope <- 0.15     # SD of standardised temperature per year
year0         <- 2000L    # year at which the trend is referenced to zero

years_train <- year0 + seq_len(n_train) - 1L  # 2000..2011
train <- data.frame(
  year = years_train,
  x    = warming_slope * (years_train - year0) +
         rnorm(n_train, 0, 0.5)               # noisy observed temperature
)
train$y <- true_alpha + true_beta * train$x +
           rnorm(n_train, 0, true_sigma)
```

We have 12 noisy training observations along the warming trajectory.
The data-generating slope is $\beta = 4$ days per SD of temperature;
everything below should recover something close to this.

# Stage 1: turn a quick frequentist fit into a prior

`extract_priors()` is the bridge from a "fast, regularized" world to
a "slow, Bayesian" world. The recipe is the same for `lm`, `glm`,
`cv.glmnet`, and `ranger`: take the point estimate (or coefficient)
as the prior mean and scale the prior SD by the standard error
(or by the regularised coefficient magnitude, for `glmnet`).

```{r}
lm_fit  <- lm(y ~ x, data = train)
summary(lm_fit)$coefficients
priors  <- extract_priors(lm_fit, multiplier = 2.0, min_sd = 0.1)
priors
```

Concretely, with $m = 2$ (the `multiplier` default) and
$\sigma_{\min} = 0.1$, the prior on $\beta$ is

$$
\beta \sim \mathcal{N}\!\left(\hat\beta_{\text{lm}},
                              \bigl[\max(2 \cdot \mathrm{SE},\,
                                         \sigma_{\min})\bigr]^2\right).
$$

For our simulated data, $\hat\beta_{\text{lm}} \approx 4.02$ with
$\mathrm{SE} \approx 0.39$, so the prior is roughly
$\mathcal{N}(4.02, 0.79^2)$.

# Stage 2: Bayesian refit with the informed prior

```{r}
bfit <- et_fit(
  formula = y ~ x,
  data    = train,
  priors  = priors,
  chains  = 4L, iter = 2000L, warmup = 1000L,
  cores   = 4L, seed = 1L, silent = 2L, refresh = 0
)
print(bfit)
```

Under the hood `et_fit()` calls `brms::brm()` with the prior we
supplied for $\beta$ plus weakly-informative defaults for the
intercept and the residual scale. The output is a full posterior:

```{r}
beta_draws <- as.matrix(bfit$fit, variable = "b_x")[, 1]
c(mean = mean(beta_draws), sd = sd(beta_draws))
```

For our seed, the posterior collapses to $\beta \approx 4.01$ with SD
$\approx 0.38$. The data didn't drag the prior very far because the
prior was *already* centred on the lm estimate — but the posterior
SD is sharper than the prior SD, which is the regular Bayesian
shrinkage we wanted.

A picture is worth a thousand draws. The package ships a helper:

```{r}
et_plot_prior_posterior(bfit)
```

# Stage 3: posterior prediction across the forecast horizon

The forecast window is years 2012–2045. At each future year, the
projected temperature is the trend-line value
$x(t) = \mathrm{warming\_slope}\cdot(t - 2000)$. We pass
`env_noise = list(x = 0.3)` to tell `et_predict()` that the projected
temperature is itself uncertain by $\pm 0.3$ SD (a reasonable error
for a downscaled climate projection).

```{r}
year_min     <- year0
year_max     <- year0 + 45L
years_grid   <- seq(year_min, year_max, by = 1L)
forecast_df  <- data.frame(
  year = years_grid,
  x    = warming_slope * (years_grid - year0)
)

pred <- et_predict(
  model             = bfit,
  newdata           = forecast_df,
  env_noise         = list(x = 0.3),
  n_draws           = 2000L,
  n_perturb         = 500L,
  ci_levels         = 0.90,
  include_env_in_ci = TRUE
)
```

Setting `include_env_in_ci = TRUE` folds the environmental noise into
the reported credible intervals. With it `FALSE` (the default), the CI
captures parameter + residual uncertainty only; environmental
variance is still reported separately in `decompose_uncertainty()`,
but the user-facing CI is "what would the response be if we knew the
future temperature exactly". The choice depends on whether you're
forecasting *under* projection uncertainty or *given* a known
driver.

# Stage 4: decompose the variance over time

`decompose_uncertainty()` returns one row per forecast year with the
three (or four, if you have autocorrelation) variance components:

```{r}
decomp <- decompose_uncertainty(pred)
decomp$year <- forecast_df$year
head(decomp)
```

Three things to notice as the forecast horizon extends into the
future (year increases, so $x$ on the warming trajectory increases
too):

1. `param_var` grows roughly as $\hat\sigma_\beta^2 \cdot x(t)^2$.
   That's the *extrapolation cost* of being uncertain about $\beta$:
   the further past the training window you forecast, the more it
   matters.
2. `env_var` is roughly constant at $\hat\beta^2 \cdot \sigma_x^2$.
   It scales with how steep the dose–response is *and* how poorly
   we know the future dose, but it does not grow with horizon.
3. `residual_var` is the biological-noise floor. It does not change
   with the forecast year because it's a property of the response,
   not of when you forecast.

A quick numerical look at a representative mid-forecast year:

```{r}
mid <- which(decomp$year == 2020L)
decomp[mid, ]
```

# Stage 5: when does the forecast become uninformative?

A forecast is *informative* when the predictive CI is narrower than
the plausible range of the response. The plausible range here is "any
phenology shift between $-4$ and $+4$ days" — anything wider than
that and we are basically saying "we have no idea".

```{r}
sl <- shelf_life(
  predictions    = pred,
  response_scale = c(-4, 4),
  ci_level       = 0.90,
  threshold      = 1.0,
  time_col       = "year"
)
attr(sl, "horizon")
```

`shelf_life()` walks along the forecast time axis and reports the
first *year* at which the 90 % CI width first exceeds the response
range. For this example it lands at $t^{\star} \approx 2027$ — i.e.
the model still produces a useful forecast for roughly 16 years past
the end of training; beyond that, the credible interval is wider
than the plausible response and should not be over-interpreted.
That answer — a calendar year, not a predictor value — is the
quantity ecologists and conservation practitioners actually need.

# The whole-pipeline figure

The four diagnostic panels — prior vs posterior, the nested CI fan,
the stacked variance components, and the shelf-life curve — are
produced together by the script that generates Box 1 of the paper.
The exact same code is what appears in the box's figure; see the
package source under `src/box1_worked_example.R` in the
proof-of-concept repository (or run the chunk below in your own
session):

```{r, eval = FALSE}
# Reproduce the four-panel Box 1 figure used in the methods paper.
source(system.file("extdata", "box1_worked_example.R",
                   package = "ErrorTracer"))
run_box1_worked_example(figures_dir = tempdir(),
                        tex_path    = tempfile(fileext = ".tex"))
```

# What this example does *not* show

Two limits of this minimal walk-through deserve flagging:

- **Small $n$, single predictor.** With 12 observations and one
  driver, the Bayesian fit is doing real work — informative priors
  matter. Real workflows usually have larger $n$, multiple correlated
  predictors, and a regularised feature-selection step (elastic net)
  upstream of `extract_priors()`. The fundamental decomposition is
  the same.
- **No grouping.** `et_fit(..., grouping = "g")` runs the same
  pipeline independently per level of a grouping factor (per SNP
  cluster, per population, per site). The toy example fits a single
  global model; the multi-group form returns a named list of
  `et_model` objects, and `decompose_uncertainty()` / `shelf_life()`
  preserve the grouping.

# Session info

```{r, eval = TRUE}
sessionInfo()
```

---
title: "Getting Started with ErrorTracer"
subtitle: "Bayesian error propagation and forecast uncertainty decomposition"
author: "Luis Javier Madrigal-Roca, John Kelly"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 3
    number_sections: true
vignette: >
  %\VignetteIndexEntry{Getting Started with ErrorTracer}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = FALSE,  # Stan/brms chunks are too slow for automated builds
  fig.width = 7,
  fig.height = 4.5
)
```

> **How to use this vignette.**
> All interactive code chunks have `eval = FALSE` so the vignette can build
> without running Stan.  Copy and run each chunk in sequence in your own R
> session.
> Before the first use, generate the bundled dataset by running once from the
> package root:
>
> ```r
> source("data-raw/generate_et_sim.R")
> ```
>
> For real analyses use `chains = 4` and `iter` ≥ 2000.  The demo below uses
> 2 chains × 1 000 iterations to keep runtime short.

---

# Introduction

**ErrorTracer** addresses a gap in the ecological and genomic forecasting
toolkit: existing Bayesian packages (`brms`, `rstanarm`) do not decompose
forecast uncertainty into its contributing sources, and no R package quantifies
the *forecast shelf life* — the time horizon at which a credible interval
becomes too wide to be informative.

The package provides a seven-step pipeline:

| Step | Function | What it does |
|---|---|---|
| 1 | `extract_priors()` | Convert a regularized model to `brms` priors |
| 2 | `et_fit()` | Fit the informed Bayesian model |
| 3 | `et_diagnose()` | Convergence, ESS, LOO-CV |
| 4 | `et_predict()` | Posterior prediction + environmental noise propagation |
| 5 | `decompose_uncertainty()` | Variance decomposition (param / env / residual, plus temporal for autocor models) |
| 6 | `shelf_life()` | Forecast horizon / shelf-life analysis |
| 7 | `et_calibrate()` | Observed vs. nominal coverage |

Visualisation functions (`et_plot_*`) return `ggplot2` objects at every step.

## The simulated dataset

We work with `et_sim`, a list bundled with the package.  It represents a
hypothetical mountain plant site where allele-frequency change on the
**arcsin-sqrt scale** (`z_diff`) in two SNP clusters (A and B) is linked to
three growing-season climate predictors.

> **Note on the arcsin-sqrt transformation.**
> Raw allele frequencies $f \in [0,\,1]$ are transformed via
> $z = \arcsin(\sqrt{f})$.  The result is unbounded (theoretically
> $[0,\,\pi/2] \approx [0,\,1.57]$ for a single chromosome, but changes
> $\Delta z$ are unbounded in principle).  **Do not use** `plausible_range =
> c(-1, 1)` for this response — instead derive the range from the training
> data or use a biologically motivated interval.

| Predictor | Description | Units (raw) |
|---|---|---|
| `Tmean` | Mean growing-season temperature | °C (standardised) |
| `PPT`   | Total growing-season precipitation | mm (standardised) |
| `SWE`   | Peak snow water equivalent | mm (standardised) |

The dataset has a training period (1995–2014, 20 years) and a forecast period
(2015–2019, 5 years).  The forecast climate is slightly warmer than the
training period on average (Tmean ~1 SD above training mean), representing a
mild warming trend.  Because the data are simulated, we know the **true
regression coefficients** — stored in `et_sim$true_params` — which lets us
validate parameter recovery and calibration.

**True data-generating model:**

$$
z\_diff_t = \alpha + \beta_1 \cdot Tmean_t + \beta_2 \cdot PPT_t +
             \beta_3 \cdot SWE_t + \varepsilon_t, \quad
\varepsilon_t \sim \mathcal{N}(0,\, \sigma^2)
$$

| Parameter | Cluster A | Cluster B |
|---|---|---|
| $\alpha$ (intercept) | 0.00 | 0.10 |
| $\beta_1$ (Tmean) | **+0.50** | **+0.30** |
| $\beta_2$ (PPT)   | **−0.30** | **−0.20** |
| $\beta_3$ (SWE)   | **+0.20** | **−0.10** |
| $\sigma$ | 0.40 | 0.50 |

Both clusters have a positive response to warming (`Tmean`) and a negative
response to precipitation (`PPT`).  They differ in the direction of their
response to snowpack (`SWE`) and in residual variability — cluster B is noisier
(σ = 0.50 vs. 0.40), which will be reflected in its wider forecast intervals
and shelf-life ratio closer to the uninformative threshold.

---

# Setup

```{r libraries}
library(ErrorTracer)
library(glmnet)      # for elastic net (Suggests)
library(ggplot2)

# Load the bundled simulated dataset
data(et_sim)
```

```{r demo-settings}
# Use small MCMC settings for this tutorial.
# In a real analysis: CHAINS = 4, ITER = 2000 (or more).
CHAINS <- 2L
ITER   <- 1000L
SEED   <- 42L
```

---

# Exploring the dataset

```{r data-overview}
str(et_sim, max.level = 2)
# List of 7
#  $ train         :'data.frame': 40 obs. of 6 variables:
#  $ forecast      :'data.frame': 10 obs. of 5 variables:
#  $ validation    :'data.frame': 10 obs. of 3 variables:
#  $ true_params   :List of 2
#  $ env_noise     :List of 3
#  $ standardization:List of 3
#  $ description   : chr ...
```

```{r head-train}
head(et_sim$train)
#   year cluster_id      Tmean        PPT        SWE     z_diff
# 1 1995          A -0.6775963  0.9119435 -0.1386249 -0.4917161
# 2 1996          A  0.4866542 -1.7284276 -1.4391959  0.4088745
# 3 1997          A -0.4935765 -0.3945083 -0.7267905  0.2352018
# 4 1998          A -0.2397691 -0.9805886 -0.8206722  0.3194194
# 5 1999          A -0.5193449  1.1304845  1.6453631 -0.2554886
# 6 2000          A -1.5333598  0.6437605  0.6252043 -0.5864645
```

```{r true-params}
# True coefficients — known because the data are simulated
et_sim$true_params
# $A
#   intercept     Tmean       PPT       SWE     sigma
#        0.00      0.50     -0.30      0.20      0.40
#
# $B
#   intercept     Tmean       PPT       SWE     sigma
#        0.10      0.30     -0.20     -0.10      0.50
```

```{r forecast-predictors}
# Forecast predictors are slightly above training range in Tmean (warming trend)
et_sim$forecast[et_sim$forecast$cluster_id == "A", ]
#   year cluster_id     Tmean       PPT        SWE
# 1 2015          A 0.6918714  0.291...  0.102...
# 2 2016          A 0.7012... ...
# ...
# Forecast Tmean is ~0.7-1.4 SDs above training mean — mild extrapolation.
```

## Visualise the training data

```{r plot-training-data}
ggplot(et_sim$train, aes(x = year, y = z_diff, colour = cluster_id)) +
  geom_line() +
  geom_point(size = 2) +
  labs(
    title   = "Simulated allele-frequency change (training period)",
    x       = "Year",
    y       = expression(italic(z)[diff]),
    colour  = "Cluster"
  ) +
  et_theme()
```

```{r plot-predictors}
# Climate predictors over the full period — warming trend visible in Tmean
all_years <- rbind(
  transform(et_sim$train[et_sim$train$cluster_id == "A",
                          c("year", "Tmean", "PPT", "SWE")],
            period = "Training"),
  transform(et_sim$forecast[et_sim$forecast$cluster_id == "A",
                             c("year", "Tmean", "PPT", "SWE")],
            period = "Forecast")
)

all_long <- reshape(all_years, direction = "long",
                    varying = c("Tmean", "PPT", "SWE"),
                    v.names = "value", timevar = "predictor",
                    times   = c("Tmean", "PPT", "SWE"))

ggplot(all_long, aes(x = year, y = value,
                     colour = period, group = period)) +
  geom_line() +
  geom_point(size = 1.8) +
  facet_wrap(~ predictor, ncol = 3) +
  labs(
    title  = "Climate predictors: training and forecast periods",
    x      = "Year", y = "Standardised value", colour = "Period"
  ) +
  et_theme()
```

---

# Single-cluster workflow: Cluster A

We walk through the complete pipeline for cluster A, then repeat for both
clusters simultaneously using the `grouping` argument.

## Step 1 — Elastic net for prior extraction

We fit a cross-validated elastic net (`alpha = 0.5`) on the training data.
The elastic net simultaneously performs variable selection under regularisation
and provides coefficient estimates that become prior means in the subsequent
Bayesian model.

```{r enet-fit}
train_A <- et_sim$train[et_sim$train$cluster_id == "A", ]

x_train <- as.matrix(train_A[, c("Tmean", "PPT", "SWE")])
y_train <- train_A$z_diff

set.seed(42)
cv_fit_A <- glmnet::cv.glmnet(
  x     = x_train,
  y     = y_train,
  alpha = 0.5          # 0 = ridge, 1 = lasso, 0.5 = elastic net
)

# Coefficient estimates at lambda.min
coef(cv_fit_A, s = "lambda.min")
# 4 x 1 sparse Matrix of class "dgCMatrix"
#             lambda.min
# (Intercept)  0.0430
# Tmean        0.6658
# PPT         -0.4591
# SWE          0.3775
```

> **Note on small n:** With 20 observations and 10-fold CV, R will print a
> warning about fewer than 3 observations per fold.  This is expected and
> harmless — the elastic net still regularises correctly; it just uses the
> default grouped cross-validation.

The elastic net correctly recovers the signs of all three predictors.  The
magnitudes are somewhat inflated relative to the true values (Tmean = 0.50,
PPT = −0.30, SWE = 0.20) because with n = 20 the regularisation cannot
fully separate the signal from sampling noise.  This shrinkage toward the
true values will be completed by the Bayesian posterior.

```{r enet-plot}
plot(cv_fit_A)
title("Cross-validated elastic net — Cluster A", line = 2.5)
```

## Step 2 — Extract informed priors

`extract_priors()` converts the elastic net coefficients into a `brmsprior`
object.  The prior SD for each coefficient is set to
`multiplier × |coef|` (floored at `min_sd`).

```{r extract-priors}
prior_spec_A <- extract_priors(
  model      = cv_fit_A,
  lambda     = "lambda.min",
  multiplier = 2.0,   # prior SD = 2 × |elastic-net coef|
  min_sd     = 0.10   # floor: avoids spike priors on near-zero coefficients
)

print(prior_spec_A)
# ErrorTracer prior specification
#   Method      : glmnet
#   Predictors  : 3
#   Multiplier  : 2
#   Min SD      : 0.1
#   Coefficients:
#     Tmean                 mean =   0.6658  sd =   1.3316
#     PPT                   mean =  -0.4591  sd =   0.9182
#     SWE                   mean =   0.3775  sd =   0.7550
```

The large prior SDs (1.3 for Tmean, 0.9 for PPT, 0.8 for SWE) reflect the
`multiplier = 2` setting — priors are wide enough that the likelihood can
readily update them, while still encoding the directional information from the
elastic net.

> **Design note:** The intercept is excluded from the informed prior.  In this
> simulation the predictors are standardised to zero mean in the training
> period, so the intercept carries no predictive information.  The intercept
> receives a default `Normal(0, 2.5)` prior.

## Step 3 — Fit the Bayesian model

`et_fit()` wraps `brms::brm()`, attaches the prior specification, and returns
an `et_model` object.

```{r et-fit}
fit_A <- et_fit(
  formula = z_diff ~ Tmean + PPT + SWE,
  data    = train_A,
  priors  = prior_spec_A,
  chains  = CHAINS,
  iter    = ITER,
  warmup  = ITER %/% 2L,
  cores   = CHAINS,
  seed    = SEED
)

print(fit_A)
# ErrorTracer model (et_model)
#   Formula : z_diff ~ Tmean + PPT + SWE
#   n obs   : 20
#   Chains  : 2  Iter: 1000  Warmup: 500
#   Priors  : informed (glmnet, 3 predictors)
#   Rhat max: 1.004
```

```{r et-summary}
summary(fit_A)
# === ErrorTracer model summary ===
#
# --- Fixed effects ---
#              Estimate  Est.Error     Q2.5    Q97.5
# Intercept    0.041      0.104    -0.162    0.246
# Tmean        0.666      0.113     0.438    0.888
# PPT         -0.487      0.200    -0.909   -0.123
# SWE          0.406      0.206     0.023    0.827
```

The posterior means are shrunk slightly relative to the elastic-net estimates,
particularly for PPT and SWE — the likelihood is pulling the estimates toward
the true values (Tmean = 0.50, PPT = −0.30, SWE = 0.20).  All 95% credible
intervals exclude zero for Tmean and PPT; SWE is positive but its lower bound
just crosses zero, consistent with the true value being 0.20 with σ = 0.40
and only 20 training observations.

## Step 4 — Diagnose convergence and model fit

```{r et-diagnose}
diag_A <- et_diagnose(fit_A, loo = TRUE)

# Convergence
diag_A$convergence
# $rhat_max
# [1] 1.004         # all chains well below 1.05 threshold
#
# $rhat_all_ok
# [1] TRUE
#
# $neff_min
# [1] 0.275         # lowest effective-sample-size ratio
#
# $neff_all_ok
# [1] TRUE          # all Neff ratios > 0.10
#
# $n_divergences
# [1] 0             # no divergent transitions
```

```{r loo-results}
diag_A$loo[c("elpd_loo", "p_loo", "looic", "n_bad_pareto_k")]
# $elpd_loo
# [1] -14.55
#
# $p_loo
# [1] 4.00          # effective parameters ≈ 4 (intercept, 3 betas); sigma
#                   # is implicit — consistent with a 4-parameter model
# $looic
# [1] 29.10
#
# $n_bad_pareto_k
# [1] 0             # no influential observations (Pareto k > 0.7)
```

All Rhat < 1.05, zero divergences, and no Pareto-k warnings — the sampler has
converged and the posterior is not dominated by any single observation.

## Step 5 — Predict with environmental noise propagation

`et_predict()` generates posterior predictive draws and propagates uncertainty
about the predictor values themselves via the `env_noise` argument.  Each
perturbation draw adds Gaussian noise to the predictor values before computing
the linear predictor, capturing sensitivity to measurement or forecast error in
the climate inputs.

`env_noise` accepts three forms:
- A **named list of scalars** — constant noise SD per predictor across all
  observations (suitable for short validation windows).
- A **named list of vectors** of length `nrow(newdata)` — **time-varying**
  noise, where each observation gets its own SD.  This is the recommended form
  for GCM projections, where ensemble spread grows with forecast horizon.
- A **single scalar** — applied as a fraction of each predictor's empirical SD.

```{r et-predict}
fcast_A <- et_sim$forecast[et_sim$forecast$cluster_id == "A", ]

pred_A <- et_predict(
  model     = fit_A,
  newdata   = fcast_A,
  env_noise = et_sim$env_noise,   # list(Tmean = 0.30, PPT = 0.20, SWE = 0.15)
  n_draws   = 1000L,
  ci_levels = c(0.50, 0.80, 0.90, 0.95),
  n_perturb = 300L                # draws for the perturbation step
)

print(pred_A)
# ErrorTracer prediction (et_prediction)
#   Observations : 5
#   Draws        : 1000
#   CI levels    : 0.5, 0.8, 0.9, 0.95
#   Mean var decomposition (across observations):
#     Parameter    : 0.0441
#     Environmental: 0.0579
#     Residual     : 0.2173
#     Total        : 0.2673
```

In this dataset the environmental uncertainty (0.058) is comparable in
magnitude to the parameter uncertainty (0.044) — a noteworthy finding.
With only 20 training observations, measurement noise in the climate inputs
contributes almost as much uncertainty as imprecision in the estimated
regression coefficients.  Residual process noise (0.217) remains dominant
because σ = 0.40 is large relative to the signal.

```{r ci-table}
# 90% credible intervals by forecast year
ci90 <- pred_A$credible_intervals[pred_A$credible_intervals$ci_level == 0.90, ]
ci90$year <- fcast_A$year
ci90[, c("year", "lower", "median", "upper", "width")]
#   year    lower  median   upper  width
# 1 2015   -0.507   0.381   1.213  1.720
# 2 2016    0.009   0.794   1.660  1.651
# 3 2017   -0.009   0.875   1.631  1.640
# 4 2018   -0.179   0.649   1.460  1.639
# 5 2019   -0.767   0.131   1.002  1.769
```

The medians are positive across all forecast years, consistent with the mild
warming trend pushing Tmean ~0.7–1.4 SDs above the training mean.  CI widths
are 1.64–1.77, somewhat narrower than the full plausible range of 2.

## Step 6 — Uncertainty decomposition

```{r decompose}
decomp_A <- decompose_uncertainty(pred_A)
decomp_A
#   obs_id total_var param_var  env_var residual_var
# 1      1    0.2627    0.0202   0.0541       0.2173
# 2      2    0.2667    0.0361   0.0588       0.2173
# 3      3    0.2573    0.0364   0.0541       0.2173
# 4      4    0.2529    0.0427   0.0653       0.2173
# 5      5    0.2972    0.0849   0.0574       0.2173
```

`residual_var` is constant across observations — it equals the posterior mean
of σ² and represents process noise that cannot be reduced by better predictors
or more training data.  `param_var` increases slightly in observations 4–5,
reflecting that those forecast years (2018–2019) are more distant from the
training period's centroid, where the model is less certain.

```{r plot-decomp}
et_plot_decomposition(decomp_A, proportional = TRUE) +
  labs(subtitle = "Cluster A, forecast 2015–2019")
```

The proportional view makes it clear that process noise (residual) accounts
for ~80% of total forecast variance, parameter uncertainty for ~12–20%, and
environmental measurement noise for ~17–25%.  To substantially improve
forecast precision for this system, the primary target should be the residual
variance — identifying additional drivers of allele-frequency change.

> **Autocorrelation note.** If you fit the model with a residual
> autocorrelation term (e.g.\ `et_fit(y ~ x + ar(time = t, p = 1), ...)`),
> `decompose_uncertainty()` adds a fourth column `temporal_var` defined as
> `pmax(0, total_var − (param_var + env_var + residual_var))`. It captures
> the autocorrelation-induced spread that `brms::posterior_predict()`
> accumulates iteratively beyond a single innovation, and `residual_var`
> is then interpretable as the *innovation* variance rather than the
> stationary marginal variance. `et_plot_decomposition()` shows the
> fourth segment automatically when the column is present.

## Step 7 — Forecast shelf life

The shelf life analysis asks: *at which forecast year does the 90% CI width
exceed the biologically plausible response range?*  Because `z_diff` is on
the arcsin-sqrt scale (unbounded), we derive the plausible range from the
training data rather than assuming a fixed ±1 bound.

```{r shelf-life}
# Derive plausible range from training data (arcsin-sqrt scale — unbounded)
pr_A <- range(train_A$z_diff)   # e.g. c(-1.50, 2.04) => width ~ 3.54

sl_A <- shelf_life(
  predictions              = pred_A,
  plausible_range          = pr_A,
  ci_level                 = 0.90,
  threshold                = 1.0,        # uninformative when CI width >= plausible range
  time_col                 = "year",
  min_slope_for_projection = 1e-4,       # positive slope required to project
  max_extrapolation_factor = 10          # cap: project <= last_year + 10 * window_width
)

print(sl_A)
# ErrorTracer shelf life analysis
#   Observations     : 5
#   Plausible range  : 3.54
#   Threshold        : 1
#   Informative      : 5 / 5
#   Mean CI/range    : 0.488
#   Max CI/range     : 0.499
#   Shelf life       : > 2019 (lower bound — all periods informative, no trend)
#   Detail           : All 5 forecast periods informative. Linear trend
#                      (slope = 0.00241 per time unit) projects threshold
#                      crossing at ~2234.3, but this exceeds the extrapolation
#                      cap (2059). Shelf life > 2019.
```

```{r shelf-life-table}
as.data.frame(sl_A)
#   obs_id  time ci_width plausible_range  ratio informative
# 1      1  2015    1.720            3.54  0.486        TRUE
# 2      2  2016    1.651            3.54  0.466        TRUE
# 3      3  2017    1.640            3.54  0.463        TRUE
# 4      4  2018    1.639            3.54  0.463        TRUE
# 5      5  2019    1.769            3.54  0.499        TRUE
```

All five forecast years are strongly informative (ratio ≈ 0.47–0.50, well
below the threshold of 1.0).  The near-flat ratio trend means the forecast
stays reliable throughout the 2015–2019 window.  A faint positive slope
(0.00241 per year) technically implies a linear crossing at ~2234 — but since
that is far beyond the 10× extrapolation cap (2059), the result is reported
as a **lower bound**: the forecast remains informative for at least the entire
observed window, and likely far beyond.

> **Interpreting the three horizon modes:**
>
> | Mode | When it applies | What is returned |
> |---|---|---|
> | `observed` | Threshold crossed within the prediction window | First crossing time |
> | `projected` | All periods informative; positive trend within cap | Extrapolated crossing time |
> | `lower_bound` | All informative; no trend, or projection beyond cap | `> last observed time` |
>
> Increase `max_extrapolation_factor` (or set it to `Inf`) to allow longer
> projections if the biological context supports it.

```{r plot-shelf-life}
et_plot_shelf_life(sl_A, show_ratio = TRUE) +
  labs(subtitle = "Cluster A — 90% CI width / training range (arcsin-sqrt scale)")
```

## Step 8 — Calibration assessment

We compare the posterior predictive intervals against the held-out true values
in `et_sim$validation`.

```{r calibrate}
valid_A <- et_sim$validation[et_sim$validation$cluster_id == "A", ]

cal_A <- et_calibrate(
  predictions  = pred_A,
  observed     = valid_A,
  response_col = "z_diff",
  ci_levels    = c(0.50, 0.80, 0.90, 0.95)
)

print(cal_A)
#   ci_level nominal observed_coverage n_obs calibration_error
# 1     0.50    0.50              0.60     5              0.10
# 2     0.80    0.80              0.80     5              0.00
# 3     0.90    0.90              1.00     5              0.10
# 4     0.95    0.95              1.00     5              0.05
```

The 80% CI achieves exactly nominal coverage.  The 90% and 95% CIs are
slightly over-conservative (all 5 held-out values fall inside), which is
expected with only 5 validation observations and regularising priors.  Coverage
estimates based on n = 5 have high sampling variance; the important result is
that no CI is systematically under-covering.

```{r plot-calibration}
et_plot_calibration(cal_A) +
  labs(subtitle = "Cluster A — posterior predictive calibration")
```

---

# Visualisations

## Forecast fan chart

```{r plot-forecast}
et_plot_forecast(
  predictions  = pred_A,
  observed     = valid_A,
  response_col = "z_diff",
  time_col     = "year"
) +
  labs(subtitle = "Cluster A — shaded ribbons: 50%, 80%, 90%, 95% CI")
```

Nested credible-interval ribbons (darker = narrower interval) overlay the
posterior median.  Red points are the held-out validation values.  All five
observed values fall within the 95% CI.

## Prior versus posterior

```{r plot-prior-posterior}
et_plot_prior_posterior(fit_A, max_preds = 3L) +
  labs(subtitle = "Cluster A — how 20 observations update the elastic-net prior")
```

Each panel overlays the prior density (grey, wide) with the posterior density
(blue, narrower).  A tightly shifted posterior means the data are informative;
a posterior that closely mirrors the prior means that the likelihood has little
leverage on that coefficient.  `Tmean` shows the largest update, consistent
with its largest true effect size.

## Forest plot: Bayesian vs. elastic net

```{r plot-coefficients}
et_plot_coefficients(fit_A) +
  labs(subtitle = "Cluster A — Bayesian 95% CI (blue) vs. enet coefficient (red ×)")
```

Red crosses mark the elastic-net estimates used as prior means.  The Bayesian
posterior means (blue points) with their 95% CIs show how the likelihood
shifted the estimates.  For all three predictors the posterior is pulled
slightly toward zero relative to the elastic-net estimate, reflecting
regularisation by the wide but informative priors.

---

# Parameter recovery validation

A key advantage of simulated data is that we can check whether the 95%
posterior credible intervals contain the known true values.

```{r parameter-recovery}
post    <- brms::fixef(fit_A$fit)
true_A  <- et_sim$true_params$A

post_df <- data.frame(
  parameter  = c("Intercept", "Tmean", "PPT", "SWE"),
  true_value = c(true_A["intercept"], true_A["Tmean"],
                 true_A["PPT"],       true_A["SWE"]),
  post_mean  = post[, "Estimate"],
  lower_95   = post[, "Q2.5"],
  upper_95   = post[, "Q97.5"]
)
post_df$covered <- with(post_df,
                        true_value >= lower_95 & true_value <= upper_95)

post_df[, c("parameter", "true_value", "post_mean", "lower_95", "upper_95", "covered")]
#   parameter true_value post_mean lower_95 upper_95 covered
# 1 Intercept       0.00     0.041   -0.162    0.246    TRUE
# 2     Tmean       0.50     0.666    0.438    0.888    TRUE
# 3       PPT      -0.30    -0.487   -0.909   -0.123    TRUE
# 4       SWE       0.20     0.406    0.023    0.827    TRUE
```

All four true parameter values fall within their 95% posterior credible
intervals.  The posterior means for Tmean, PPT, and SWE are consistently
above the true values in absolute magnitude — this is typical over-estimation
bias with n = 20 and informative priors near the OLS estimate.  The bias is
small relative to the interval widths, and would shrink with more training
data.

---

# Multi-cluster grouped workflow

In practice you often need to fit the same model to many units (SNP clusters,
species, sites) simultaneously.  `et_fit()` handles this with the `grouping`
argument, returning an `et_model_list`.

## Fitting

```{r grouped-priors}
# Build one et_prior_spec per cluster from individual elastic nets
prior_list <- lapply(c("A", "B"), function(cl) {
  df <- et_sim$train[et_sim$train$cluster_id == cl, ]
  x  <- as.matrix(df[, c("Tmean", "PPT", "SWE")])
  set.seed(42)
  suppressWarnings(cv <- glmnet::cv.glmnet(x, df$z_diff, alpha = 0.5))
  extract_priors(cv, multiplier = 2.0, min_sd = 0.10)
})
names(prior_list) <- c("A", "B")
```

```{r grouped-fit}
# One brms model per cluster, returned as et_model_list
fit_all <- et_fit(
  formula  = z_diff ~ Tmean + PPT + SWE,
  data     = et_sim$train,
  priors   = prior_list,       # named list: one et_prior_spec per cluster
  grouping = "cluster_id",
  chains   = CHAINS,
  iter     = ITER,
  warmup   = ITER %/% 2L,
  cores    = CHAINS,
  seed     = SEED
)

print(fit_all)
# ErrorTracer grouped model list (et_model_list)
#   Grouping : cluster_id
#   Formula  : z_diff ~ Tmean + PPT + SWE
#   Groups   : 2
#   Fitted   : 2 / 2

summary(fit_all)
# --- Per-group Rhat max ---
#   A                     Rhat max = 1.004
#   B                     Rhat max = 1.007
```

## Diagnostics

```{r grouped-diagnose}
diag_all <- et_diagnose(fit_all, loo = TRUE)
diag_all$summary
#   group rhat_ok neff_ok n_divergences elpd_loo n_bad_pareto_k
# 1     A    TRUE    TRUE             0   -14.55              0
# 2     B    TRUE    TRUE             0   -17.15              1
```

Cluster B has one observation with Pareto-k > 0.7, which is a mild flag worth
investigating (it suggests that observation has an outsized influence on the
posterior).  With only 20 training points this is not unusual.

## Prediction and decomposition

```{r grouped-predict}
pred_all <- et_predict(
  model     = fit_all,
  newdata   = et_sim$forecast,
  env_noise = et_sim$env_noise,
  n_draws   = 1000L,
  ci_levels = c(0.50, 0.80, 0.90, 0.95),
  n_perturb = 300L
)

print(pred_all)
# ErrorTracer grouped predictions (et_prediction_list)
#   Grouping : cluster_id
#   Groups   : 2 / 2 predicted
```

```{r grouped-decompose}
decomp_all <- decompose_uncertainty(pred_all)
head(decomp_all)
#   group obs_id total_var param_var  env_var residual_var
# 1     A      1    0.2627    0.0202   0.0541       0.2173
# 2     A      2    0.2667    0.0361   0.0588       0.2173
# ...
# 6     B      1    0.3374    0.0252   0.0122       0.2999
# 7     B      2    0.3472    0.0488   0.0180       0.2999
```

Cluster B has a systematically larger `residual_var` (0.30 vs. 0.22),
directly reflecting its higher σ (0.50 vs. 0.40).

## Shelf life comparison

```{r grouped-shelf-life}
# Use the pooled training range across both clusters (arcsin-sqrt scale)
pr_all <- range(et_sim$train$z_diff)   # e.g. c(-1.67, 2.04) => width ~ 3.71

sl_all <- shelf_life(
  pred_all,
  plausible_range          = pr_all,
  ci_level                 = 0.90,
  threshold                = 1.0,
  time_col                 = "year",
  max_extrapolation_factor = 10
)

print(sl_all)
# ErrorTracer shelf life analysis
#   Observations     : 10
#   Plausible range  : 3.71
#   Threshold        : 1
#   [A]  100.0% informative  mean ratio = 0.465  shelf life > 2019 (lower bound)
#   [B]  100.0% informative  mean ratio = 0.544  shelf life > 2019 (lower bound)
```

Both clusters remain strongly informative across all five forecast years.
Cluster B's mean ratio (~0.54) is noticeably higher than cluster A's (~0.47),
directly reflecting cluster B's higher residual variance (σ = 0.50 vs. 0.40).
Both are well below the uninformative threshold of 1.0, and neither crosses
within the 2015–2019 window, so both are reported as lower bounds.

The difference in ratios has a practical implication: under a longer forecast
horizon or with larger environmental noise inputs, **cluster B would approach
the uninformative threshold first** — a direct consequence of its higher
biological process noise.

```{r plot-grouped-shelf-life}
et_plot_shelf_life(sl_all, show_ratio = TRUE) +
  labs(
    subtitle = "Both clusters — 90% CI width / plausible range",
    colour   = "Cluster"
  )
```

## Calibration and forecast plots

```{r grouped-calibrate}
cal_all <- et_calibrate(
  pred_all,
  observed     = et_sim$validation,
  response_col = "z_diff",
  ci_levels    = c(0.50, 0.80, 0.90, 0.95)
)

et_plot_calibration(cal_all) +
  labs(subtitle = "Both clusters — posterior predictive calibration")
```

```{r grouped-coef-plot}
# Forest plot for all groups in a single faceted figure
et_plot_coefficients(fit_all) +
  labs(subtitle = "Bayesian 95% CI (blue) vs. enet coefficient (red ×)")
```

---

# Long-horizon forecasting with GCM projections

The shelf-life analysis becomes most powerful when combined with climate
projections from Global Circulation Models (GCMs).  The idea is:

1. **Validate** that the model works on real held-out years (e.g.\ 2022–2025).
2. **Project** onto GCM-simulated climate for 2025–2075 to ask: *over what
   horizon does the forecast remain informative?*

A critical refinement is that GCM uncertainty is **not constant** — ensemble
spread (the disagreement among model runs) typically grows with the projection
horizon.  The `env_noise` argument now accepts **per-row vectors** so you can
pass growing noise SDs directly.

```{r gcm-workflow, eval = FALSE}
# Suppose you have GCM ensemble-mean predictors for 2025-2075,
# standardised using your training-period statistics.
gcm_years <- 2025:2075

gcm_df <- data.frame(
  year  = gcm_years,
  Tmean = ...,    # standardised GCM ensemble-mean temperature
  PPT   = ...,    # standardised GCM ensemble-mean precipitation
  SWE   = ...     # standardised GCM ensemble-mean snowpack
)

# GCM ensemble spread (SD across runs) grows roughly linearly with time.
# Here we model spread as starting at the validation-period noise level
# and increasing by ~0.01 SD per year for temperature.
base_year   <- 2025
tmean_noise <- 0.30 + 0.01 * (gcm_years - base_year)  # 0.30 → 0.80
ppt_noise   <- 0.20 + 0.005 * (gcm_years - base_year) # 0.20 → 0.45
swe_noise   <- 0.15 + 0.005 * (gcm_years - base_year) # 0.15 → 0.40

# Predict over the 50-year GCM window with time-varying noise
pred_gcm <- et_predict(
  model     = fit_A,
  newdata   = gcm_df,
  env_noise = list(
    Tmean = tmean_noise,   # per-row vector, length = nrow(gcm_df)
    PPT   = ppt_noise,
    SWE   = swe_noise
  ),
  ci_levels = c(0.90, 0.95),
  n_draws   = 1000L
)

# Shelf life over the 50-year horizon
sl_gcm <- shelf_life(
  pred_gcm,
  plausible_range          = range(train_A$z_diff),
  time_col                 = "year",
  max_extrapolation_factor = Inf    # allow projection beyond the 50-yr window
)

print(sl_gcm)
# If the 90% CI stays below the plausible range throughout 2025-2075:
#   Shelf life : > 2075 (lower bound — all periods informative, no trend)
# If it crosses (say, around 2058):
#   Shelf life : ~2058 (observed — threshold first exceeded)
```

The per-row noise means that `decompose_uncertainty()` now shows `env_var`
**increasing** across forecast years — directly capturing the epistemic
uncertainty from the GCM ensemble spread.

```{r gcm-decompose, eval = FALSE}
decomp_gcm <- decompose_uncertainty(pred_gcm)
# env_var should be small in 2025 and growing toward 2075 as GCM spread
# accumulates; param_var and residual_var remain roughly constant.
plot(gcm_years, decomp_gcm$env_var,
     type = "l", xlab = "Year", ylab = "Environmental variance",
     main = "Growing GCM ensemble uncertainty propagated into forecast")
```

> **Interpreting the result for conservation.**
> If the shelf life exceeds your planning horizon (e.g.\ "> 2075"), you can
> state: *"Under the [GCM scenario], the 90% credible interval for allele-
> frequency change remains within the plausible biological range throughout
> the 50-year planning horizon, supporting its use in long-term conservation
> genomic forecasting."*
> If a crossing year is observed (e.g.\ 2058), you can report the horizon
> directly and note which uncertainty source (parameter, environmental, or
> residual) drives the loss of informativeness.

---

# Using `lm` instead of `glmnet`

`extract_priors()` is not limited to elastic net.  The `lm` method uses
OLS standard errors (× `multiplier`) as prior SDs instead of coefficient
magnitudes, giving narrower priors when SEs are small.

```{r lm-priors}
lm_fit_A <- lm(z_diff ~ Tmean + PPT + SWE, data = train_A)

prior_lm_A <- extract_priors(
  model      = lm_fit_A,
  multiplier = 2.0,
  min_sd     = 0.10
)

print(prior_lm_A)
# ErrorTracer prior specification
#   Method      : lm
#   Predictors  : 3
#   Multiplier  : 2
#   Min SD      : 0.1
#   Coefficients:
#     Tmean                 mean =   0.6669  sd =   1.3338
#     PPT                   mean =  -0.4677  sd =   0.9354
#     SWE                   mean =   0.3859  sd =   0.7719
```

In this case the OLS and elastic-net estimates are very close (n = 20 is not
a sparse problem with only 3 predictors), so the two prior specifications
lead to nearly identical posteriors.  The key advantage of the elastic-net
pathway arises when the number of predictors is large relative to n, where
regularisation is essential.

---

# Standardising and back-transforming

The `standardization` slot stores the training-period mean and SD for each
predictor, enabling back-transformation to original units with `unstandardize()`.

```{r back-transform}
# Standardisation constants used when generating et_sim
et_sim$standardization
# $Tmean
#      mean        sd
# 15.5422  0.6797
#
# $PPT
#      mean        sd
# 110.5460  14.6855
#
# $SWE
#      mean        sd
# 86.9190  15.1885

# A Tmean of +1 SD corresponds to:
unstandardize(z = 1.0,
              mu = et_sim$standardization$Tmean["mean"],
              s  = et_sim$standardization$Tmean["sd"])
# [1] 16.22   ← 1 SD above training mean = 16.2 °C
```

---

# Summary

This vignette has demonstrated the complete ErrorTracer pipeline on simulated
data where the ground truth is known:

1. **`extract_priors()`** converted elastic-net coefficients into
   `brms`-compatible informed priors, transferring regularised estimates as
   Bayesian prior means.

2. **`et_fit()`** fitted the Bayesian model for a single cluster and for
   both clusters simultaneously via `grouping = "cluster_id"`.

3. **`et_diagnose()`** confirmed convergence (Rhat < 1.05, zero divergences)
   and adequate model fit (no high Pareto-k observations for cluster A).

4. **`et_predict()`** generated posterior predictive draws and propagated
   environmental measurement uncertainty.  Environmental noise contributed
   ~17–25% of total variance — comparable to parameter uncertainty in this
   small-sample setting.

5. **`decompose_uncertainty()`** partitioned forecast variance into three
   components, identifying residual process noise as the dominant bottleneck
   (~80%) in both clusters.

6. **`shelf_life()`** quantified that cluster A forecasts (mean ratio ~0.47)
   are more informative than cluster B (~0.54), consistent with cluster B's
   higher residual variance.  Both clusters are well below the uninformative
   threshold (1.0), so the shelf life exceeds the entire 2015–2019 validation
   window — reported as lower bounds.  The three-mode horizon logic
   (`observed` / `projected` / `lower_bound`) and the
   `max_extrapolation_factor` cap prevent spurious far-future projections
   when the CI/range trend is nearly flat.

7. **`et_calibrate()`** confirmed near-nominal posterior predictive coverage
   on held-out validation data.

8. **Parameter recovery** confirmed that all four true coefficients fell
   within their 95% posterior credible intervals, validating the pipeline's
   statistical integrity.

---

# Session information

```{r session-info, eval = TRUE, echo = FALSE}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{et_plot_prior_posterior}
\alias{et_plot_prior_posterior}
\title{Plot prior vs posterior distributions for model coefficients}
\usage{
et_plot_prior_posterior(model, max_preds = 8L, n_prior_draws = 4000L)
}
\arguments{
\item{model}{An \code{et_model} object.}

\item{max_preds}{Integer.  Maximum number of predictors to show
(default 8).  Predictors are shown in the order they appear in the
prior specification.}

\item{n_prior_draws}{Integer.  Number of random draws for the prior
density (default 4000).}
}
\value{
A \code{ggplot2} object.
}
\description{
Overlays prior and posterior density for each predictor coefficient,
visualising how much the data update the priors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decompose.R
\name{decompose_uncertainty}
\alias{decompose_uncertainty}
\title{Extract or recompute uncertainty decomposition}
\usage{
decompose_uncertainty(predictions, ...)
}
\arguments{
\item{predictions}{An \code{et_prediction} object from
\code{\link{et_predict}}, or an \code{et_prediction_list} (grouped).}

\item{...}{Unused.}
}
\value{
A \code{data.frame} with columns
  \code{obs_id, param_var, env_var, residual_var, total_var}
  (plus \code{temporal_var} for autocorrelation models, and a leading
  \code{group} column for grouped predictions).
}
\description{
Returns a \code{data.frame} with the uncertainty decomposition stored
inside an \code{et_prediction} object:
\describe{
  \item{param_var}{Variance of the posterior linear predictor — captures
    uncertainty in fitted regression coefficients.}
  \item{env_var}{Additional variance arising from measurement or
    prediction uncertainty in the predictor values (estimated via
    perturbation in \code{\link{et_predict}}).  Zero when
    \code{env_noise = NULL}.}
  \item{residual_var}{Posterior mean of \eqn{\sigma^2} (or its
    family-specific analogue) — biological process noise, unmeasured
    drivers, and drift. For autocorrelation models this is the
    \emph{innovation} variance, not the stationary marginal variance;
    the autocorrelated accumulation is reported separately in
    \code{temporal_var}.}
  \item{temporal_var}{(Only present when the model formula contains an
    autocorrelation term such as \code{ar()}, \code{ma()},
    \code{arma()}, \code{cosy()}, \code{unstr()}, \code{sar()}, or
    \code{car()}.) Variance attributable to residual temporal or
    spatial dependence beyond the iid \code{param + residual} sum,
    computed as
    \code{pmax(0, total_var - (param_var + residual_var))}.
    \code{env_var} is deliberately excluded from this gap because it
    is an additive perturbation-based augmentation measured outside
    of \code{posterior_predict}.}
  \item{total_var}{Variance of the full posterior predictive draws,
    including any autocorrelation structure modelled by brms.}
}
}
\details{
All variance components are guaranteed non-negative. When
\code{temporal_var} is present, \code{param_var + residual_var +
temporal_var} reconstructs \code{total_var} (modulo Monte Carlo error);
when it is absent, \code{param_var + residual_var} does.
\code{env_var} is always additive on top, representing the extra
variance that would be contributed by perturbing predictors with
\code{env_noise}.
}
\examples{
\donttest{
set.seed(1)
df  <- data.frame(y = rnorm(20), x1 = rnorm(20))
fit <- et_fit(y ~ x1, data = df,
              chains = 1, iter = 500, warmup = 250,
              cores = 1, refresh = 0)
new_df <- data.frame(x1 = rnorm(5))
pred   <- et_predict(fit, newdata = new_df,
                     env_noise = list(x1 = 0.2),
                     n_draws = 200, n_perturb = 50)
decomp <- decompose_uncertainty(pred)
head(decomp)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{et_plot_shelf_life}
\alias{et_plot_shelf_life}
\title{Plot forecast shelf life}
\usage{
et_plot_shelf_life(sl, show_ratio = TRUE)
}
\arguments{
\item{sl}{An \code{et_shelf_life} object from \code{\link{shelf_life}}.}

\item{show_ratio}{Logical.  If \code{TRUE} (default), plots the
ratio (CI width / plausible range) rather than raw CI width.}
}
\value{
A \code{ggplot2} object.
}
\description{
Shows how credible interval width grows over the forecast horizon and
marks the threshold beyond which the forecast is uninformative.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calibrate.R
\name{et_calibrate}
\alias{et_calibrate}
\title{Assess calibration of posterior predictive intervals}
\usage{
et_calibrate(predictions, observed, response_col = NULL, ci_levels = NULL, ...)
}
\arguments{
\item{predictions}{An \code{et_prediction} or \code{et_prediction_list}.}

\item{observed}{A \code{data.frame} with true response values.  Must have
the same number of rows as \code{predictions$newdata} (rows are matched
positionally) and a column with the true response values.}

\item{response_col}{Character.  Name of the response column in
\code{observed}.  Defaults to the left-hand side of the model formula
if it can be inferred, otherwise must be specified.}

\item{ci_levels}{Numeric vector.  CI levels to assess.  Defaults to all
levels present in the \code{et_prediction} object.}

\item{...}{Unused.}
}
\value{
A \code{data.frame} with columns:
  \describe{
    \item{ci_level}{Nominal CI level.}
    \item{nominal}{Same as ci_level.}
    \item{observed_coverage}{Fraction of true values falling inside the CI.}
    \item{n_obs}{Number of observations used.}
    \item{calibration_error}{Signed difference: observed - nominal.
      Positive = over-coverage (CIs too wide / conservative).
      Negative = under-coverage (CIs too narrow / overconfident).}
    \item{sharpness}{Mean CI width across observations.  Sharpness and
      calibration are complementary: a model can be calibrated but useless
      if sharpness is poor (very wide CIs).}
  }
  For grouped predictions, a \code{group} column is prepended.
}
\description{
Computes observed coverage probability at multiple nominal CI levels.
A well-calibrated Bayesian model should produce 90\% CIs that contain
the true value 90\% of the time, etc.
}
\examples{
\donttest{
set.seed(1)
df  <- data.frame(y = rnorm(20), x1 = rnorm(20))
fit <- et_fit(y ~ x1, data = df,
              chains = 1, iter = 500, warmup = 250,
              cores = 1, refresh = 0)
valid_df <- data.frame(y = rnorm(5), x1 = rnorm(5))
pred <- et_predict(fit, newdata = valid_df,
                   n_draws = 200, n_perturb = 50)
cal <- et_calibrate(pred, observed = valid_df, response_col = "y")
print(cal)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{et_plot_calibration}
\alias{et_plot_calibration}
\title{Plot calibration: observed vs nominal coverage}
\usage{
et_plot_calibration(cal, group_col = NULL)
}
\arguments{
\item{cal}{A \code{data.frame} from \code{\link{et_calibrate}}.}

\item{group_col}{Optional character.  Name of a column in \code{cal} that
identifies sub-groups (e.g.\ \code{"species"}, \code{"cluster_id"}).
When \code{NULL} (default), \code{et_plot_calibration} uses the
\code{group} column if present; otherwise it auto-detects any single
non-canonical column with more than one unique value and treats it as
the grouping.  Set to \code{NA} to force a single un-grouped series.}
}
\value{
A \code{ggplot2} object.
}
\description{
A well-calibrated model produces points along the 1:1 diagonal.
Points above the diagonal indicate over-coverage (conservative);
below indicates under-coverage (anti-conservative).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{et_plot_forecast}
\alias{et_plot_forecast}
\title{Plot posterior predictive fan chart}
\usage{
et_plot_forecast(
  predictions,
  observed = NULL,
  response_col = NULL,
  time_col = NULL
)
}
\arguments{
\item{predictions}{An \code{et_prediction} object.}

\item{observed}{Optional \code{data.frame} with true response values.
If provided, points are overlaid.}

\item{response_col}{Character.  Name of the response column in
\code{observed}.}

\item{time_col}{Character.  Column in \code{predictions$newdata} used
as the x-axis.  Defaults to observation index.}
}
\value{
A \code{ggplot2} object.
}
\description{
Overlays nested credible interval ribbons on the median forecast, with
optional observed values for calibration assessment.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit.R
\name{et_fit}
\alias{et_fit}
\title{Fit a Bayesian regression model with informed priors}
\usage{
et_fit(
  formula,
  data,
  priors = NULL,
  chains = 4L,
  iter = 2000L,
  warmup = floor(iter/2),
  cores = min(chains, parallel::detectCores()),
  seed = 42L,
  adapt_delta = 0.95,
  max_treedepth = 12L,
  grouping = NULL,
  eiv = NULL,
  silent = 2L,
  ...
)
}
\arguments{
\item{formula}{An R formula, e.g.\ \code{response ~ .} or
\code{y ~ x1 + x2}.}

\item{data}{A \code{data.frame} with all predictors and the response.}

\item{priors}{An \code{et_prior_spec} object from
\code{\link{extract_priors}}, or a \code{brmsprior} object, or
\code{NULL} for brms defaults.}

\item{chains}{Integer.  Number of MCMC chains (default 4).}

\item{iter}{Integer.  Total iterations per chain, including warmup
(default 2000).}

\item{warmup}{Integer.  Warmup iterations per chain (default
\code{floor(iter / 2)}).}

\item{cores}{Integer.  Parallel cores (default
\code{min(chains, parallel::detectCores())}).}

\item{seed}{Integer.  Random seed for reproducibility (default 42).}

\item{adapt_delta}{Numeric.  Target acceptance probability for HMC
(default 0.95).}

\item{max_treedepth}{Integer.  Maximum tree depth (default 12).}

\item{grouping}{Character.  Name of a column in \code{data} to use for
grouping.  If non-\code{NULL}, one model is fitted per unique group
value and an \code{et_model_list} is returned.}

\item{eiv}{Optional errors-in-variables specification.  A named list /
vector mapping predictor names to either a scalar SD or a vector of
per-row SDs (length \code{nrow(data)}).  For each entry, the formula
term for that predictor is rewritten as \code{brms::me(pred, se_pred)}
(an auxiliary \code{se_<pred>} column is appended to \code{data}), so
the posterior reflects measurement error in the predictor as well as
coefficient uncertainty.  The beta posteriors widen accordingly, which
partially absorbs what ErrorTracer's downstream \code{env_var}
component would otherwise report.  When \code{eiv} is supplied together
with an \code{et_prior_spec} from \code{\link{extract_priors}}, the
informed priors are \emph{dropped} because they target \code{class = "b"}
terms and \code{me()} terms live under \code{class = "bsp"}; brms
defaults are used instead (and a warning is logged).}

\item{silent}{Integer passed to \code{brms::brm()} (default 2, no Stan
output).}

\item{...}{Additional arguments passed to \code{brms::brm()}.}
}
\value{
An \code{et_model} object (or an \code{et_model_list} if
  \code{grouping} is specified).
}
\description{
Wraps \code{brms::brm()} and attaches the prior specification,
training data reference, and configuration for downstream uncertainty
decomposition.  Pass \code{priors} from \code{\link{extract_priors}} to
use regularized-model coefficients as prior means; omit it for
default (weakly informative) priors.
}
\examples{
\donttest{
set.seed(1)
df  <- data.frame(y = rnorm(20), x1 = rnorm(20), x2 = rnorm(20))
ps  <- extract_priors(lm(y ~ x1 + x2, data = df))
fit <- et_fit(y ~ x1 + x2, data = df, priors = ps,
              chains = 1, iter = 500, warmup = 250,
              cores = 1, refresh = 0)
print(fit)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{et_plot_coefficients}
\alias{et_plot_coefficients}
\title{Forest plot of regression coefficients}
\usage{
et_plot_coefficients(model)
}
\arguments{
\item{model}{An \code{et_model} or \code{et_model_list}.}
}
\value{
A \code{ggplot2} object.
}
\description{
Compares Bayesian posterior estimates (95\% CI) with the regularized
coefficient values used as prior means (shown as crosses).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{extract_priors}
\alias{extract_priors}
\alias{extract_priors.lm}
\alias{extract_priors.glm}
\alias{extract_priors.cv.glmnet}
\alias{extract_priors.glmnet}
\alias{extract_priors.ranger}
\title{Extract brms prior specification from a regularized or standard regression model}
\usage{
extract_priors(
  model,
  multiplier = 2,
  min_sd = 0.1,
  intercept_prior_sd = NULL,
  sigma_prior_scale = 1,
  ...
)

\method{extract_priors}{lm}(
  model,
  multiplier = 2,
  min_sd = 0.1,
  intercept_prior_sd = NULL,
  sigma_prior_scale = 1,
  ...
)

\method{extract_priors}{glm}(
  model,
  multiplier = 2,
  min_sd = 0.1,
  intercept_prior_sd = NULL,
  sigma_prior_scale = 1,
  ...
)

\method{extract_priors}{cv.glmnet}(
  model,
  multiplier = 2,
  min_sd = 0.1,
  intercept_prior_sd = NULL,
  sigma_prior_scale = 1,
  lambda = "lambda.min",
  ...
)

\method{extract_priors}{glmnet}(
  model,
  multiplier = 2,
  min_sd = 0.1,
  intercept_prior_sd = NULL,
  sigma_prior_scale = 1,
  s = 1L,
  ...
)

\method{extract_priors}{ranger}(
  model,
  multiplier = 2,
  min_sd = 0.1,
  intercept_prior_sd = NULL,
  sigma_prior_scale = 1,
  ...
)
}
\arguments{
\item{model}{A fitted model.  Supported classes:
\itemize{
  \item \code{cv.glmnet} / \code{glmnet} — elastic net / lasso
  \item \code{lm} — ordinary least squares
  \item \code{glm} — generalized linear model
  \item \code{ranger} — random forest (importance-scaled flat priors)
}}

\item{multiplier}{Numeric scalar.  Prior SD is set to
\code{multiplier * |coef|} (for signed-coefficient methods) or
\code{multiplier * importance_normalised} (for ranger).  Default 2.0.}

\item{min_sd}{Numeric scalar.  Minimum prior SD to avoid degenerate
(spike) priors on near-zero coefficients.  Default 0.1.}

\item{intercept_prior_sd}{Optional prior SD for the intercept term.  The
default \code{NULL} emits \emph{no} explicit intercept prior and lets
\code{brms} pick its data-aware default
(\code{student_t(3, median(y), mad(y))}).  If you supply a numeric value,
the prior becomes \code{normal(0, intercept_prior_sd)} --- this is only
appropriate when the response has been centred (or nearly so) before
fitting, since \code{brms}'s \code{class = "Intercept"} refers to the
intercept at the centre of the predictors, whose posterior mean equals
\code{E[y]} there.  Supplying a small \code{intercept_prior_sd} for an
uncentred response will pull the intercept toward zero and cripple the
fit.}

\item{sigma_prior_scale}{Scale parameter for the half-Cauchy prior on the
residual SD sigma.  Default 1.0.}

\item{...}{Additional arguments passed to methods.}

\item{lambda}{Which lambda to extract coefficients from when the model is a
\code{cv.glmnet} object.  Either \code{"lambda.min"} (default) or
\code{"lambda.1se"}.}

\item{s}{Index (integer) or value of lambda to extract coefficients at when
the model is a plain \code{glmnet} object.  Defaults to the first lambda
(smallest regularisation).}
}
\value{
An \code{et_prior_spec} list containing:
  \describe{
    \item{\code{prior}}{A \code{brmsprior} object for use in \code{brms::brm()}.}
    \item{\code{pred_names}}{Character vector of included predictor names.}
    \item{\code{coefs}}{Named numeric vector of regularized coefficients
      (NULL for ranger, which uses importance instead).}
    \item{\code{method}}{Character: the dispatch method used.}
    \item{\code{multiplier}, \code{min_sd}}{Settings echo.}
  }
}
\description{
Converts a fitted model object into a \code{brms} prior specification
suitable for \code{\link{et_fit}}.  The coefficient estimates (or
importance weights for \pkg{ranger}) from the regularized or standard fit are used as
informative prior means, so the Bayesian model starts close to the
regularized or standard solution while remaining open to data-driven revision.
}
\details{
For \code{ranger} models, signed coefficients are not available.  Priors
are centred at zero (direction unknown) and the prior SD for each predictor
is set to \code{multiplier * importance_normalised}, where importance is
normalised to the \code{[min_sd, 1]} interval.  Only variables with
positive permutation importance are included.
}
\examples{
fit_lm <- lm(mpg ~ wt + hp + cyl, data = mtcars)
ps <- extract_priors(fit_lm, multiplier = 2, min_sd = 0.1)
print(ps)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{unstandardize}
\alias{unstandardize}
\title{Reverse standardization}
\usage{
unstandardize(z, mu, s)
}
\arguments{
\item{z}{Standardized values.}

\item{mu}{Original mean.}

\item{s}{Original standard deviation.}
}
\value{
Values on the original scale.
}
\description{
Reverse standardization
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{et_sim}
\alias{et_sim}
\title{Simulated allele-frequency-change dataset for ErrorTracer tutorials}
\format{
A named \code{list} with seven elements:
\describe{
  \item{\code{train}}{
    \code{data.frame} with 100 rows and 6 columns (2 clusters \eqn{\times}
    50 training years, 1970–2019):
    \describe{
      \item{year}{Integer. Calendar year.}
      \item{cluster_id}{Character. SNP cluster identifier: \code{"A"} or
        \code{"B"}.}
      \item{Tmean}{Numeric. Standardised mean growing-season temperature
        (original unit: °C; standardised using training-period mean and SD).}
      \item{PPT}{Numeric. Standardised total growing-season precipitation
        (original unit: mm).}
      \item{SWE}{Numeric. Standardised peak snow water equivalent
        (original unit: mm).}
      \item{z_diff}{Numeric. Simulated allele-frequency change on the
        arcsin-sqrt scale (\eqn{\arcsin(\sqrt{f})} transformation).
        This transformation is unbounded and has no fixed
        [\eqn{-1}, 1] constraint; the plausible range should be derived
        from the training data or from the theoretical bounds
        (\eqn{[0,\, \pi/2]}\ on the arcsin scale).}
    }
  }
  \item{\code{forecast}}{
    \code{data.frame} with 30 rows and 5 columns (2 clusters \eqn{\times}
    15 forecast years, 2020–2034).  Same columns as \code{train} except
    \code{z_diff} is absent — these are the prediction targets.  Predictors
    are standardised using the training-period statistics stored in
    \code{standardization}.
  }
  \item{\code{validation}}{
    \code{data.frame} with 30 rows and 3 columns (\code{year},
    \code{cluster_id}, \code{z_diff}).  True response values for the
    forecast period; used with \code{\link{et_calibrate}} to assess
    posterior predictive coverage.
  }
  \item{\code{true_params}}{
    Named \code{list} with one element per cluster.  Each element is a named
    numeric vector of the true data-generating parameters:
    \code{intercept}, \code{Tmean}, \code{PPT}, \code{SWE}, and
    \code{sigma} (residual SD).
    \describe{
      \item{Cluster A}{\code{intercept = 0.00}, \code{Tmean = 0.50},
        \code{PPT = -0.30}, \code{SWE = 0.20}, \code{sigma = 0.30}}
      \item{Cluster B}{\code{intercept = 0.10}, \code{Tmean = 0.30},
        \code{PPT = -0.20}, \code{SWE = -0.25}, \code{sigma = 0.35}}
    }
  }
  \item{\code{env_noise}}{
    Named \code{list}.  Suggested per-predictor environmental noise SDs (on
    the standardised scale) for use with the \code{env_noise} argument of
    \code{\link{et_predict}}: \code{Tmean = 0.30}, \code{PPT = 0.20},
    \code{SWE = 0.15}.
  }
  \item{\code{standardization}}{
    Named \code{list} with one element per predictor.  Each element is a
    named numeric vector \code{c(mean = ..., sd = ...)} giving the
    training-period statistics used to standardise the data.  Needed if you
    wish to back-transform predictions to the original (unstandardised)
    scale with \code{\link{unstandardize}}.
  }
  \item{\code{description}}{
    Character string.  Human-readable description of the dataset and how it
    was generated.
  }
}
}
\source{
Generated by \code{data-raw/generate_et_sim.R} with
  \code{set.seed(111)}.
}
\usage{
et_sim
}
\description{
A named list containing training data, forecast predictors, validation
responses, and ground-truth parameters for two synthetic SNP clusters (A
and B) at a hypothetical mountain plant site.  The dataset is designed to
exercise every step of the ErrorTracer pipeline and to support
parameter-recovery validation (true coefficients are known).
}
\details{
The climate time series are simulated as AR(1) processes with a warming
trend in \code{Tmean} (+0.015 °C yr\eqn{^{-1}}, cumulating to ~0.30 SD
above the training mean over the 15-year forecast window).  \code{SWE}
is generated with a weak dependence on \code{Tmean} (negative coupling)
and \code{PPT} (positive coupling) plus a dominant independent noise
component, so that the three predictors have only mild pairwise
correlation (|R| < 0.2 on the training subset).  This makes all
regression coefficients reliably identifiable in a single replicate.
All predictors are standardised using training-period statistics only.

In addition to the three real climate predictors, the dataset includes
ten independent standardised nuisance predictors \code{d1}, ..., \code{d10}
with zero true effect on the response.  They give the regularized-prior
extraction step a real selection job: cv.glmnet (alpha = 0.5) is expected
to identify the three real predictors and shrink the ten dummies toward
zero.  Without them, regularization on three well-identified predictors
merely biases the true coefficients without selecting anything.

Responses are generated from a linear model with Gaussian noise:
\deqn{z\_diff_t = \alpha + \beta_1 Tmean_t + \beta_2 PPT_t +
      \beta_3 SWE_t + \varepsilon_t, \quad \varepsilon_t \sim
      \mathcal{N}(0, \sigma^2)}
The true parameters are stored in \code{et_sim$true_params} for
parameter-recovery validation.

The dataset was generated with \code{set.seed(111)}.  The full generation
script is at \code{data-raw/generate_et_sim.R}.
}
\examples{
data(et_sim)

# Inspect structure
str(et_sim, max.level = 2)

# Training data
head(et_sim$train)

# True parameters for cluster A
et_sim$true_params$A

# Suggested noise SDs for et_predict()
et_sim$env_noise

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calibrate.R
\name{et_diagnose}
\alias{et_diagnose}
\title{Diagnose a fitted ErrorTracer model}
\usage{
et_diagnose(model, loo = TRUE, ...)
}
\arguments{
\item{model}{An \code{et_model} or \code{et_model_list}.}

\item{loo}{Logical.  Whether to run LOO-CV (can be slow; default
\code{TRUE}).}

\item{...}{Unused.}
}
\value{
A list with elements:
  \describe{
    \item{convergence}{List: \code{rhat_max}, \code{rhat_all_ok},
      \code{neff_min}, \code{neff_all_ok}, \code{n_divergences}.}
    \item{loo}{List or \code{NULL}: \code{elpd_loo}, \code{p_loo},
      \code{looic}, \code{n_bad_pareto_k}, \code{loo_object}.}
    \item{summary}{Printed summary from \code{brms::summary()}.}
  }
  For \code{et_model_list}, a named list of per-group diagnostic lists
  plus an aggregated \code{summary} data.frame.
}
\description{
Computes Rhat, effective sample size ratios, divergent transitions, and
leave-one-out cross-validation (LOO-CV) for a fitted \code{et_model}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{standardize}
\alias{standardize}
\title{Standardize a numeric vector (mean-centre, unit-variance)}
\usage{
standardize(x)
}
\arguments{
\item{x}{Numeric vector.}
}
\value{
Standardized numeric vector. Returns zeros if variance is zero.
}
\description{
Standardize a numeric vector (mean-centre, unit-variance)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{et_plot_decomposition}
\alias{et_plot_decomposition}
\title{Plot uncertainty decomposition}
\usage{
et_plot_decomposition(decomp, proportional = TRUE, group_col = NULL)
}
\arguments{
\item{decomp}{A \code{data.frame} from \code{\link{decompose_uncertainty}}
or directly the \code{$decomposition} slot of an \code{et_prediction}.}

\item{proportional}{Logical.  If \code{TRUE} (default), bars are scaled to
sum to 1 (proportional contribution).  If \code{FALSE}, raw variances
are shown.}

\item{group_col}{Character.  Optional name of a grouping column in
\code{decomp} (present for grouped predictions).}
}
\value{
A \code{ggplot2} object.
}
\description{
Produces a stacked bar chart showing the relative contributions of
parameter, environmental, and residual variance for each observation,
plus a fourth temporal-autocorrelation component when present in
\code{decomp} (see \code{\link{decompose_uncertainty}}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{et_predict}
\alias{et_predict}
\title{Posterior prediction with uncertainty decomposition}
\usage{
et_predict(
  model,
  newdata,
  env_noise = NULL,
  env_cov = NULL,
  env_dist = NULL,
  n_draws = 2000L,
  ci_levels = c(0.5, 0.8, 0.9, 0.95),
  n_perturb = NULL,
  n_env_draws = 1L,
  interval_type = c("predictive", "linpred"),
  include_env_in_ci = FALSE,
  ...
)
}
\arguments{
\item{model}{An \code{et_model} or \code{et_model_list} object from
\code{\link{et_fit}}.}

\item{newdata}{A \code{data.frame} containing the predictor columns
named in the model formula.  For grouped models, must also contain
the grouping column.}

\item{env_noise}{Environmental measurement / prediction uncertainty.
Can be:
\itemize{
  \item \code{NULL} (default): no environmental noise.
  \item A single numeric: applied as a fraction of each predictor's
    empirical SD in \code{newdata} (e.g.\ \code{0.1} means 10\% noise,
    constant across all observations).
  \item A named \code{list} or named numeric vector with one scalar per
    predictor: constant absolute noise SD per predictor, e.g.\
    \code{list(Tmean = 0.5, PPT = 10)}.
  \item A named \code{list} where each entry is a \strong{numeric vector
    of length \code{nrow(newdata)}}: \emph{time-varying} (per-row) noise
    SDs.  Use this when predictor uncertainty grows with forecast horizon,
    e.g.\ from a GCM ensemble spread that increases over time:
    \code{list(Tmean = 0.30 + 0.01 * (years - base_year))}.
    Entries not supplied default to zero (no noise for that predictor).
}}

\item{env_cov}{Correlation structure of the environmental noise.  The
\emph{magnitudes} of the noise come from \code{env_noise}; \code{env_cov}
supplies the \emph{correlation} between predictors, so that a perturbation
on row \eqn{i} is drawn from
\eqn{\mathcal{N}(0,\; D_i R D_i)} with
\eqn{D_i = \mathrm{diag}(\sigma_{i1}, \ldots, \sigma_{ip})} and
\eqn{R} = the correlation matrix.  One of:
\itemize{
  \item \code{NULL} (default): independent noise, \eqn{R = I} --- equivalent
    to ErrorTracer behaviour prior to this feature and the right choice
    when predictor measurement errors are genuinely independent (e.g.\
    separate instruments on unrelated variables).
  \item \code{"empirical"}: compute the correlation of the predictor
    columns in the \strong{training data} (\code{model$data}).  Use this
    when predictor \emph{errors} are expected to inherit the correlation
    structure of the predictors themselves --- e.g.\ temperature and
    humidity that co-vary in the underlying climate system.
  \item \code{"newdata"}: compute the correlation of the predictor columns
    in \code{newdata}.  Useful when the forecast window has a different
    covariance structure than training (e.g.\ scenario runs).
  \item A numeric \eqn{p \times p} matrix with \code{dimnames} matching the
    model's predictors.  Entries with an off-diagonal exceeding 1 are
    rescaled to a correlation matrix.  Use this to supply an independent
    estimate of the \emph{error} correlation structure (e.g.\ from a
    reanalysis product or a sensor covariance report).
}
A correlation derived from training data is a working assumption: the
structure of the \emph{errors} is assumed to mirror the structure of the
\emph{values}.  When this is implausible, pass a matrix directly.}

\item{env_dist}{Distributional form of the per-predictor noise.  The
\code{env_noise} SDs set the \emph{magnitude} of the perturbation;
\code{env_dist} sets its \emph{shape}.  For every distribution other than
\code{"gaussian"}, the noise is calibrated so that (approximately)
\eqn{E[\tilde x] = x} and \eqn{\mathrm{Var}[\tilde x] = \sigma^2}, using a
Gaussian copula to honour \code{env_cov}.  One of:
\itemize{
  \item \code{NULL} (default): \code{"gaussian"} for every predictor ---
    additive Gaussian noise, legacy behaviour.
  \item A single string (\code{"gaussian"}, \code{"lognormal"},
    \code{"gamma"}, \code{"beta"}): applied to all predictors.
  \item A named list / character vector with one entry per predictor to
    override the default, e.g.\
    \code{list(PPT = "gamma", tmax = "gaussian")}.
}
Distributions:
\describe{
  \item{\code{"gaussian"}}{Additive normal noise (\eqn{\tilde x = x + \varepsilon},
    \eqn{\varepsilon \sim N(0, \sigma^2)}).  Appropriate for symmetric
    measurement error on a continuous, potentially negative scale
    (temperature, anomalies).}
  \item{\code{"lognormal"}}{Multiplicative noise: \eqn{\log \tilde x \sim
    N(\log x - s^2/2,\; s^2)} with \eqn{s^2 = \log(1 + (\sigma/x)^2)}.
    Preserves positivity; right-tail skewed.  Natural for strictly
    positive continuous variables whose error scales with magnitude
    (e.g.\ enzyme activity, biomass).  Rows with \eqn{x \le 0} are
    left unperturbed.}
  \item{\code{"gamma"}}{\eqn{\tilde x \sim \mathrm{Gamma}(\mathrm{shape} =
    (x/\sigma)^2,\; \mathrm{rate} = x/\sigma^2)}.  Positive support,
    right-skewed, analytic mean/variance match.  Natural for precipitation,
    rates, and other non-negative continuous variables.  Rows with
    \eqn{x \le 0} are left unperturbed.}
  \item{\code{"beta"}}{\eqn{\tilde x \sim \mathrm{Beta}(\alpha,\beta)} with
    \eqn{\alpha + \beta = x(1-x)/\sigma^2 - 1}.  Support in \eqn{(0,1)};
    appropriate for proportions and probabilities (allele frequencies,
    presence rates).  Rows with \eqn{x \not\in (0,1)} or
    \eqn{\sigma^2 \ge x(1-x)} are left unperturbed.}
}
Correlation (\code{env_cov}) is applied to the latent standard-normal
draws before the marginal quantile transform, so rank correlations are
preserved across distributions.}

\item{n_draws}{Integer.  Number of posterior draws to use (default 2000;
capped at the number of draws available in the fit).}

\item{ci_levels}{Numeric vector.  Credible interval levels to compute
(default \code{c(0.5, 0.8, 0.9, 0.95)}).}

\item{n_perturb}{Integer.  Number of posterior draws used for the
environmental perturbation step (default \code{min(500, n_draws)}).
Reducing this speeds up computation.}

\item{n_env_draws}{Integer.  Number of independent environmental
perturbations averaged \emph{per posterior draw} when estimating
\code{env_var} (default \code{1}).  Increasing this reduces Monte Carlo
noise on the environmental-variance estimate at the cost of proportional
computation.  The decomposition always reports a Monte Carlo SE
(\code{v_env_mcse}) alongside \code{env_var}; it decreases roughly like
\eqn{1/\sqrt{n\_env\_draws \cdot n\_perturb}}.}

\item{interval_type}{Character.  Which draws to use when computing credible
intervals:
\itemize{
  \item \code{"predictive"} (default): draws from \code{posterior_predict},
    which include sigma (residual noise).  Use this when forecasting
    \strong{individual observations} --- e.g. a single population's allele
    frequency, one site's ozone reading on a specific day.
  \item \code{"linpred"}: draws from \code{posterior_linpred}, which
    capture only parameter uncertainty (no sigma).  Use this when
    forecasting the \strong{mean response} --- e.g. the expected ozone
    across many similar days, or mean delta f across replicate populations.
    These intervals are always narrower; they will under-cover individual
    observations unless sigma is negligible.
}
The decomposition components and \code{posterior_predict} /
\code{posterior_linpred} matrices are always computed regardless of this
setting.}

\item{include_env_in_ci}{Logical.  When \code{TRUE} and
\code{interval_type = "predictive"}, credible intervals are constructed
from \strong{environmentally inflated} draws
\eqn{\tilde y = \tilde{\mathrm{lp}} + \varepsilon}, with
\eqn{\tilde{\mathrm{lp}}} the perturbed linear predictor and
\eqn{\varepsilon \sim N(0, \sigma^2)} using posterior draws of
\eqn{\sigma}.  This folds the environmental uncertainty component back
into the CI, which is typically what you want for sensitivity analyses
or whenever the reported interval should cover predictor-measurement
error.  When \code{FALSE} (default, backward compatible), CIs are based
on \code{posterior_predict} only --- parameter + residual, without
predictor noise.}

\item{...}{Passed to methods.}
}
\value{
An \code{et_prediction} object (list) containing:
\describe{
  \item{\code{posterior_predict}}{Matrix \code{[n_draws x n_obs]}: full
    posterior predictive draws (parameter + residual uncertainty).}
  \item{\code{posterior_linpred}}{Matrix \code{[n_draws x n_obs]}: linear
    predictor draws on the \strong{link scale} (parameter uncertainty only).}
  \item{\code{lp_perturbed}}{Matrix \code{[n_perturb x n_obs]}: linear
    predictor on the link scale computed on environmentally perturbed inputs.}
  \item{\code{sigma_draws}}{Numeric vector: posterior draws of sigma
    (\code{NA} for families without a sigma parameter, e.g.\ Binomial).}
  \item{\code{credible_intervals}}{data.frame with columns
    \code{row_id, ci_level, lower, median, upper, width}.}
  \item{\code{decomposition}}{data.frame with columns
    \code{obs_id, total_var, param_var, env_var, v_env_mcse, residual_var}.
    All components are on the \strong{response scale}.
    \code{v_env_mcse} is the Monte Carlo SE of \code{env_var}.
    \code{residual_var} is per-observation for non-Gaussian families
    (e.g.\ Binomial: \eqn{E_s[\mu^{(s)}(1-\mu^{(s)})]}) and constant
    for Gaussian.}
  \item{\code{newdata}}{The input \code{newdata}.}
  \item{\code{model}}{Reference to the \code{et_model} used.}
  \item{\code{env_cov}}{The \eqn{p \times p} correlation matrix actually
    used for perturbation (identity for \code{env_cov = NULL}).}
  \item{\code{env_dist}}{Named character vector mapping each predictor to
    the distribution actually used for its perturbation.}
}
}
\description{
Generates posterior predictive draws for new observations, propagates
environmental measurement uncertainty through the model, and computes
credible intervals.  The resulting \code{et_prediction} object is the
input to \code{\link{decompose_uncertainty}}, \code{\link{shelf_life}},
and the plotting functions.
}
\seealso{
\code{\link{decompose_uncertainty}}, \code{\link{shelf_life}},
  \code{\link{et_calibrate}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{et_theme}
\alias{et_theme}
\title{Minimal ggplot2 theme for ErrorTracer plots}
\usage{
et_theme(base_size = 12)
}
\arguments{
\item{base_size}{Base font size (default 12).}
}
\value{
A \code{ggplot2} theme object.
}
\description{
Minimal ggplot2 theme for ErrorTracer plots
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shelf_life.R
\name{shelf_life}
\alias{shelf_life}
\title{Compute the forecast shelf life}
\usage{
shelf_life(
  predictions,
  response_scale,
  ci_level = 0.9,
  threshold = 1,
  time_col = NULL,
  min_slope_for_projection = 1e-04,
  max_extrapolation_factor = 10,
  ...,
  plausible_range = NULL
)
}
\arguments{
\item{predictions}{An \code{et_prediction} or \code{et_prediction_list}.}

\item{response_scale}{Numeric vector of length 2 (\code{c(min, max)})
giving the response scale used as the denominator in the
CI-width / range ratio.  For unbounded responses use
\code{range(training_data$response)} as a conservative default, or
supply a biologically motivated interval.
The effective range is \code{diff(response_scale)}.}

\item{ci_level}{Numeric.  The credible interval level to use (default
0.90).  Must be present in the \code{et_prediction} object.}

\item{threshold}{Numeric.  CI width / response scale above which the
forecast is uninformative (default 1.0).}

\item{time_col}{Character.  Optional column in
\code{predictions$newdata} to use as the time axis.  If \code{NULL},
observation index is used.}

\item{min_slope_for_projection}{Numeric.  Minimum linear slope (of
ratio vs.\ time) required to attempt extrapolation when all periods
are informative.  Below this value the shelf life is reported as a
lower bound only.  Default \code{1e-4}.}

\item{max_extrapolation_factor}{Numeric.  Cap on how far the linear
projection may reach beyond the observed window.  If the projected
crossing time exceeds
\code{max(time) + max_extrapolation_factor * (max(time) - min(time))},
the result is reported as a lower bound instead of a projection.
Set to \code{Inf} to disable the cap.  Default \code{10}.}

\item{...}{Unused.}

\item{plausible_range}{Deprecated.  Use \code{response_scale} instead.}
}
\value{
An \code{et_shelf_life} object (a \code{data.frame}) with columns:
  \describe{
    \item{obs_id}{Observation index.}
    \item{time}{Time axis value.}
    \item{ci_width}{Width of the credible interval at \code{ci_level}.}
    \item{plausible_range}{Effective response scale (scalar diff).}
    \item{ratio}{CI width / response scale.}
    \item{informative}{Logical; \code{TRUE} when ratio < threshold.}
  }
  For grouped predictions a \code{group} column is prepended.
  The object carries three attributes that drive \code{print()}:
  \describe{
    \item{\code{horizon}}{Named list with elements \code{value},
      \code{type} (\code{"observed"}, \code{"projected"}, or
      \code{"lower_bound"}), \code{last_informative},
      \code{description}, and (for projected) \code{se_t_star}.}
    \item{\code{horizon_by_group}}{For grouped objects: named list of
      per-group horizon lists.}
    \item{\code{threshold}}{The threshold value used.}
  }
}
\description{
Quantifies \emph{when} a forecast becomes uninformative by comparing the
width of credible intervals to a response scale.  A forecast is
uninformative when its CI width exceeds \code{threshold * response_scale}.
}
\details{
The function operates in three modes depending on the available data and
whether the uninformative threshold is crossed within the forecast window:

\describe{
  \item{\strong{Observed}}{The threshold is crossed within the
    forecast/validation window.  The shelf life is the first time point
    at which \code{ratio >= threshold}.}
  \item{\strong{Projected}}{All forecast periods remain informative but
    the CI/range ratio is trending upward.  A linear trend is fitted to
    the ratios and extrapolated to estimate when the threshold would be
    reached.  The projected crossing time \eqn{t^* = (\tau - a) / b}
    (where \eqn{\tau} is the threshold, \eqn{a} the fitted intercept,
    \eqn{b} the fitted slope) is reported together with a Monte Carlo
    standard error \code{se_t_star} derived via the delta method.}
  \item{\strong{Lower bound}}{All forecast periods are informative with
    no upward trend in the ratio.  The shelf life is reported as a lower
    bound: \code{> last observed time}.}
}

The intended workflow is:
\enumerate{
  \item Fit the model (\code{\link{et_fit}}).
  \item Predict on held-out data or a future time window
    (\code{\link{et_predict}}).
  \item Call \code{shelf_life()} on those predictions.
  \item If the threshold is not crossed in the held-out window, the
    \emph{projected} mode extrapolates the horizon automatically.
}
}
\examples{
\donttest{
set.seed(1)
df  <- data.frame(y = rnorm(20), year = 2001:2020, x1 = rnorm(20))
fit <- et_fit(y ~ x1, data = df,
              chains = 1, iter = 500, warmup = 250,
              cores = 1, refresh = 0)
new_df <- data.frame(x1 = rnorm(10), year = 2021:2030)
pred   <- et_predict(fit, newdata = new_df,
                     n_draws = 200, n_perturb = 50)
sl <- shelf_life(pred,
                 response_scale            = range(df$y),
                 ci_level                  = 0.90,
                 threshold                 = 1.0,
                 time_col                  = "year",
                 max_extrapolation_factor  = 10)
print(sl)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sensitivity.R
\name{et_sensitivity_profile}
\alias{et_sensitivity_profile}
\title{Sensitivity profile of the environmental uncertainty component}
\usage{
et_sensitivity_profile(
  model,
  newdata,
  response_scale,
  fraction_grid = NULL,
  absolute_grid = NULL,
  env_cov = NULL,
  env_dist = NULL,
  include_env_in_ci = TRUE,
  ci_level = 0.9,
  ci_levels = c(0.5, 0.8, 0.9, 0.95),
  threshold = 1,
  time_col = NULL,
  n_draws = 2000L,
  n_perturb = NULL,
  max_extrapolation_factor = 10,
  verbose = TRUE,
  plausible_range = NULL
)
}
\arguments{
\item{model}{An \code{et_model} from \code{\link{et_fit}}.}

\item{newdata}{\code{data.frame} passed to \code{et_predict}.}

\item{response_scale}{Two-element numeric vector \code{c(min, max)}, as in
\code{\link{shelf_life}}.  The effective range is
\code{diff(response_scale)}.}

\item{fraction_grid}{Optional numeric vector of scalar noise fractions.}

\item{absolute_grid}{Optional \code{list} of \code{env_noise} arguments
(each a named list / vector).  If supplied together with
\code{fraction_grid}, \code{fraction_grid} is ignored.}

\item{env_cov, env_dist}{Passed through to \code{et_predict}.}

\item{include_env_in_ci}{Logical.  If \code{TRUE} (default), credible
intervals — and hence the shelf-life horizon — are computed from
environmentally inflated draws (\code{et_predict(..., include_env_in_ci
= TRUE)}).  Set \code{FALSE} to see how \code{env_var} evolves while
holding the CI fixed at parameter + residual.}

\item{ci_level}{Numeric.  Used for shelf life and CI width tracking
(default 0.90).  Must be in \code{ci_levels}.}

\item{ci_levels}{Numeric vector of CI levels computed per iteration
(default \code{c(0.5, 0.8, 0.9, 0.95)}).}

\item{threshold}{Numeric.  Shelf-life threshold (default 1.0).}

\item{time_col}{Character.  Optional time column for shelf life.}

\item{n_draws, n_perturb}{Passed to \code{et_predict}.}

\item{max_extrapolation_factor}{Passed to \code{shelf_life}.}

\item{verbose}{Logical.  If \code{TRUE} (default), logs each iteration.}

\item{plausible_range}{Deprecated.  Use \code{response_scale} instead.}
}
\value{
An \code{et_sensitivity} object: a \code{data.frame} with one row
  per grid point and columns
  \describe{
    \item{\code{grid_id}}{1-indexed step.}
    \item{\code{label}}{Short descriptor of the env_noise used
      (e.g.\ \code{"fraction = 0.05"}).}
    \item{\code{fraction}}{Scalar fraction when applicable; \code{NA}
      for absolute-grid rows.}
    \item{\code{param_var, env_var, residual_var, total_var}}{Mean across
      forecast observations.}
    \item{\code{env_share}}{\code{env_var / (env_var + total_var)} ---
      the fraction of combined predictive variance attributable to
      predictor noise (bounded in \eqn{[0, 1]}).}
    \item{\code{ci_width_mean, ci_width_max}}{At \code{ci_level}.}
    \item{\code{ratio_mean, ratio_max}}{CI width / plausible range.}
    \item{\code{horizon_type}}{One of \code{"observed"}, \code{"projected"},
      \code{"lower_bound"}.}
    \item{\code{horizon}}{Numeric horizon (observed or projected) or
      \code{NA} for lower-bound rows.}
    \item{\code{horizon_description}}{The long description returned by
      \code{shelf_life}.}
  }
  The returned object carries the original call and \code{response_scale}
  as attributes for use by \code{\link{et_plot_sensitivity}}.
}
\description{
Sweeps a grid of noise magnitudes, re-runs \code{\link{et_predict}} for
each, and summarises how the environmental variance component and the
forecast shelf life respond.  This turns the subjective choice of
\code{env_noise} into an auditable \emph{curve}: instead of reporting a
single shelf life at one (often hand-picked) measurement-error level, the
user sees how the horizon degrades as assumed noise grows.
}
\details{
Three argument styles are supported for the grid:
\itemize{
  \item \code{fraction_grid}: scalar fractions of each predictor's SD in
    \code{newdata}.  These are passed straight to the scalar form of
    \code{\link{et_predict}}'s \code{env_noise}.  Use this for an apples-
    to-apples sweep that scales all predictors together.
  \item \code{absolute_grid}: a \code{list} of named numeric lists /
    vectors; each list element becomes a single \code{env_noise} argument
    (so you can sweep absolute SDs for one predictor while holding others
    fixed).
  \item Neither supplied (\code{NULL}): a default log-spaced fraction grid
    \code{c(0, 0.01, 0.025, 0.05, 0.1, 0.2, 0.4, 0.8)} is used.
}

Each iteration calls \code{et_predict(..., env_noise = g)} then
\code{\link{shelf_life}} and records:
\itemize{
  \item Mean parameter / environmental / residual / total variance across
    forecast observations.
  \item The horizon description (\code{"observed"}, \code{"projected"}, or
    \code{"lower_bound"}) and numeric horizon if any.
  \item Mean / max CI width / plausible-range ratio.
}
}
\seealso{
\code{\link{et_predict}}, \code{\link{shelf_life}},
  \code{\link{et_plot_sensitivity}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{et_plot_sensitivity}
\alias{et_plot_sensitivity}
\title{Plot a sensitivity profile}
\usage{
et_plot_sensitivity(sens, show = c("horizon", "env_share", "ratio"))
}
\arguments{
\item{sens}{An \code{et_sensitivity} object from
\code{\link{et_sensitivity_profile}}.}

\item{show}{\code{"horizon"} (default) shows the shelf-life horizon;
\code{"env_share"} shows \code{env_var / total_var}; \code{"ratio"}
shows the mean CI width / plausible range.}
}
\value{
A \code{ggplot2} object.
}
\description{
Visualises the output of \code{\link{et_sensitivity_profile}}: for each
noise grid point, shows how the \strong{environmental share} of total
variance and the \strong{forecast horizon} respond.  Observed horizons
are drawn as solid points; projected horizons are hollow points;
lower-bound rows are drawn as upward arrows at the last informative time.
}
\details{
The x-axis is the noise fraction (when the grid was built from
\code{fraction_grid}) or the grid step label otherwise.  When both a
numeric fraction and a descriptive label exist, the fraction is preferred
for continuous x-positioning.
}
\seealso{
\code{\link{et_sensitivity_profile}}
}
